unit UnitACPopupClicks;

interface

uses UnitClipQueue, UnitFrmConfig, UnitFramePopup, System.Types;

type TACPopupClicks = class(TObject)
    private
        fLastClip : TClipItem;
        fLastString : string;
        MenuPoint : TPoint;
    protected
        procedure SendText(s : string; ci : TClipItem=nil; noplaintext : boolean = false);
        function HandleKeypress(ca: TClickAction; p : TACPopupItem): boolean;
        function IsKeypressExecute(p : TACPopupItem) : boolean;
    public
    	function StayOpenOnClick(p : TACPopupItem) : boolean;
        function PinClick(p : TACPopupItem) : boolean;
        procedure StayOpenClick(p : TACPopupItem);

        procedure SetMenuPoint(p : TPoint);

    	procedure PasteAllClick(Sender : TObject);
        procedure PopupClipClick(Sender : TObject);
        procedure RecentlyRemovedClicke(Sender : TObject);
        procedure PasteMethodClick(Sender : TObject);
        procedure PermanentClipClick(Sender : TObject);
        procedure PermanentGroupClick(Sender : TObject);

        procedure ProgramOptionDisableMonitoring(Sender : TOBject);
        procedure ProgramOptionDisableRightClickPopup(Sender : TObject);

        procedure ToggleClipboardBar(Sender : TObject);
       	procedure EraseHistory(Sender : TObject);
        procedure PopupClipRightClip(Sender : TObject);
        procedure OpenWindowCommand(Sender : TObject);
        procedure RightClickHello(Sender : TObject);
        procedure ClipTypeClick(Sender : TObject);
end;

var ACPopupClicks : TACPopupClicks;

implementation

uses Classes, UnitMisc, UnitPaste, Windows, UnitPopupGenerate,
StrUtils, SysUtils, UnitFrmPermanentNew, UnitFrmClipboardManager, Graphics, UnitFrmEditTextExternal,
UnitMenuItemTagdata, UnitFrmSysTrayMenu, Forms, UnitFrmAbout, UnitFrmSearch,
UnitFrmNewEditHistory, UnitFrmPasteSelected, UnitFrmDummyClipboardBar,
Clipbrd, dialogs, Controls, Menus, UnitKeyboardQuery, UnitTWideChar,
  UnitClipMenu, UnitFrmTooltipNew;


function TACPopupClicks.PinClick(p : TACPopupItem) : boolean;
const theset = [CA_Pin_Unpin_Clip];
begin
	result := false;
	if p.ItemType in TPastesOnClickj then begin
		result := (p.CtrlPressed) and (FrmConfig.GetCtrlClickAction in theset);
        result := result or
        		((p.ShiftPressed) and (FrmConfig.GetShiftClickAction in theset));
        result := result or
            ((p.MiddleClicked) and (FrmConfig.GetMiddleClickAction in theset))
    end;
end;

function TACPopupClicks.StayOpenOnClick(p : TACPopupItem) : boolean;
const theset = [CA_Paste_and_Keep_Popup_Open];
begin
	result := false;
	if p.ItemType in TPastesOnClickj then begin
		result := (p.CtrlPressed) and (FrmConfig.GetCtrlClickAction in theset);
        result := result or
        		((p.ShiftPressed) and (FrmConfig.GetShiftClickAction in theset));
        result := result or
            ((p.MiddleClicked) and (FrmConfig.GetMiddleClickAction in theset));
        result := result or
            ((p.AltPressed) and (FrmConfig.GetAltClickAction in theset));
    end;
end;
procedure TACPopupClicks.StayOpenClick(p : TACPopupItem);
begin
	if p.CtrlPressed then begin
    	p.CtrlPressed := false;
        Paste.SendCTRL_Release;
    end else if p.ShiftPressed then begin
    	p.ShiftPressed := false;
        Paste.SendShift_Release;
    end else if p.MiddleClicked then begin
        p.MiddleClicked := false;
    end else if p.AltPressed then begin
        p.AltPressed := false;
        Paste.SendALT_Release;
    end;

    p.OnClick(p);
end;
function TACPopupClicks.IsKeypressExecute(p : TACPopupItem): boolean;
var
    ca : TClickAction;
    function TestForMimicWait(ca : TClickAction; vk : integer) : boolean;
    begin
        result := false;
        if ca = CA_Paste_using_Mimic_Typing then begin
            KeyboardQuery.WaitUntilRelease(vk);
            result := true;
        end;
    end;
begin
    result := false;
    if p.ctrlpressed or p.shiftpressed or p.MiddleClicked or p.AltPressed then begin
        result := true;
        if p.ctrlpressed then begin
            ca := FrmConfig.GetCtrlClickAction;
            if not TestForMimicWait(ca, VK_CONTROL) then begin
                Paste.SendCTRL_Release;
            end;
            self.HandleKeypress(ca, p);
        end;
        if p.shiftpressed then begin
            ca := FrmConfig.GetShiftClickAction;
            if not TestForMimicWait(ca, VK_SHIFT) then begin
        	    Paste.SendShift_Release;
            end;
            self.HandleKeypress(ca, p);
        end;
        if p.AltPressed then begin
            ca := FrmConfig.GetAltClickAction;
            if not TestForMimicWait(ca, VK_MENU) then begin
                Paste.SendALT_Release;
            end;
            self.HandleKeypress(ca, p)
        end;
        if p.MiddleClicked then begin
            self.HandleKeypress(FrmConfig.GetMiddleClickAction, p);
        end;
    end;
end;

procedure TACPopupClicks.OpenWindowCommand(Sender: TObject);
var ow : TACOpenWindowFlag;
    procedure ShowForm(f : TForm);
    begin
        f.show;
        ForceForeground(f.Handle )
    end;


begin
    FrmMainPopup.SkipFocusReturnOnce;
    with Sender as TACPopupItem do begin
        ow := TACOpenWindowFlag(IntegerData);
        case ow of
        owConfiguration: ShowForm(FrmConfig);
        owEditHistory: begin
            frmEditHistory.ShowPopup;
            ShowForm(frmEditHistory);
        end;
        owRemovedClips: begin
            frmEditHistory.ShowRemoved;
            ShowForm(frmEditHistory);
        end;
        owPermanentClips: ShowForm(FrmPermanent);
        owPasteSelected: FrmPasteSelected.ShowaAutomated;
        owSearch: begin
            frmSearch.ShowAutomatted(FrmMainPopup.TargetData.ForegroundWindow);
        end;
        owEditClipboard: begin
            frmEditTextExternal.EditClip(clip);
        end;
        owAbout: ShowForm(frmAbout);
        end;
    end;
end;

procedure TACPopupClicks.PasteAllClick(Sender: TObject);
var
	s : string;
	i : integer;
    ci : TClipItem;
    wc : TWideChar;
    cft : TClipFormatType;
begin
    with Sender as TACPopupItem do begin
        s := '';
        for i := 0 to (ClipQueue.GetQueueCount  - 1) do begin
            ci := ClipQueue.GetClipItem(i);

            s := s + ci.GetAsPlaintext;

            if (Cardinal(i) = (ClipQueue.GetQueueCount  - 1)) then begin
            end else begin
                s := s + #13#10;
            end;
        end;

        self.SendText(s, nil);
    end;
end;
procedure TACPopupClicks.PasteMethodClick(Sender: TObject);
begin
    with Sender as TACPopupItem do begin
        Paste.AssignPaste(FrmMainPopup.TargetData.ExeName,TPasteMethod(IntegerData));

        case TPasteMethod(IntegerData) of
        PASTE_DEFAULT: Paste.ClearOnceFlags;
        PASTE_CTRL_V: Paste.SetUsePasteCVOnce;
        PASTE_SHIFT_INS: paste.SetUsePasteSIOnce;
        PASTE_MIMIC: paste.SetKeyboardMimicOnce;
        PASTE_CLIPBOARD: paste.SetClipboardOnlyOnce;
        end;
    end;
end;
procedure TACPopupClicks.SetMenuPoint(p : TPoint);
begin
    MenuPoint := p;
end;

procedure TACPopupClicks.PermanentClipClick(Sender: TObject);
var
    p : TACPopupItem;
    ci : TClipITem;
begin
    p := TACPopupItem(sender);
    with Sender as TACPopupItem do begin
        if IsKeypressExecute(p) then EXIT;

        if PermanentGroupID <> -1 then begin
            FrmPermanent.PermFolderPush;
            FrmPermanent.SetPermanentPath(FrmPermanent.PermFoldersGetItem(PermanentGroupID));
        end;
        ci := nil;
        if FrmPermanent.IsComplexItem(p.PermanentText)  then begin
        	ci := FrmPermanent.GetComplexItem(p.PermanentText);
        end;
        if PermanentGroupID <> -1 then begin
            FrmPermanent.PermFolderPop;
        end;

        if ci <> nil then begin
			self.SendText('',ci);
            myfree(ci);
        end else begin
        	self.SendText(p.PermanentText);
        end;


    end;
end;
procedure TACPopupClicks.PermanentGroupClick(Sender: TObject);
begin
     with Sender as TACPopupItem do begin
         FrmPermanent.SetPermanentGroup(PermanentGroupID);
         FrmPermanent.AssignPIG(FrmMainPopup.TargetData.ExeName); 
     end;
end;
procedure TACPopupClicks.PopupClipClick(Sender: TObject);
var pasted : boolean;
begin


    with Sender as TACPopupItem do begin

    self.fLastClip := clip;
    if IsKeypressExecute(TACPopupItem(sender)) then EXIT;

    if clip.HasText or (clip.GetFormat = CF_HDROP)  then begin
        fLastString := clip.GetAsPlaintext;
    end else begin
        fLastString := '';
        //fLastString := clip.GetAsText;
    end;


    UnitMisc.AppendLog('MenuItem format=' + clip.GetFormatName );


     if frmconfig.cbMoveFirst.Checked then
        ClipQueue.MoveToStart( ClipQueue.IndexOf(clip) );

// This is being depricated
//     ClickedClipQueue.InsertAtStart( clip );
     pasted := false;
    {Force as clip only if it's unicode, or if Formatted is the default}
    if (clip  <> nil) and (clip.GetFormatType  <> FT_UNKNOWN) then begin

        // handle FormattedOnly and PlainAndFormatted here
        if (not Paste.GetKeyboardMimicOnce) then begin
            if (not FrmConfig.rbPlainOnly.Checked) and
            (clip.GetFormatType in [FT_UNICODE, FT_RICHTEXT, FT_HTML]) then begin
                UnitMisc.AppendLog('Sending UNICODE or Formatted');
                self.SendText('', clip, FrmConfig.rbFormattedOnly.Checked);
                pasted := true;
            end;
        end;
    end;
    {}
    if (not pasted) then begin

        if fLastString = '' then begin
            // TODO: check for non-text item paste method
            UnitMisc.AppendLog('Sending as Normal clip');
            self.SendText('', clip);
        end else begin
        	UnitMisc.AppendLog('forching Plain text only (CF_HDROP)');
            self.SendText(fLastString);
        end;
        UnitMisc.AppendLog('inserting done');
    end;

    end;
end;

procedure TACPopupClicks.PopupClipRightClip(Sender: TObject);
begin
     with Sender as TACPopupItem do begin
     end;
end;

procedure TACPopupClicks.ProgramOptionDisableMonitoring(Sender: TOBject);
begin
    FrmSysTrayMenu.ToggletDisabledEXE(FrmMainPopup.TargetData.ExeName);
end;
procedure TACPopupClicks.ProgramOptionDisableRightClickPopup(Sender: TObject);
begin
    frmSysTrayMenu.ToggleNoRightClickEXE(FrmMainPopup.TargetData.ExeName);
end;

procedure TACPopupClicks.RecentlyRemovedClicke(Sender: TObject);
begin
    with Sender as TACPopupItem do begin
        if IsKeypressExecute(TACPopupItem(sender)) then EXIT;

        if frmconfig.cbMoveFirst.checked then begin
            frmClipboardManager.DisablePasteProtectionOnce;
        end;
        self.SendText('', clip);
    end;
end;

procedure TACPopupClicks.RightClickHello(Sender: TObject);
begin

end;

procedure TACPopupClicks.ToggleClipboardBar(Sender: TObject);
begin
    with Sender as TACPopupItem do begin
    if assigned(frmClipboardBar) then begin
        if checked then frmconfig.LoadClipboardBarState;
        frmClipboardBar.SetSizePos(
        	BoundRect.Right,frmClipboardBar.Top,
            frmClipboardBar.width,frmClipboardBar.height
        );

        frmClipboardBar.SetVisible(Checked);
    end;
    end;
end;
procedure TACPopupClicks.ClipTypeClick(Sender: TObject);
begin
    with Sender as TACPopupItem do begin
        SendText('', clip, TRUE);
    end;
end;

procedure TACPopupClicks.EraseHistory(Sender: TObject);
var s : string;
begin

    with Sender as TACPopupItem do begin
        case TPopupEraseFlag(IntegerData) of
        peCLipboard: s := 'the Clipboard';
        pePopupClips: s := 'the Popup Clips';
        peRemovedClips: s := 'the Removed Clips';
        peEverything: s := 'All Clips';
        end;

        if MessageDlg('Erase '+s+'?',mtConfirmation,mbYesNo, 0) <> Controls.mrYes then EXIT;



        case TPopupEraseFlag(IntegerData) of
        peCLipboard: begin
            frmClipboardManager.ClearClipboard;
        end;
        pePopupClips: begin
            frmEditHistory.SkipConfirmationOnce;
            frmEditHistory.btnPopupFlush.click;
        end;
        peRemovedClips: begin
            frmEditHistory.SkipConfirmationOnce;
            frmEditHistory.btnRemovedFlushItems.click;
        end;
        peEverything: begin
            frmClipboardManager.ClearClipboard;
            frmEditHistory.SkipConfirmationOnce;
            frmEditHistory.btnPopupFlush.click;
            frmEditHistory.SkipConfirmationOnce;
            frmEditHistory.btnRemovedFlushItems.click;
        end;
        end;
    end;
end;

procedure TACPopupClicks.SendText(s: string; ci: TClipItem=nil; noplaintext : boolean=false);
var prefix : string;
    b : boolean;
    tt : TFrmTooltipNew;
begin
    Windows.SetLastError(ERROR_SUCCESS);

    //
    // Since the popup may have stolen focus from the target,
    // it must be given keyboard focus again
    //

    if (FrmMainPopup.GetNeedsFocus) then begin

        b := (UnitMisc.ThreadAttach(FrmMainPopup.TargetData.ForegroundWindow));
        Windows.SetForegroundWindow(FrmMainPopup.TargetData.ForegroundWindow);
        if (FrmMainPopup.TargetData.focuscontrol <> 0) then begin
            Windows.SetFocus(FrmMainPopup.TargetData.focuscontrol);
        end;
        if b then ThreadDetach;
        if FrmMainPopup.TargetData.ForegroundWindow <> GetForegroundWindow then begin
            ForceForeground(FrmMainPopup.TargetData.ForegroundWindow);
        end;

//        UnitMisc.AppendLog('Returning Focus to - ' + WindowHandleToEXEName(FrmMainPopup.TargetData.ForegroundWindow));
//        UnitMisc.AppendLog('Current Foreground - ' + WindowHandleToEXEName(GetForegroundWindow));
    end;

    b := Paste.GetClipboardOnlyOnce;
    prefix :=uppercase(leftstr(s,6));
    if (prefix  = '[KEYS]') then begin
        Paste.SendTextWithKeystrokes(s);
    end else begin
        Paste.SendText(s, ci, noplaintext);
    end;

    if (Paste.GetPasteMethod(FrmMainPopup.TargetData.ExeName) = PASTE_CLIPBOARD)
        or b
    then begin
        tt := TFrmTooltipNew.Create(application);

        tt.HideHeader;
        tt.CloseOnLostFocus := False;
        tt.SmallFontOnce := true;
        tt.OnClick := nil;
        tt.DodgeMouse := false;
        tt.ShowTooltip('  Clipboard Only'+ WideChar($00A0)+ WideChar($00A0), MenuPoint);
//        ForceForeground(tt.Handle);
        MySleep(800);
        tt.Hide;
    end;

    FrmMainPopup.SkipFocusReturnOnce;
    if (FrmMainPopup.GetNeedsFocus) then begin
        UnitMisc.ThreadDetach;
    end;
end;

function TACPopupClicks.HandleKeypress(ca: TClickAction; p : TACPopupItem): boolean;
var
    i : integer;
    procedure MoveToFirstCheck;
    begin
        if not frmConfig.cbMoveFirst.Checked then EXIT;

        case p.ItemType of
        IT_MENU:
                ClipQueue.MoveToStart( ClipQueue.IndexOf(p.Clip) );
        IT_TEMP:
                frmClipboardManager.DisablePasteProtectionOnce;
        end;
    end;
begin

    result := true;
    case ca of
        CA_Run:; //self.Run1Click(self);
        CA_RemoveUndelete: begin
            if p.clip = nil then EXIT;

            if p.ItemType = IT_TEMP then begin
                // Undelete
                i := TempClipQueue.IndexOf(p.clip);
                if (i <> -1) then begin
                    TempClipQueue.DeleteItem2(i, true); {delete without a free}
                    ClipQueue.InsertAtStart(p.clip);
                end;
            end else begin
                if (ClipQueue.IndexOf(p.clip) = -1)  then begin
                    // Can't Undelete or Remove
                end else begin
                    i := ClipQueue.IndexOf(p.clip);
                     ClipQueue.DeleteItem(i);
                end;
            end;
        end;
        CA_Destroy_Clip: begin
            // Destroy Item -- bad programmer being lazy and re-using and old menu item
            if p.Clip = nil then EXIT;
            ForceForeground(FrmMainPopup.Handle);
            if (MessageDlg('Destroy: Permanently delete this item?', mtConfirmation, mbYesNo,0) = mrYes) then begin
                if ClipQueue.IndexOf(p.clip) <> -1 then begin
                    clipqueue.DestroyItem(ClipQueue.IndexOf(p.clip));
                end else if TempClipQueue.IndexOf(p.clip) <> -1  then begin
                    TempClipQueue.DeleteItem(TempClipQueue.IndexOf(p.clip));
                end;
            end;
        end;                        //self.DestroyClick(self);
        CA_Edit_Clip: Begin
            frmEditTextExternal.EditClip(p.clip);
        end;                           //self.EditText1Click(self);
        CA_Edit_Clip_as_file: begin
            frmEditTextExternal.EditClip(p.clip);
        end;
        CA_Paste_using_Mimic_Typing: begin      {self.MimicTyping1Click(self); result := false;}
            Paste.SetKeyboardMimicOnce;
            self.SendText('', p.clip);
        end;
        CA_Paste_on_Clipboard_only: begin       {self.PasteusingMimicTyping1Click(self); result := false;}
            Paste.SetClipboardOnlyOnce;
            MoveToFirstCheck;
            self.SendText('', p.clip);
        end;
        CA_Paste_using_ShiftInsert: begin       {self.PasteusingSHIFTINS1Click(self); result := false;}
            Paste.SetUsePasteSIOnce;
            MoveToFirstCheck;
            self.SendText('', p.clip);
        end;
        CA_Paste_using_CtrlV: begin             {self.PasteusingCTRLV1Click(self); result := false;}
            paste.SetUsePasteCVOnce;
            MoveToFirstCheck;
            self.SendText('', p.clip);
        end;
        CA_Make_Permanent_Clip: begin
            if p.ItemType <> IT_PERMANENT then
                FrmPermanent.ShowWithNewComplexItem(p.clip);
        end;
        CA_Make_Permanent_Clip_as_Plain_Text: begin
            if p.ItemType <> IT_PERMANENT then
                FrmPermanent.ShowWithNewItem(p.clip.GetAsPlaintext);
        end;
        CA_Show_RightClick_Menu: begin

        end;
        CA_Paste_as_Plaintext_only: begin
            MoveToFirstCheck;
            self.SendText(p.Clip.GetAsPlaintext, nil);
        end;
        CA_Pin_Unpin_Clip: begin
            case p.ItemType of
            IT_MENU: begin
                PinnedClipQueue.InsertAtStart(p.clip);
            end;
            IT_PINNED: begin
                PinnedClipQueue.DeleteItem( PinnedClipQueue.IndexOf(p.clip));
            end;
            end;
        end;
        CA_Paste_as_Formatted_only: begin
            ACPopupClicks.ClipTypeClick(p);
        end
    else
        result := false;
    end;
end;


initialization
begin
    ACPopupClicks := TACPopupClicks.Create;
end;
end.
