unit UnitFramePopup;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, ComCtrls, UnitClipQueue,
  UnitMenuItemTagdata, unitMisc, Contnrs, UnitClipMenu, UnitTWideChar,
  UnitFrmTooltipNew;

const USE_POPUP_KEYBOARD_HOOK = FALSE;

const TUsesClipMenu = [IT_TEMP, IT_MENU, IT_CLIPBOARD, IT_CLICKED, IT_PERMANENT, IT_PINNED];
const TContainsAClip = [IT_TEMP, IT_MENU, IT_CLIPBOARD, IT_CLICKED, IT_PINNED];
const TPastesOnClickj = TUsesClipMenu;
const TClickableLeftIcons = TUsesClipMenu + [IT_FULL];
const TDoubleSized = [IT_MENU, IT_CLIPBOARD, IT_PINNED, IT_TEMP];

type
TACPopupItem = class;
TACPopupItemCollection = class;
TACPopupItemClass = class of TACPopupItem;
TACPopup = class;


TPopupItemPosition = (
    pipLeftIcon, pipRightIcon, pipCaption, pipNone
);

TPopupItemStyle = (
    psNormal, psLine, psBreak, psSubmenu, psExpandable, psCheckable, psGap
);

TPopupEraseFlag = (
    peCLipboard, pePopupClips, peRemovedClips, peEverything
);

TACOpenWindowFlag = (
    owConfiguration, owEditHistory, owRemovedClips, owPermanentClips,
    owPasteSelected, owSearch, owAbout, owEditClipboard
);



TCustomTransparentControl = class(TCustomControl)
  private
    FInterceptMouse: Boolean;
  protected
    procedure WMNCHitTest(var Message: TWMNCHitTest); message WM_NCHITTEST;
    procedure CreateParams(var Params: TCreateParams); override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Invalidate; override;
    property InterceptMouse: Boolean read FInterceptMouse write FInterceptMouse default False;
  end;

TACPopup = class(TCustomTransparentControl)
    procedure CreateParams(var Params: TCreateParams);  override;
private
	crit : RTL_CRITICAL_SECTION;
    scroll : THandle;
    timHover: TTimer;
    timShowSubmenu: TTimer;
    timHideSubmenu: TTimer;
    timLeftIconHover : TTimer;
    timIgnoreMouseMove : TTimer;
    lastHitTest : integer;
    ignoreHitTestOnce : boolean;

    FirstPaint : boolean;
    FormMode : boolean;
    LastMouseX, LastMouseY : integer;

    fItems : TACPopupItem;
    fLastVisible : TACPopupItem;

    fKeystrokes : TStringList;
    fAltKeyDown : boolean;
    fSpacing : integer;
    fMaxItemWidth : integer;

    fFullMode : boolean;
    fHover : TACPopupItem;
    fDownShift : TShiftState;
    FLastPosition : TPopupItemPosition;
    //fToolTip : TToolTipWindow;
    fToolTipNew : TFrmTooltipNew;

    //fLastItemRect : TRect;
    fAcceleratorCount : integer;
    fAccelCharsUsed : string;
    fReservedCount : integer;
//    fClipboard : TClipItemCurrentClipboard;
    fLastSubmenu : TACPopupItem;
    fQueuedSubmenu : TACPopupItem;
    fMainRect : TRect;


    fPoint : TPoint;
    fTopFixed, fLeftFixed : boolean;
    fim : TClipMenu;
    region1, region2 : HRGN;
    fMouseDownOn : TACPopupItem;

    fNewCanvas : TCanvas;
    fDragStarted : boolean;
    fDragMenuItem : TACPopupItem;
    fDropMenuItem : TACPopupItem;
    fDragX, fDragY : integer;

    fOnHideEvent : TNotifyEvent;
    fShowing : boolean;
    fIncludeEmpty : boolean;

    fIsTooTall : boolean;
    fIgnoreMouseMove : boolean;
    fIgnoreKeypresses : boolean;
    fLastKeyDown : word;

    fBackgroundColor,
    fHighlightColor,
    fFontColor,
    fClickedColor,
    fDisabledColor,
    fExpandedBackgroundColor,
    fColumnColor,
    fBorderColor,
    fColumnEdgeColor : TColor;

    fGroup : TObjectList;
    fUseGrouping : boolean;
    fGroupingCount : integer;
    fCurrentGroup : Integer;
    fUseGroupingOnce : boolean;
    fUseGroupingOnceFailed : boolean;
    fAutoExpandOnce : TItemType;

    fLastClipSequence : cardinal;
    fTempBitmap : TBitmap;

    fCaptionPoint : TPoint;

    fLastExpandedSet : set of TItemType;

    fKeyActivated : boolean;
    fShowExpandedShortcuts : Boolean;
    fShowExpandedType : TItemType;
    fExpandedKeystrokes : TStringList;

    function GetClipHeader(ci : TClipItem) : string;

    procedure timHoverTimer(Sender: TObject);
    procedure timShowSubmenuTimer(Sender: TObject);
    procedure timHideSubmenuTimer(Sender: TObject);
    procedure timLeftIconHoverTimer(Sender: TObject);
    procedure timIgnoreMouseMoveTimer(Sender: TObject);
    procedure WMGetDlgCode(var Msg: TMessage); message WM_GETDLGCODE;
    procedure WMActivate(Var msg:tMessage); message WM_ACTIVATE;
    procedure WMNCHitTest(var Message: TWMNCHitTest); message WM_NCHITTEST;
    procedure WMSetCursor(Var M: TWMSetCursor); message WM_SETCURSOR;
	procedure WMMove(var Msg: TMessage) ; message WM_MOVE;

    function CalcWidth(fpi : TACPopupItem; ClipIt : boolean=true) : integer;
    function CalcHeight(fpi : TACPopupItem) : integer;
    procedure CalcTotalSize(FixSubmenu:boolean=false);
    procedure EnsureVisibility(FixSubmenu:boolean=false);
    procedure WMKeyDown(var Message: TWMKeyDown);
    procedure MouseLeave(Sender: TObject);
    function IsLegalDrop(p : TACPopupItem) : boolean;
    function GetAccelerator(i: integer): char;
    function GetAcceleratorPopupClips(i: integer) : Char;
    procedure ReserveAccel;
    procedure AddAScrollbar;
    procedure DetectAutoScroll(top,bottom : integer);

    procedure CheckForAutoExpand(p : TACPopupItem);

    procedure SaveExpandedState;
    procedure LoadExpandedState;

    procedure AssignSubmenuShortcuts(p : TACPopupItem);

    function UseKeyboard : boolean;
protected
    procedure AddPopupClips;
    procedure AddCancel;
    procedure AddFullConfigured;
    procedure AddCurrentClipboard;
    procedure AddLast;
    procedure AddCurrentPermanentItems;
    procedure AddSearch;
    procedure AddRecentlyRemoved;
    procedure AddPastingTools;
    procedure AddProgramOptions;
    procedure AddRecentlyClicked;
    procedure AddSwitchToItems;
    procedure AddSwitchToSubmenu;
    procedure AddAllItemsSubmenu;
    procedure AddSystem;

    procedure paint; override;
    procedure resize; override;
    {procedure invalidate; override;}


    procedure SetPrefix(p : TACPopupItem);
    procedure DetectKeystroke(p : TACPopupItem);
    procedure DrawItem(p : TACPopupItem);
    procedure DrawCheckgroup(p : TACPopupItem);
    procedure FullRedraw;
    procedure AddSubmenu(p : TACPopupItem);
    procedure DrawDragmark;

    procedure ClipMenuShow(rightclick : boolean = false);
    procedure ClipMenuHide;
    procedure ClipMenuReportKey(key : Char); overload;
    procedure ClipMenuReportKey(key : word); overload;
    function ClipMenuVisible : Boolean;
    procedure HoverUp;
    procedure HoverLeft;
    procedure HoverRight;
    procedure HandleOnClick(p : TACPopupItem);
    procedure HandleMouseClick(p : TACPopupItem);
    procedure HandleCaptionClick(p : TACPopupItem);
    procedure HoverClick;
    procedure ShowSubmenuDelayed(p : TACPopupItem);
    procedure CancelShowSubmenu;
    procedure MouseMove(Shift: TShiftState; X: Integer; Y: Integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X: Integer; Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X: Integer; Y: Integer); override;


    function GetItemAt(x,y : integer) : TACPopupItem;

    procedure SetHover(value : TACPopupItem);
    property Hover : TACPopupItem read fHover;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;

    procedure EditCallback(Sender : Tobject);
    procedure DeleteClipCallback(Sender : TObject);
    procedure DestroyCallback(Sender : Tobject);
    procedure PasteCallback(Sender : Tobject);
    procedure MakePermanentCallback(Sender : Tobject);



    procedure DragMouseDown(p : TACPopupItem);
    function DragMouseMove(p, last : TACPopupItem) : boolean;
    function DragMouseUp(p : TACPopupItem) : boolean;


    function GetCanvas : TCanvas;
    procedure SetShowExpandedShortcuts(value : Boolean);
    property ShowExpandedShortcuts : boolean read fShowExpandedShortcuts write SetShowExpandedShortcuts;
public

    { Creation }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;  override;

    function Add : TACPopupItem;
    function AddLine : TACPopupItem;
    function AddBreak : TACPopupItem;

    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;

    procedure AutoPopulate;
    procedure Clear;
    procedure ShowPopup(x,y : integer); overload;
    procedure ShowPopup(pt : TPoint); overload;
    procedure DodgePoint(pt : TPoint);
    procedure HoverDown;
    procedure Hide;

    {External Draw Routines}
    procedure SetCanvas(c : TCanvas);
    procedure DrawOnCanvas(fullmode : boolean=false);
    property UseCanvas : TCanvas read GetCanvas;
    property IncludeEmpty : boolean read fIncludeEmpty write fIncludeEmpty;

    { Cosmetics }
    property Spacing : integer read fSPacing write fSpacing;
    property MaxItemWidth : integer read fMaxItemWidth write fMaxItemWidth;
    property OnHideEvent : TNotifyEvent read fOnHideEvent write fOnHideEvent;
    property Showing : boolean read fShowing;
    property FullMode : boolean read fFullMode write fFullMode;
    property UseGrouping : boolean read fUseGrouping write fUseGrouping;
    property GroupingCount : integer read fGroupingCount write fGroupingCount;
end;

TACPopupItem = class(TObject)
private
    fClip : TClipItem;
    fIcon : HICON;
    fLeftBitmap : TBitmap;
    fRightBitmap : TBitmap;
    fCaption : string;
    fPrefix : string;
    fHasRightIcon : boolean;
    fHasLeftIcon : boolean;
    fLeftIconOnlyOnHover : boolean;
    fItemType : TItemType;
    fVisible : boolean;

    fPermanentID, fPermanentGroupID : integer;
    fPermanentText : string;

    fMaxHeight : integer;
    fColWidth : TIntList;
    fSubMenu : TACPopupItemCollection;
    fCheckgroup : TObjectList;
    fParent : TACPopupItem;
    fMyColumn : integer;

    fSubShowing : TACPopupItem;

    fExpanded : boolean;
    fCollapseParentOnClick : boolean;
    fChecked : boolean;

    fStyle : TPopupItemStyle;
    fBoundsRect : TRect;
    fDoubleHeight : boolean;
    fDisabled : boolean;
    fShowCollapseIcon : Boolean;
protected
    fBottom : integer;
    fLeft : integer;
    fTop : integer;
    fRight : integer;
    fOnClick : TNotifyEvent;
    fOnRightClick : TNotifyEvent;
    fctrlpressed, fShiftPressed, fMiddleClicked, fAltpressed : boolean;
    fIntegerData : integer;
    fHint : string;
    
    procedure SetCaption(value : string);
    procedure SetLeftIcon(value : hicon);
    procedure SetLeftBitmap(value : TBitmap);
    procedure SetRightIcon(value : TBitmap);
    function CalcRight : integer;
    function GetIsInSubMenu : boolean;
    function GetIsInExpandedMenu : boolean;
    function GetColumnWidth : integer;
    function GetChecked : boolean;
    procedure SetChecked(value : boolean);

    function GetIsLine : boolean;
    procedure SetIsLine(value : boolean);
    function GetIsBreak : boolean;
    procedure SetIsBreak(value : boolean);
    function GetIsExpandable : boolean;
    procedure SetIsExpandable(value : boolean);
    function GetIsSubmenu : boolean;
    procedure SetIsSubmenu(value : boolean);
    function GetIsGap : boolean;
    procedure SetIsGap(value : boolean);

    function GetItem(Index : integer) : TACPopupItem;
    procedure SetItem(Index : integer; value : TACPopupItem);
    function GetExpanded : boolean;
    procedure SetExpanded(value : boolean);
    function GetIndex : integer;
    function GetOwnerIndex : integer;
    function GetVisibleParent : TACPopupItem;

    procedure SetBoundsRect(parenttop, parentleft : integer);
    function GetBoundsRect : TRect;
    procedure SetCaptionFromClip(value : TClipItem);


public

    constructor Create();
    destructor Destroy; override;
    function Add : TACPopupItem;
    function AddLine : TACPopupITem;
    function AddBreak : TACPopupITem;
    procedure CheckGrouped;
    function IsCheckGrouped : boolean;
    procedure SetShowingSubMenu;
    property Right : integer read CalcRight;
    property Left : integer read fLeft;
    property Top : integer read fTOp;
    property Bottom : integer read fBottom;
    property BoundRect : TRect read GetBoundsRect;
    property Clip : TClipItem read fClip write fClip;
    property LeftIcon : HICON read fIcon write SetLeftIcon;
    property LeftBitmap : TBitmap write SetLeftBitmap;
    property Caption : String read fCaption write SetCaption;
    property CaptionFromClip : TClipItem write SetCaptionFromClip;
    property RightIcon : TBitmap read fRightBitmap write SetRightIcon;

    property IsLine : boolean read GetIsLine write SetIsLine;
    property IsBreak : boolean read GetIsBreak write SetIsBreak;
    property IsExpandable : boolean read GetIsExpandable write SetIsExpandable;
    property IsSubmenu : boolean read GetIsSubmenu write SetIsSubmenu;
    property isGap : boolean read GetIsGap write SetIsGap;
    property Checked : boolean read GetChecked write SetChecked;
    property Expanded : boolean read GetExpanded write SetExpanded;
    property Style : TPopupItemStyle read fStyle;

    property OnClick : TNotifyEvent read fOnClick write fOnClick;
    property OnRightClick : TNotifyEvent read fOnRightClick write fOnRightClick;

    property Prefix : string read fPrefix write fPrefix;
    property ItemType : TItemType read fItemType write fItemType;
    property Visible : boolean read fVisible write fVisible;
    property IsInSubmenu : boolean read GetIsInSubMenu;
    property IsInExpandedMenu : boolean read GetIsInExpandedMenu;
    property ColumnWidth : integer read GetColumnWidth;
    property ColumnIndex : integer read fMyColumn;
    property Items[Index : integer] : TACPopupItem read GetItem write Setitem; default;
    property Index : integer read GetIndex;
    property OwnerIndex : integer read GetOwnerIndex;
    property VisibleParent : TACPopupItem read GetVisibleParent;
    property CtrlPressed : boolean read fCtrlPressed write fCtrlPressed;
    property ShiftPressed : boolean read fShiftPressed write fShiftPressed;
    property MiddleClicked : Boolean read fMiddleClicked write fMiddleClicked;
    property AltPressed : boolean read fAltpressed write fAltpressed;

    property IntegerData : integer read fIntegerData;
    property PermanentGroupID : integer read fPermanentGroupID;
    property PermanentID : integer read fPermanentID;
    property PermanentText : string read fPermanentText;
    property Hint : string read fHint write fHint;
    property DoubleHeight : boolean read fDoubleHeight write fDoubleHeight;
    property Disabled : boolean read fDisabled write fDisabled;

    property ShowCollapseIcon : Boolean read fShowCollapseIcon write fShowCollapseIcon;
end;
TACPopupItemCollection = class(TObjectList)
protected
    function GetItem(Index: Integer): TACPopupItem;
    procedure SetItem(Index: Integer; const Value: TACPopupItem);
public
    function Add : TACPopupItem; overload;
    property Items[Index: Integer]: TACPopupItem read GetItem write SetItem; default;
end;



implementation

uses Math, GraphUtil,  UnitFrmConfig, UnitPopupGenerate,
  UnitFrmClipboardManager, StrUtils, UnitToken, UnitFrmPermanentNew,
  UnitFrmMainPopup, UnitPaste, UnitFrmSysTrayMenu, UnitACPopupClicks,
  UnitKeyboardQuery, Clipbrd, UnitFrmEditTextExternal, UnitFrmEditItem,
  UnitFrmDummyClipboardBar, UnitFrmClipMenuNew;

{$R *.dfm}


{ TACPopup }
procedure TACPopup.CreateParams(var Params: TCreateParams);
begin
    inherited CreateParams(params);

   	Params.Style := WS_POPUP {or WS_TABSTOP or WS_CLIPSIBLINGS};
    Params.ExStyle := WS_EX_TOOLWINDOW or WS_EX_TOPMOST {or WS_EX_LAYERED};


    {WS_EX_TRANSPARENT screws up in WinXP - causing a blank redraw}
end;

constructor TACPopup.Create(AOwner: TComponent);
begin
    self.InterceptMouse := true;
    self.Visible := false;
    inherited;
    if Win32MajorVersion >= 6 then begin
    	self.DoubleBuffered := true; // CANT use this with transparrent on XP
    end;



    self.OnExit := self.MouseLeave;

    self.Width := 1;
    self.Height := 1;
    
    fToolTipNew := TFrmTooltipNew.Create(application);
    timhover := TTimer.Create(self);
    timHover.Enabled := false;
    //timHover.Interval := 300;
    timHover.OnTimer := self.timHoverTimer;
    timHideSubmenu := TTimer.Create(self);
    timHideSubmenu.Enabled := false;
    timHideSubmenu.Interval := 300;
    timHideSubmenu.OnTimer := self.timHideSubmenuTimer;
    timShowSubmenu := TTimer.Create(self);
    timShowSubmenu.Enabled := false;
    timShowSubmenu.Interval := 300;
    timShowSubmenu.OnTimer := self.timShowSubmenuTimer;

    timLeftIconHover := TTimer.Create(self);
    timLeftIconHover.Enabled := false;
    timLeftIconHover.Interval := 190;
    timLeftIconHover.ontimer :=  timLeftIconHoverTimer;

    timIgnoreMouseMove := TTimer.Create(self);
    timIgnoreMouseMove.Enabled := false;
    timIgnoreMouseMove.Interval := 100;
    timIgnoreMouseMove.OnTimer := timIgnoreMouseMoveTimer;

    fItems := TACPopupItem.Create();
    fKeystrokes := TStringList.create;
    fExpandedKeystrokes := TStringList.Create;
    fExpandedKeystrokes.OwnsObjects := false;

    fTempBitmap := TBitmap.Create;

    fim := nil;

    self.fMaxItemWidth := 300;
    fSpacing := 2;
    self.Width := 1;
    self.Height := 1;

    self.ParentColor := false;         /// MUST BE FALSE for transparent
    self.ParentBackground := false;    /// ditoo++

    timHover.Interval := Application.HintPause;
    fBorderColor := saturate(cl3DDkShadow,0.10);

    self.fGroup := TObjectList.Create;
    self.fGroup.OwnsObjects := false;

    Windows.InitializeCriticalSection(crit);

    scroll := 0;
end;

destructor TACPopup.Destroy;
begin
    fToolTipNew.Close;
    Windows.DeleteCriticalSection(crit);
    inherited;
end;



function TACPopup.GetClipHeader(ci : TClipItem) : string;
begin
	result := 'right-click for Clip Menu';
	if ci <> nil then begin
    	result := result + '    ('+IntToStr(ci.GetDataSize div 2)+' chars)';
    end;
end;



function TACPopup.Add: TACPopupItem;
begin
    result := fItems.add;
end;



procedure TACPopup.ShowPopup(pt : TPoint);
begin
    self.ShowPopup(pt.x, pt.y);
end;
procedure TACPopup.ShowPopup(x, y: integer);
var c, t : Cardinal;
    procedure SetNoActivate;
    begin
        SetWindowLong(self.Handle, GWL_EXSTYLE,
            GetWindowLong(self.Handle, GWL_EXSTYLE) or WS_EX_NOACTIVATE);
    end;
    procedure ClearNoActivate;
    begin
        SetWindowLong(self.Handle, GWL_EXSTYLE,
            GetWindowLong(self.Handle, GWL_EXSTYLE) and not WS_EX_NOACTIVATE);
    end;
begin
	if fshowing then begin
    	appendlog('ACPopup.ShowPopup called twice');
    	EXIT;
    end;


    windows.EnterCriticalSection(crit);
    try
        t := Windows.GetTickCount;
        fShowing := true;
        fIgnoreMouseMove := false;
        fIsTooTall := false;
        fLastKeyDown := 0;
        fToolTipNew.MaxWidth := frmConfig.UDToolWidth.Position;
        fToolTipNew.MaxHeight := frmConfig.UDToolHeight.Position;
        AppendLog('acpopup start');

        self.fNewCanvas := self.Canvas;
        UseCanvas.Font := FrmConfig.GetFont;
        UseCanvas.Font.Size := FrmConfig.GetFont.Size;
        UseCanvas.Font.Style := FrmConfig.getfont.style;

        timHover.Interval := FrmConfig.tbTooltipDelay.Position;
        timHover.Enabled := false;

        fMaxItemWidth := FrmConfig.UDWidth.Position;
        fShowExpandedShortcuts := false;

        if fim = nil then begin
            fim := TClipMenu.Create(nil);
            fim.Hide;
            fim.OnRemoveClip := self.DeleteClipCallback;
            fim.OnDestroyClip := self.DeleteClipCallback;
        end;
        //fim.ParentWindow := GetDesktopWindow;

        fpoint.x := x;
        fpoint.y := y;
        self.Left := X;
        self.Top := Y;
        self.CalcTotalSize;
        self.fIgnoreKeypresses := true;



        if FrmMainPopup.GetNeedsFocus then begin
            ClearNoActivate;
        end else begin
            SetNoActivate;
        end;
        self.visible := true;
        Windows.SetWindowPos(
            self.Handle,HWND_TOPMOST,
            self.left,self.top,self.width,self.Height,
            SWP_SHOWWINDOW or SWP_NOZORDER or SWP_NOACTIVATE
        );
        Windows.WaitForInputIdle(Windows.GetCurrentProcess, INFINITE);

//        ForceForeground(self.handle);

        If FrmMainPopup.GetNeedsFocus then
            ForceForeground(self.handle);

        self.HoverDown;
        AppendLog('acpopup end - milliseconds=' + IntToStr(Windows.GetTickCount - t));

        self.AddAScrollbar;
    finally
        if fitems.fsubmenu.count = 0 then begin
            AppendLog('acpopup error - empty popup, hiding');
            self.fShowing := true;
            self.Hide;
        end;

        self.fIgnoreKeypresses := False;
        windows.LeaveCriticalSection(crit);
    end;
end;
procedure TACPopup.DodgePoint(pt : TPoint);
var m : TMonitor;
begin
    m := screen.MonitorFromPoint(pt);
    self.Top := pt.Y - 10 - self.Height;
    if (m <> nil) then begin
        if self.Top < m.Top then begin
            self.Top := m.Top;
        end;
    end;
end;

function SetLeftIcon(p : TACPopupItem; ci : TClipItem; forceType : boolean = false) : HICON;
var
    cdf : TClipFormatType;
begin


    if (frmconfig.cbCopyIcons.Checked = false) or forcetype then begin
        cdf := ci.GetFormatType;
        case cdf of
            FT_PICTURE:		p.LeftBitmap := FrmClipboardManager.iPic.Picture.Bitmap;
            FT_HTML: 		p.LeftBitmap  := FrmClipboardManager.iFormat.Picture.Bitmap;
            FT_RICHTEXT:	p.LeftBitmap := FrmClipboardManager.iFormat.Picture.Bitmap;
            FT_FILE: 		p.LeftBitmap := FrmClipboardManager.iFiles.Picture.Bitmap;
            else
            	p.LeftBitmap := FrmClipboardManager.iText.Picture.Bitmap;
        end;
    end else begin
        p.LeftIcon := ci.CData.GetHICON;
    end;
end;

const LEFT_SPACE = 18 + 4;
const LEFT_BORDER = 1;
const LEFT_EDGE = 2;
const RIGHT_SPACE = 20;
const RIGHT_BORDER = 1;
const TOP_BORDER = 1;
const BOTTOM_BORDER = 1;
const CAPTION_LEFT = LEFT_SPACE+LEFT_EDGE; {missing fSpacing}

function ImgToBitmap(img : TImage) : TBitmap;
begin
    result := img.Picture.Bitmap;
end;



{Popuplate Menu}
procedure TACPopup.DetectKeystroke(p: TACPopupItem);
var s : string;
	c : char;
    i : integer;
begin
    s := p.Prefix + p.Caption;
    s := StringReplace(s,'&&','',[rfReplaceAll]);
    i := Pos('&', s);
    if i <> 0 then begin
        c := s[i+1];
        c := upcase(c);
        p.fcaption := StringReplace(p.fcaption, '&'+c,'&'+c,[rfIgnoreCase]);
        fKeystrokes.AddObject(c, p);

        if p.ItemType = IT_PERMANENT then begin
            p.Prefix := '&' + c;
            if Pos('&'+c+' ', p.fCaption) = 1 then begin
                p.fCaption := UnitToken.clipManyChars(p.fCaption, 3);
            end;
        end;
        //fAccelCharsUsed := fAccelCharsUsed + c;
    end;
end;
function TACPopup.GetAcceleratorPopupClips(i: integer) : Char;
begin
    if FrmConfig.cbPopupCharLimit.Checked then begin
        Result := self.GetAccelerator(
            i - (fAcceleratorCount+fReservedCount)
        );
    end else begin
        if FrmConfig.cbPinnedShortcuts.checked then begin
            Result := self.GetAccelerator(
                i - (ClipQueue.GetQueueSize+PinnedClipQueue.GetQueueCount+fReservedCount)
            );
        end else begin
            Result := self.GetAccelerator(
                i - (ClipQueue.GetQueueSize+fReservedCount)
            );
        end;
    end;
end;
function TACPopup.GetAccelerator(i: integer): char;
{ const
 	other : array[37..46] of char = (
        '-', '=', '[', ']', '\', ';', '''', ',', '.', '/' );}
var other : string;
	alphanumeric : string;
begin
    result := ' ';
    inc(i, fAcceleratorCount+fReservedCount);
    if (i<=35) then begin
        alphanumeric := frmconfig.GetAccelAlphaNumeric;
    	result := alphanumeric[i+1];
    end else begin
    	other := FrmConfig.GetAccelSymbols;
        if (i-35) < length(other) then begin {exclude the last character}
        	result := other[i-35]; {start with 1 as an index}
        end;
    end;
end;
procedure TACPopup.CheckForAutoExpand(p : TACPopupItem);
begin
    if p.ItemType = fAutoExpandOnce then begin
        p.Expanded := true;
    end;
end;

procedure TACPopup.SaveExpandedState;
var i : integer;
begin
    fLastExpandedSet := [];

    for i := 0 to self.fItems.fSubMenu.count - 1 do begin
        with fItems.fSubMenu.Items[i] as TACPopupItem do begin
            if IsExpandable and Expanded then begin
                fLastExpandedSet := fLastExpandedSet + [ItemType];
            end;
        end;
    end;
end;
procedure TACPopup.LoadExpandedState;
var i : Integer;
begin
    for i := (self.fItems.fSubMenu.count - 1) downto 0 do begin
        with fItems.fSubMenu.Items[i] as TACPopupItem do begin
            if (IsExpandable) and (ItemType in fLastExpandedSet) then begin
                Expanded := true;
            end;
        end;
    end;
end;


function TACPopup.AddBreak: TACPopupItem;
begin
    result := fItems.AddBreak;
end;
function TACPopup.AddLine: TACPopupItem;
begin
    result := fitems.AddLine;
end;
procedure TACPopup.AddCancel;
var p : TACPopupItem;
begin
    p := self.Add;
    p.ItemType := IT_CANCEL;
    if FrmMainPopup.GetNeedsFocus then begin
        p.Caption := 'Cancel   (ESC)';
    end else begin
        p.Caption := 'Cancel';
    end;
    p.fIntegerData := Integer(MOT_CANCEL);
end;
procedure TACPopup.AddPastingTools;
var p, sub : TACPopupItem;
begin
    p := self.Add;
    p.IsExpandable := true;
    p.fLeftIconOnlyOnHover := true;

    p.Caption := 'Pasting Tools';
    p.LeftBitmap := ImgToBitmap(frmmainpopup.ImgMisc);
    SetPrefix(p);
    p.fIntegerData := Integer(MOT_PASTEUTILS);
    p.ItemType := IT_PASTING_TOOLS;
    {sub := p.Add;
    sub.Caption := 'Edit History...';
    sub.LeftIcon := GetIconFromBitmap(frmmainpopup.imgEdit.Picture.Bitmap);
    sub.fIntegerData := Integer(owEditHistory);
    sub.OnClick := ACPopupClicks.OpenWindowCommand;

    sub := p.addline;}

    sub := p.add;
    sub.caption := 'Paste All';
    sub.LeftBitmap := ImgToBitmap(frmmainpopup.imgPasteAll);
    sub.Hint := 'Pastes all Popup Clips as a single clip';
    sub.fLeftIconOnlyOnHover := true;
    sub.OnClick  := ACPopupClicks.PasteAllClick;

    sub := p.add;
    sub.caption := 'Paste Selected Text...';
    sub.LeftBitmap := ImgToBitmap(frmmainpopup.imgPaste);
    sub.fIntegerData := Integer(owPasteSelected);
    sub.OnClick := ACPopupClicks.OpenWindowCommand;
    sub.Hint := 'Shows the ''Paste Selected'' Window';
    sub.fLeftIconOnlyOnHover := true;

    p.addLine;

    sub := p.add;
    //frmmainpopup.UseFormMode;
    sub.Caption := 'Form Mode';
    //sub.LeftIcon := GetIconFromBitmap(frmmainpopup.imgEdit.Picture.Bitmap);
    sub.Checked := self.FormMode;
    sub.hint := 'Enabled pasting of multiple clips';
	sub.ItemType := IT_FORMMODE;

    sub := p.add;
    sub.caption := '>> Press TAB';
    sub.Hint := 'Simulates a Tab keypress';
	sub.ItemType := IT_NEXTFIELD;

    if self.FormMode then begin
        p.Expanded := true;
    end;

    self.CheckForAutoExpand(p);
    //sub.LeftIcon := GetIconFromBitmap(frmmainpopup.imgEdit.Picture.Bitmap);
end;
procedure TACPopup.AddAllItemsSubmenu;
var i, j : integer;
    m, main, sub, subm2 : TACPopupItem;

    s : string;
    wc : TWideChar;
    ci : TClipItem;
begin
    main := nil;

    if (not FrmConfig.cbShowAllAsSubmenus.checked) then begin
        main := self.Add;
        main.Caption := 'All Permanent';
        main.leftbitmap := ImgToBitmap(frmmainpopup.imgPermCasc);
        main.IsExpandable := true;
        main.fIntegerData := Integer(MOT_ALLPERMANENT);
        main.ItemType := IT_ALLPERM_MENU;
        SetPrefix(main);
    end;
    s := lowercase(frmPermanent.GetPermanentPath);

    FrmPermanent.PermFolderPush;
    unitMisc.AppendLog('PermFolders count='+IntToStr(frmPermanent.PermFoldersGetCount));
    for i := 0 to frmPermanent.PermFoldersGetCount - 1 do begin
        //if (lowercase(frmPermanent.PermFoldersGetItem(i)) <> s) {or (not self.permshown)} then begin

            if (FrmConfig.cbShowAllAsSubmenus.checked) then begin
                m := self.add;
                m.LeftBitmap := ImgToBitmap(frmmainpopup.imgPermCasc);
                SetPrefix(m);
            end else begin
                m := main.Add;
            end;
            m.Caption := frmPermanent.PermFoldersGetItem(i);
            m.itemtype := IT_POPUP;
            AddSubmenu(m);
            FrmPermanent.SetPermanentPath(m.Caption);

            unitmisc.AppendLog('PermCount='+IntToSTr(FrmPermanent.GetCount));
            for j := 0 to FrmPermanent.GetCount - 1 do begin
                subm2 := m.add;
                subm2.Caption := FrmPermanent.GetItemName(j);
                subm2.itemtype := IT_PERMANENT;
                subm2.fPermanentGroupID := i;
                subm2.fPermanentID  := j;
                subm2.fPermanentText := FrmPermanent.GetItemText(j);

                subm2.leftbitmap := ImgToBitmap(FrmMainPopup.ImgEdit);
                subm2.fLeftIconOnlyOnHover := true;

                if FrmPermanent.IsComplexItem(subm2.fPermanentText) then begin
                    ci := FrmPermanent.GetComplexItem(subm2.fPermanentText);
//                    subm2.fHasRightIcon := true;
                    case ci.GetFormatType of
                    FT_FILE: 		subm2.RightIcon := ImgToBitmap(FrmMainPopup.imgFile);
                    FT_PICTURE: 	subm2.RightIcon := ImgToBitmap(FrmMainPopup.imgPic);
                    FT_HTML:		subm2.RightIcon := ImgToBitmap(FrmMainPopup.imgHTML);
                    FT_RICHTEXT:	subm2.RightIcon := ImgToBitmap(FrmMainPopup.imgRichText);
                    FT_UNICODE: 	subm2.fHasRightIcon := false;
                    end;

                    myfree(ci);
                end;

                subm2.OnClick := ACPopupClicks.PermanentClipClick;
            end;
        //end;
    end;

    if (not FrmConfig.cbShowAllAsSubmenus.checked) then begin
        if FrmConfig.cbPermanentItemsSubmenu.Checked then begin
            subm2 := main.Add;
            self.AddSubmenu(subm2);
            subm2.Caption := '[Switch Group]';
        end;
        for i := 0 to frmPermanent.PermFoldersGetCount - 1 do begin
            if (not FrmConfig.cbPermanentItemsSubmenu.Checked) and (i=0) then begin
                main.AddLine;
            end;
            if FrmConfig.cbPermanentItemsSubmenu.Checked then begin
                sub := subm2.add;
            end else begin
                sub := main.add;
            end;
            //sub.Caption := FrmPermanent.PermFoldersGetItem(i);
            sub.fPermanentGroupID := i;
            sub.itemtype := IT_SWITCH;
            sub.fCollapseParentOnClick := true;
            sub.OnClick := ACPopupClicks.PermanentGroupClick;
            wc := TWideChar.create;
            wc.Append(WideChar($2192)); {right arrow}
            if (lowercase(FrmPermanent.PermFoldersGetItem(i)) <> s) then begin
                sub.Checked := false;
                sub.CheckGrouped;
            end else begin
                sub.checked := true;
                sub.CheckGrouped;
            end;
            wc.AppendUnicode('[' +FrmPermanent.PermFoldersGetItem(i) + ']');
            sub.fcaption := wc.GetString;
            sub.hint := 'Changes the current Permanent Clips group';
        end;

	end;
    FrmPermanent.PermFolderPop;

    if (not FrmConfig.cbShowAllAsSubmenus.checked) then begin
        if (main.fSubMenu.Count>0) then begin

            if (FrmConfig.cbShowAllAsSubmenus.checked) then begin
                self.AddLine;
                m := self.add;
            end else begin
                main.addline;
                m := main.Add;
            end;
            m.Caption := 'Edit ...';
            m.LeftBitmap := ImgToBitmap(FrmMainPopup.imgPermCasc);
            m.fIntegerData := Integer(owPermanentClips);
            m.OnClick := ACPopupClicks.OpenWindowCommand;
            m.fLeftIconOnlyOnHover := true;

            self.CheckForAutoExpand(main);
        end;
    end;
end;


procedure TACPopup.AddSwitchToSubmenu;
var i : integer;
    p, sub : TACPopupItem;
    s : string;
    wc : TWideChar;
begin
    p := self.add;
    p.Caption := 'Permanent Groups';
    p.LeftBitmap := ImgToBitmap(FrmMainPopup.imgPermCasc);
    p.IsExpandable := true;
    p.fIntegerData := Integer(MOT_SWITCH);
    p.ItemType := IT_SWITCH_MENU;
    SetPrefix(p);
    s := lowercase(frmPermanent.GetPermanentPath);


    for i := 0 to frmPermanent.PermFoldersGetCount - 1 do begin
        sub := p.add;
        //sub.Caption := FrmPermanent.PermFoldersGetItem(i);
        sub.fPermanentGroupID := i;
        sub.itemtype := IT_SWITCH;
        sub.fCollapseParentOnClick := true;
        sub.OnClick := ACPopupClicks.PermanentGroupClick;
        wc := TWideChar.create;
        wc.Append(WideChar($2192)); {right arrow}
        if (lowercase(FrmPermanent.PermFoldersGetItem(i)) <> s) then begin

            sub.Checked := false;
            sub.CheckGrouped;
        end else begin
            sub.checked := true;
            sub.CheckGrouped;
        end;
        wc.Append('[' +FrmPermanent.PermFoldersGetItem(i) + ']');
        sub.fcaption := wc.GetString;
        sub.hint := 'Changes the current Permanent Clips group';
        //SetPrefix(sub);
    end;
    {
    for i := 0 to frmPermanent.PermFoldersGetCount - 1 do begin
        sub := p.add;
        sub.Caption := FrmPermanent.PermFoldersGetItem(i);
        sub.fPermanentGroupID := i;
        sub.itemtype := IT_SWITCH;
        sub.fCollapseParentOnClick := true;
        sub.OnClick := ACPopupClicks.PermanentGroupClick;
        if (lowercase(sub.caption) <> s) then begin
            sub.Caption := 'switch to [' + sub.caption + ']';
        end else begin
            sub.Caption := 'switch to [' + sub.caption + '] (Current)';
        end;
        sub.hint := 'Changes the current Permanent Clips group';
    end;
    }

    p.addline;
    sub := p.add;
    sub.Caption := 'Edit ...';
    sub.fIntegerData := Integer(owPermanentClips);
    sub.OnClick := ACPopupClicks.OpenWindowCommand; 
    sub.LeftBitmap :=  ImgToBitmap(FrmMainPopup.imgPermCasc);
    sub.fLeftIconOnlyOnHover := true;


    self.CheckForAutoExpand(p);
end;
procedure TACPopup.AddCurrentClipboard;
var
    p : TACPopupItem;
begin
    CurrentClipboard.GetClipboardItem(0);
    if CurrentClipboard.GetDataSize <> 0 then begin
        p := self.Add;
        p.ItemType := IT_CLIPBOARD;

        p.Clip := CurrentClipboard;
        self.SetPrefix(p);
        p.fIntegerData := Integer(MOT_CURRENT);

        p.CaptionFromClip := p.clip;


//        wc.Append('[Clipboard]: ');

{
        wc := TWideChar.create;
        if p.fWideCaption = nil then begin
            wc.appendUnicode(p.fCaption);
        end else begin
            wc.Append(p.fWideCaption);
        end;
        myfree(p.fWideCaption);
        p.fWideCaption := wc;
 }

        p.OnClick := ACPopupClicks.PopupClipClick;
        p.OnRightClick := ACPopupClicks.PopupClipRightClip;

        SetLeftIcon(p, p.clip, true);
    end else begin
        if self.fIncludeEmpty or FrmConfig.cbShowEmptyMenus.checked then begin
            p := self.Add;
            p.ItemType := IT_CLIPBOARD;
            self.SetPrefix(p);
            p.fIntegerData := Integer(MOT_CURRENT);
        	p.fcaption := '[empty]';
            p.LeftBitmap := FrmClipboardManager.iText.Picture.Bitmap;
            p.Disabled := true;
        end else begin
        	self.ReserveAccel;

        end;
    end;

end;
procedure TACPopup.AddCurrentPermanentItems;
var i : integer;
    p : TACPopupItem;
    ci : TClipItem;
begin
	for i := 0 to frmPermanent.GetCount - 1 do begin
    	p := self.add;
        p.fCaption := FrmPermanent.GetItemName(i); {bypass the clip cleaning routings}

		p.LeftBitmap  := ImgToBitmap(FrmMainPopup.ImgEdit);
        p.fLeftIconOnlyOnHover := true;
        p.itemtype := IT_PERMANENT;
        p.fPermanentGroupID := -1;
        p.fPermanentID := i;
        p.fPermanentText := FrmPermanent.GetItemText(i);
        p.OnClick := ACPopupClicks.PermanentClipClick;
        p.OnRightClick := ACPopupClicks.PopupClipRightClip;

        if FrmPermanent.IsComplexItem(p.fPermanentText) then begin
            ci := FrmPermanent.GetComplexItem(p.fPermanentText);
            case ci.GetFormatType of
            FT_FILE: 		p.RightIcon := ImgToBitmap(FrmMainPopup.imgFile);
            FT_PICTURE: 	p.RightIcon := ImgToBitmap(FrmMainPopup.imgPic);
            FT_HTML:		p.RightIcon := ImgToBitmap(FrmMainPopup.imgHTML);
            FT_RICHTEXT:	p.RightIcon := ImgToBitmap(FrmMainPopup.imgRichText);
            FT_UNICODE: 	p.fHasRightIcon := false;
            end;

            myfree(ci);
        end;

        SetPrefix(p);
	end;
end;
procedure TACPopup.AddFullConfigured;
var p : TACPopupItem;
begin
    p := self.Add;
    p.ItemType := IT_FULL;
    if self.UseKeyboard then begin
    	p.Prefix := '&' +FrmConfig.GetAccelSymbols[length(frmconfig.GetAccelSymbols)]+  ' ';
    end;
    if fFullMode then begin
        p.Caption := 'Configured Mode';
        p.Hint := 'Show only configured items';
    end else begin
        p.Caption := 'Full Mode';
        p.Hint := 'Show all possible items';
    end;
    SetPrefix(p);


    p.LeftBitmap := ImgToBitmap(frmmainpopup.imgFull);
end;
procedure TACPopup.AddLast;

begin
{
    if PData.LastClickedText <> '' then begin
        p := self.add;
        p.Caption := 'Last: ' + PData.LastClickedText;
        self.AddPrefix(p);
    end else begin
        if fFullMode = false then
            inc(fReservedCount);
    end;
}
end;
procedure TACPopup.SetPrefix(p: TACPopupItem);
var accel : char;
	i, j  : integer;
    s : string;
begin
	if self.UseKeyboard = false then EXIT;

    s := stringreplace(p.Caption,'&&','',[rfReplaceAll]);
    if (Pos('&', s)=0) and (pos('&', p.Prefix)=0)  then begin

        j := 0;
        repeat
//        	if j <> 0 then appendlog('Accel key take: ' + accel);

        	accel := GetAccelerator(j);
            if accel = ' ' then begin
                AppendLog('out of accelerator keys');
                EXIT;
            end;
            
            i := fKeystrokes.IndexOf(accel);
            if (i=-1) then begin
                i := Pos(accel, fAccelCharsUsed);
                if i=0 then i:=-1;
            end;
            inc(j);
        until i = -1;
		inc(fAcceleratorCount);
        
        p.Prefix := '&' + accel  + ' ';
        fKeystrokes.AddObject(accel, p);
    end else begin
		if (p.ItemType in [IT_MENU, IT_PERMANENT, IT_FULL])  then begin
            DetectKeystroke(p);
        end;
    end;
end;
procedure TACPopup.AddProgramOptions;
var p, sub : TACPopupItem;
    i : integer;
begin
    p := self.Add;
    p.caption := 'Program options: [' + FrmMainPopup.TargetData.ExeName +']';
    p.IsExpandable := true;
    p.LeftIcon := FrmMainPopup.TargetData.icon;
    p.ItemType := IT_PROGRAM_MENU;
    p.fIntegerData := Integer(MOT_PROGRAM);
    SetPrefix(p);
    p.Hint := 'Per-program options for Pasting and Monitoring';
    i := Integer(Paste.GetPasteMethod(FrmMainPopup.TargetData.ExeName));


    sub := p.add;
    sub.Caption := 'CTRL+V';
    sub.ItemType := IT_PASTE_METHOD;
    sub.Checked := (i = 0);
    sub.fIntegerData :=0;
    sub.CheckGrouped;
    sub.OnClick := ACPopupClicks.PasteMethodClick;


    sub := p.add;
    sub.Caption := 'SHIFT+INSERT';
    sub.ItemType := IT_PASTE_METHOD;
    sub.Checked := (i = 1);
    sub.fIntegerData := 1;
    sub.CheckGrouped;
    sub.OnClick := ACPopupClicks.PasteMethodClick;

    sub := p.add;
    sub.Caption := 'Mimic Typing';
    sub.ItemType := IT_PASTE_METHOD;
    sub.Checked := (i = 2);
    sub.fIntegerData := 2;
    sub.CheckGrouped;
    sub.OnClick := ACPopupClicks.PasteMethodClick;

    sub := p.add;
    sub.Caption := 'Clipboard Only';
    sub.ItemType := IT_PASTE_METHOD;
    sub.Checked := (i = 3);
    sub.fIntegerData := 3;
    sub.CheckGrouped;
    sub.OnClick := ACPopupClicks.PasteMethodClick;

    sub := p.add;
    sub.Caption := 'Default';
    sub.ItemType := IT_PASTE_METHOD;
    sub.Checked := (i = 4);
    sub.fIntegerData := 4;
    sub.CheckGrouped;
    sub.OnClick := ACPopupClicks.PasteMethodClick;

    p.AddLine;

    sub := p.add;
    sub.caption := 'Disable Monitoring';
    sub.ItemType := IT_PASTE_METHOD;
    sub.Checked := FrmSysTrayMenu.IsDisabledEXE(FrmMainPopup.TargetData.ExeName);
    sub.OnClick := ACPopupClicks.ProgramOptionDisableMonitoring;

    sub := p.add;
    sub.caption := 'Disable Right-Click Trigger';
    sub.ItemType := IT_PASTE_METHOD;
    sub.Checked := FrmSysTrayMenu.IsNoRightClickEXE(FrmMainPopup.TargetData.ExeName);
    sub.OnClick := ACPopupClicks.ProgramOptionDisableRightClickPopup; 

    self.CheckForAutoExpand(p);
end;
procedure TACPopup.AddRecentlyClicked;
var p, sub : TACPopupItem;
    i : integer;
    ci : TClipItem;
begin
	for i := (ClickedclipQueue.GetQueueCount - 1) downto 0 do begin
        ci := ClickedclipQueue.GetClipItem(i);
        if ci.CData = nil then begin
            AppendLog('Screwed up Recently Clicked clip');
            ClickedClipQueue.DeleteItem(i);
            CONTINUE;
        end;
    end;

    if (ClickedclipQueue.GetQueueCount = 0) and (not fIncludeEmpty) and (not FrmConfig.cbShowEmptyMenus.checked ) then begin
        self.ReserveAccel;
    end else begin
        p := self.Add;
        p.Caption := 'Recently Clicked';
        p.LeftBitmap := ImgToBitmap(frmmainpopup.imgFlush);
        p.itemtype := IT_CLICKED_MENU;
        p.IsExpandable := true;
        p.fIntegerData := Integer(MOT_CLICKED);
        SetPrefix(p);
    	p.Disabled := (ClickedclipQueue.GetQueueCount = 0);

        p.hint := 'History of clips recently clicked on';


        for i := 0 to (ClickedclipQueue.GetQueueCount - 1) do begin
        	ci := ClickedclipQueue.GetClipItem(i);
            if ci.CData = nil then begin
            	AppendLog('Screwed up Recently Clicked clip');
            	CONTINUE;
            end;

            sub := p.add;
            sub.fClip := ci;
            sub.CaptionFromClip := sub.clip;
            sub.itemtype := IT_CLICKED;
            SetLeftIcon(sub, sub.fclip);
            if i = (ClickedclipQueue.GetQueueCount - 1) then begin
                p.AddLine;
            end;
            sub.OnClick := ACPopupClicks.PopupClipClick;
            p.OnRightClick := ACPopupClicks.PopupClipRightClip; 
        end;

        if p.fSubMenu.Count = 0 then begin
            sub := p.add;
            sub.Caption := 'none';
        end;

        self.CheckForAutoExpand(p);
    end;
end;
procedure TACPopup.AddRecentlyRemoved;
var p, p2 : TACPopupItem;
    i : integer;


begin
    p := self.Add;
    p.Caption := 'Recently Removed Clips';
    //AddSubmenu(p);
    p.IsExpandable := true;
    SetPrefix(p);
    p.LeftBitmap := ImgToBitmap(FrmMainPopup.imgRemoved);
    p.fIntegerData := Integer(MOT_REMOVED);
    p.ItemType := IT_RECENTLY_REMOVED;
    p.Disabled := (TempClipQueue.GetQueueCount=0);


    for i := 0 to (TempClipQueue.GetQueueCount - 1) do begin
        p2 := p.Add;

        p2.fClip := TempClipQueue.GetClipItem(i);
        //p2.Caption := TempClipQueue.GetItemText(i);
        p2.CaptionFromClip := p2.clip;
        p2.ItemType := IT_TEMP;
        SetLeftIcon(p2, p2.fClip);
        p2.OnClick := ACPopupClicks.RecentlyRemovedClicke;
        p.OnRightClick := ACPopupClicks.PopupClipRightClip; 
        if i = (TempClipQueue.GetQueueCount - 1) then begin
            p2 := p.add;
            p2.IsLine := true;
        end;
    end;

    p2 := p.Add;

    p2.Caption := 'Removed Clips ...';
    p2.LeftBitmap := ImgToBitmap(FrmMainPopup.imgRemoved);
    p2.fIntegerData := Integer(owRemovedClips);
    p2.OnClick := ACPopupClicks.OpenWindowCommand;
    p2.fLeftIconOnlyOnHover := true;

    self.CheckForAutoExpand(p);
end;
procedure TACPopup.AddSearch;
var p : TACPopupItem;
begin
    p := self.Add;
    p.Caption := 'Search and Paste ...';
    p.LeftBitmap := ImgToBitmap(FrmMainPopup.imgSearch);
    p.ItemType := IT_SEARCH;
    p.fIntegerData := Integer(owSearch);
    p.OnClick := ACPopupClicks.OpenWindowCommand;
    p.hint := 'Shows the ''Search'' Window';
    SetPrefix(p);
end;
procedure TACPopup.AddSubmenu(p: TACPopupItem);
begin
    p.IsSubMenu := true;
    p.Disabled := true;
    p.ItemType := IT_POPUP;
    p.fLeftIconOnlyOnHover := true;
end;
procedure TACPopup.AddSwitchToItems;
var i : integer;
    sub : TACPopupItem;
    s : string;
    wc : TWideChar;
begin
    s := lowercase(frmPermanent.GetPermanentPath);
    for i := 0 to frmPermanent.PermFoldersGetCount - 1 do begin
    	if (lowercase(frmPermanent.PermFoldersGetItem(i)) = s) then CONTINUE;
        
        sub := self.add;
        sub.fPermanentGroupID := i;
        sub.itemtype := IT_SWITCH;
        sub.OnClick := ACPopupClicks.PermanentGroupClick;
        wc := TWideChar.create;
        if (lowercase(FrmPermanent.PermFoldersGetItem(i)) <> s) then begin
            wc.Append(WideChar($2192)); {right arrow}
        end;
        wc.AppendUnicode('[' +FrmPermanent.PermFoldersGetItem(i) + ']');
        sub.fcaption := wc.GetString;
        sub.hint := 'Changes the current Permanent Clips group';
        sub.LeftBitmap := ImgToBitmap(frmmainpopup.imgPermCasc);
        SetPrefix(sub);
    end;
end;
procedure TACPopup.AddSystem;
var p, sub, sub2 : TACPopupItem;
begin
    p := self.add;
    p.Caption := 'System';
    p.LeftBitmap := ImgToBitmap(frmmainpopup.imgAICon);
    p.IsExpandable := true;
    p.fLeftIconOnlyOnHover := true;
    SetPrefix(p);
    p.fIntegerData := Integer(MOT_SYSTEM);
    p.ItemType := IT_SYSTEM_MENU;
    
    sub := p.Add;
    sub.Caption := 'Configuration ...';
    sub.LeftBitmap := ImgToBitmap(frmmainpopup.imgA);
    sub.ItemType := IT_SYSTEM;
    sub.fIntegerData := Integer(owConfiguration);
    sub.OnClick := ACPopupClicks.OpenWindowCommand;
    sub.hint := 'Shows the ''Configuration'' Window';
    sub.fLeftIconOnlyOnHover := true;

    p.addline;

    sub := p.Add;
    sub.Caption := 'Permanent Clips/Macros ...';
    sub.LeftBitmap := ImgToBitmap(frmmainpopup.imgPermCasc);
    sub.ItemType := IT_SYSTEM;
    sub.fIntegerData := Integer(owPermanentClips);
    sub.OnClick := ACPopupClicks.OpenWindowCommand;
    sub.hint := 'Shows the ''Permanent Clips'' Window';
    sub.fLeftIconOnlyOnHover := true;

    sub := p.Add;
    sub.Caption := 'Edit History ...';
    sub.LeftBitmap := ImgToBitmap(frmmainpopup.imgEdit);
    sub.ItemType := IT_SYSTEM;
    sub.fIntegerData := Integer(owEditHistory);
    sub.OnClick := ACPopupClicks.OpenWindowCommand;
    sub.hint := 'Shows the ''Edit History'' Window';
    sub.fLeftIconOnlyOnHover := true;

    sub := p.Add;
    sub.Caption := 'Edit Clipboard ...';
    sub.LeftBitmap := ImgToBitmap(frmmainpopup.imgPaste);
    sub.ItemType := IT_SYSTEM;
    sub.fIntegerData := Integer(owEditClipboard);
    sub.OnClick := ACPopupClicks.OpenWindowCommand;
    sub.fClip := CurrentClipboard;
    sub.hint := 'Edits the clipboard using an external program';
    sub.fLeftIconOnlyOnHover := true;

    sub := p.Add;
    sub.caption := 'Flush History';
    sub.LeftBitmap := ImgToBitmap(FrmMainPopup.imgFlush);
    sub.ItemType := IT_POPUP;
    sub.fLeftIconOnlyOnHover := true;
    AddSubmenu(sub);
    sub.Hint := 'Erase Clipboad History';

    sub2 := sub.add;
    sub2.Caption := 'Erase Clipboard';
    sub2.fIntegerData := Integer(peClipboard);
    sub2.OnClick := ACPopupClicks.EraseHistory;
    sub2 := sub.add;
    sub2.Caption := 'Erase Popup Clips';
    sub2.fIntegerData := Integer(pePopupClips);
    sub2.OnClick := ACPopupClicks.EraseHistory;
    sub2 := sub.add;
    sub2.Caption := 'Erase Removed Clips';
    sub2.fIntegerData := Integer(peRemovedClips);
    sub2.OnClick := ACPopupClicks.EraseHistory;
    sub.AddLine;
    sub2 := sub.add;
    sub2.Caption := 'Erase Everything';
    sub2.fIntegerData := Integer(peEverything);
    sub2.OnClick := ACPopupClicks.EraseHistory;    


    sub := p.Add;
    sub.Caption := 'Toggle Clipboard Bar';
    sub.LeftBitmap := ImgToBitmap(frmmainpopup.imgPaste);
    sub.Checked := frmClipboardBar.IsVisible;
    sub.OnClick := ACPopupClicks.OpenWindowCommand;
    sub.hint := 'Shows or hides the Clipboard Bar';
    sub.fLeftIconOnlyOnHover := true;
    sub.onclick := ACPopupClicks.ToggleClipboardBar;

    p.addline;

    sub := p.Add;
    sub.Caption := 'About ...';
    sub.ItemType := IT_SYSTEM;
    sub.fIntegerData := Integer(owAbout);
    sub.OnClick := ACPopupClicks.OpenWindowCommand;
    sub.hint := 'Shows the ''About'' Window';

    sub := p.Add;
    sub.Caption := 'Exit';
    sub.ItemType := IT_SYSTEM;
    sub.OnClick := frmSysTrayMenu.ExitMenuItemClickEvent;
    sub.hint := 'Closes the program';

    self.CheckForAutoExpand(p);
end;
procedure TACPopup.AddPopupClips;
var i, groupcnt : integer;
    p : TACPopupItem;
    group : TACPopupItem;

    procedure SetPopupPrexif(p : TACPopupItem);
    begin
        if FrmConfig.cbPinnedShortcuts.Checked then begin
            p.prefix :=  '&' + GetAcceleratorPopupClips(i+PinnedClipQueue.GetQueueCount) +' ';
        end else begin
            p.prefix :=  '&' + GetAcceleratorPopupClips(i) +' ';
        end;
    end;
begin
    CurrentClipboard.GetClipboardItem(0);

    if PinnedClipQueue.GetQueueCount > 0 then begin
        for i := (PinnedClipQueue.GetQueueCount - 1) downto 0 do begin
            p := self.add;
            p.Clip := PinnedClipQueue.GetClipItem(i);
            p.CaptionFromClip := p.clip;
            p.Prefix := '  ';
            if self.UseKeyboard and frmconfig.cbPinnedShortcuts.checked then begin
                p.prefix := '&' + GetAcceleratorPopupClips((PinnedClipQueue.GetQueueCount-1)-i) +' ';
            end;
            p.ItemType := IT_PINNED;
            SetLeftIcon(p, p.Clip);
            p.RightIcon := ImgToBitmap(FrmMainPopup.imgUnkown);
            p.OnClick := ACPopupClicks.PopupClipClick;
            p.OnRightClick := ACPopupClicks.PopupClipRightClip;
            p.fIntegerData := i;


            case p.fclip.GetFormatType of
            FT_FILE: 		p.RightIcon := ImgToBitmap(FrmMainPopup.imgFile);
            FT_PICTURE: 	p.RightIcon := ImgToBitmap(FrmMainPopup.imgPic);
            FT_HTML:		p.RightIcon := ImgToBitmap(FrmMainPopup.imgHTML);
            FT_RICHTEXT:	p.RightIcon := ImgToBitmap(FrmMainPopup.imgRichText);
            FT_UNICODE: 	p.fHasRightIcon := false;
            end;

            if p.fRightBitmap <> nil then
                p.fRightBitmap.TransparentColor := $AAAAAA;

            DetectKeystroke(p);
        end;
        Self.AddLine;
    end;


    //
    // use group when enabled and history > set size
    // use group ONCE when popup too tall and a large enough history
    //
    self.fUseGrouping := (ClipQueue.GetQueueCount > FrmConfig.udClipsPerSet.Position)
        and FrmConfig.cbEnableClipSets.Checked;
    if self.fUseGroupingOnce then begin
        Self.fUseGroupingOnce := false;
        self.fUseGrouping := (ClipQueue.GetQueueCount > FrmConfig.udClipsPerSet.Position);
        self.fUseGroupingOnceFailed := not self.fUseGrouping;
    end;

    self.GroupingCount := FrmConfig.udClipsPerSet.Position;

    self.fGroup.Clear;
    groupcnt := 0;
    // extract Clip, Caption, Prefix, and 2 possible icons
    for i := 0 to ClipQueue.GetQueueCount - 1 do begin

        if self.fUseGrouping then begin
            if (i mod self.fGroupingCount = 0) then begin
                group := self.Add;
                group.IsExpandable := true;
                group.ItemType := IT_CLIPSET_HEADER;
                group.Caption := '[ Clips '+IntToStr(i+1)+'-'+IntToStr(i+self.fGroupingCount) +' ]';
                group.Expanded := false;
                group.ShowCollapseIcon := false;

                SetPrefix(group);
                Dec(Self.fAcceleratorCount);   // kludge to keep the shortcut key for the first submenu right

                fGroup.Add(group);

                Inc(groupcnt);
            end;
            p := group.Add;
        end else begin
            p := self.add;
        end;
        p.Clip := ClipQueue.GetClipItem(i);
        p.CaptionFromClip := p.clip;
        if self.UseKeyboard then begin
            if (FrmConfig.cbPopupCharLimit.Checked) then begin
                if i < frmconfig.udPopupKeysLimit.Position then begin
                    SetPopupPrexif(p);
                end;
            end else begin
        	    SetPopupPrexif(p);
            end;
        end;
        p.ItemType := IT_MENU;
        SetLeftIcon(p, p.Clip);
        p.RightIcon := ImgToBitmap(FrmMainPopup.imgUnkown);
        p.OnClick := ACPopupClicks.PopupClipClick;
        p.OnRightClick := ACPopupClicks.PopupClipRightClip;
        p.fIntegerData := i;


        case p.fclip.GetFormatType of
        FT_FILE: 		p.RightIcon := ImgToBitmap(FrmMainPopup.imgFile);
        FT_PICTURE: 	p.RightIcon := ImgToBitmap(FrmMainPopup.imgPic);
        FT_HTML:		p.RightIcon := ImgToBitmap(FrmMainPopup.imgHTML);
        FT_RICHTEXT:	p.RightIcon := ImgToBitmap(FrmMainPopup.imgRichText);
        FT_UNICODE: 	p.fHasRightIcon := false;
        end;

        if p.fRightBitmap <> nil then
        	p.fRightBitmap.TransparentColor := $AAAAAA;

        DetectKeystroke(p);
    end;

    try
        if self.UseGrouping then begin
            TACPopupItem(fGroup.Items[self.fCurrentGroup]).Expanded := true;
        end;
    except
        self.fCurrentGroup := 0;
    end;
end;

procedure TACPopup.AutoPopulate;
var i, j,x, cnt : integer;
	c : char;
    linebefore : boolean;
    mot : TMenuOrderType;
    s : string;
    b : boolean;
begin
    AppendLog('Clear Popup');

    if (FrmConfig.lvOrder.items.Count=0) then begin
        FrmConfig.lvOrder.HandleNeeded;
        if (FrmConfig.lvOrder.items.Count=0) then begin
            raise Exception.Create('Error: [Configuration>Popup>Menu Order] is empty.');
            EXIT;
        end;
    end;

    self.Clear;
    fitems.fSubShowing := nil;

    // default colors
        fBackgroundColor := clBtnFace;
        fHighlightColor := ColorToRGB(clHighlight);
        fFontColor := ColorToRGB(clBtnText);
        fClickedColor := $00790079;


        if FrmConfig.cbFontColor.Checked then begin
            fFontColor := ColorToRGB(FrmConfig.pnlFontColor.color);
        end;
        if FrmConfig.cbHighlightColor.Checked then begin
            fHighlightColor := ColorToRGB(FrmConfig.pnlHighlightColor.color);
        end;
        if FrmConfig.cbBackgroundColor.Checked then begin
            fBackgroundColor := ColorToRGB(FrmConfig.pnlBackgroundColor.color);
        end;
        if frmconfig.cbClickedColor.Checked then begin
            fClickedColor := ColorToRGB(FrmConfig.pnlClickedColor.color);
        end;
        fExpandedBackgroundColor := DimColorOrReverse(fBackgroundColor, 1.04);
        fColumnColor := DimColor(fBackgroundColor,0.95);
        fColumnEdgeColor := dimColor(fBackgroundColor,0.86);
        fDisabledColor := DimColorOrReverse(fFontColor, 0.90);

    timHover.Enabled := false;
    fKeystrokes.Clear;
    fExpandedKeystrokes.Clear;

    fHover := nil; {Don't want to trigger timer}
    fReservedCount := 0;

    if FrmConfig.cbPopupCharLimit.Checked then begin
        if frmConfig.cbPinnedShortcuts.Checked then begin
            fAcceleratorCount := Min(ClipQueue.GetQueueSize, FrmConfig.udPopupKeysLimit.position) +
                 PinnedClipQueue.GetQueueCount;
        end else begin
            fAcceleratorCount := Min(ClipQueue.GetQueueSize, FrmConfig.udPopupKeysLimit.position);
        end;
    end else begin
        if frmConfig.cbPinnedShortcuts.Checked then begin
            fAcceleratorCount := ClipQueue.GetQueueSize + PinnedClipQueue.GetQueueCount;
        end else begin
            fAcceleratorCount := ClipQueue.GetQueueSize;
        end;
    end;
    fAccelCharsUsed := '';

    // reserve user set permanent keys
	for i := 0 to FrmConfig.lvOrder.items.Count - 1 do begin
    	mot := TMenuOrderType(frmConfig.lvOrder.Items[i].data);
        if frmConfig.lvOrder.Items[i].Checked or
        	fFullMode or
            ((mot = MOT_PASTEUTILS) and self.FormMode)  then begin
        	case mot of
                MOT_PERMANENT: begin
                	for j := 0 to FrmPermanent.GetCount - 1 do begin
                    	s := FrmPermanent.GetItemName(j);
                        s := StringReplace(s,'&&','',[rfReplaceAll]);
                        x := pos('&', s);
                        if (x<>0) then begin
                        	if x<length(s) then begin
                                c := upcase(s[x+1]);
                                fAccelCharsUsed := fAccelCharsUsed + c;
                            end;
						end;
                    end;
                end;
            end;
    	end;
    end;
    AppendLog('PermKeys= '+fAccelCharsUsed);
    AppendLog('FrmConfig.lvOrder.items.Count='+IntToStr(FrmConfig.lvOrder.items.Count));

    for i := 0 to FrmConfig.lvOrder.items.Count - 1 do begin
        linebefore := (FrmConfig.lvOrder.items[i].SubItems[0] = 'Yes');

        mot := TMenuOrderType(frmConfig.lvOrder.Items[i].data);
        b := fFullMode;
        if MOT = MOT_CANCEL then begin
            // always show CANCEL when forced keyboard off detected
            if (not FrmMainPopup.GetNeedsFocus) then begin
                fFullMode := true;
            end;
        end;

        if frmConfig.lvOrder.Items[i].Checked or
        	fFullMode or
            ((mot = MOT_PASTEUTILS) and self.FormMode)  then begin
            if linebefore and (fitems.fSubMenu.Count > 0) then begin
               self.AddLine;
            end;
            cnt := fItems.fSubMenu.Count;
            case mot of
                MOT_MENU : self.AddPopupClips; {
                MOT_OTHER : if frmConfig.cbSeparateNontextList.checked then self.AddOtherItems;  }
                MOT_CANCEL : self.AddCancel;
                {MOT_LAST : self.AddLast;}
                MOT_CURRENT : self.AddCurrentClipboard;
                {MOT_SELECTED : self.AddSelected;}
                MOT_REMOVED : self.AddRecentlyRemoved;
                MOT_PROGRAM : self.AddProgramOptions;
                MOT_SYSTEM : self.AddSystem;
                MOT_PASTEUTILS : self.AddPastingTools;
                MOT_PERMANENT : begin
                    // don't show in full mode, unless configured to show normally
                    if not (fFullMode and (not frmConfig.lvOrder.Items[i].Checked)) then begin
                        self.AddCurrentPermanentItems;
                    end;
                end;
                MOT_SWITCH :
                    if frmconfig.cbPermanentItemsSubmenu.checked then
                        self.AddSwitchToSubmenu
                    else
                        self.AddSwitchToItems;
                MOT_FULL : self.AddFullConfigured;
                MOT_ALLPERMANENT : self.AddAllItemsSubmenu;
//              This is being depricated
//                MOT_CLICKED: self.AddRecentlyClicked;
                MOT_SEARCH: self.AddSearch;
                MOT_BREAK:  if not (ffullMode and (not frmConfig.lvOrder.Items[i].Checked)) then begin
                     self.AddBreak; // don't add breaks in Full Mode
                end;

            end;
            // remove lines for items that generate no new menuitems
            if linebefore and (cnt = fItems.fSubMenu.count) then begin
                if cnt <> 0 then begin
                    fItems.fsubmenu.Delete(fitems.fsubmenu.count-1);
                end;
            end;
        end;
        fFullMode := b;
    end;

    fAutoExpandOnce := IT_NONE;

    AppendLog('End of AutoPopulate = '+ IntToStr(fitems.fsubmenu.count));
    fIncludeEmpty := false;
end;
procedure TACPopup.Clear;
begin
    fitems.fsubmenu.Clear;
end;



{utility routines}
function TACPopup.IsLegalDrop(p: TACPopupItem): boolean;
begin
	result := false;

	if p = nil then EXIT;
    if fDragMenuItem = nil then EXIT;
	if p = fDragMenuItem then EXIT;

    case fdragmenuitem.itemtype of
        IT_MENU, IT_PERMANENT,IT_CLICKED, IT_TEMP: begin
            if fDragMenuItem.ItemType = p.ItemType then begin
                result := true;
            end;
        end;
        else begin
            if fDragMenuItem.IsExpandable or
                (fdragmenuitem.ItemType in [IT_CLIPBOARD, {IT_LAST,} IT_SEARCH, IT_FULL,IT_MENU,IT_PERMANENT]) then begin
                if p.IsExpandable or
                (p.ItemType in [IT_CLIPBOARD, {IT_LAST,} IT_SEARCH, IT_FULL,IT_MENU, IT_PERMANENT]) then begin
                    result := true;
                end;
            end;
        end;
    end;
end;
function TACPopup.GetItemAt(x, y: integer): TACPopupItem;
var i : integer;
    p : TACPopupItem;

    function CheckSubmenus(p : TACPopupItem) : TACPopupItem;
    var i : integer;
        sub : TACPopupItem;
    begin
        result := nil;
        if p = nil then EXIT;

        if (p.fSubShowing <> nil) then begin
            // check nested submenu first
            result := CheckSubmenus(p.fSubShowing);
            if result <> nil then EXIT;
        end;
        for i := 0 to p.fsubmenu.count-1 do begin
            sub := p.fSubMenu[i];

            if (x >= sub.fLeft) and (x <= sub.fRight) then begin
                if (y <= sub.fBottom) and (y >= sub.fTop) then begin
                    result := sub;
                    EXIT;
                end;
            end;
        end;
    end;
begin
    // check submenus first
    result := CheckSubmenus(fitems.fSubShowing);
    if result <> nil then EXIT;

    for i := 0 to fitems.fsubmenu.count-1 do begin
        p := fitems.fsubmenu.Items[i];
        {same column}
        if (x >= p.fLeft) and (x <= p.fRight) then begin
            if (y <= p.fBottom) then begin
            	if p.isbreak then EXIT;

                result := p;
                EXIT;
            end;
        end;
    end;
end;
procedure TACPopup.AssignSubmenuShortcuts(p : TACPopupItem);
var i,j : integer;
begin
    if p.ItemType = IT_CLIPSET_HEADER then EXIT;

    if self.fKeyActivated and fShowExpandedShortcuts then begin
        if p.IsSubmenu or p.IsExpandable then begin
            self.fShowExpandedType := p.ItemType;
            j := 0;
            fExpandedKeystrokes.Clear;
            for i := 0 to p.fSubMenu.Count - 1 do begin
                with p.fSubMenu.Items[i] do begin
                    if not IsLine and not IsGap then begin
                        Prefix := '&' +self.GetAcceleratorPopupClips(j);
                        fExpandedKeystrokes.AddObject(Prefix[2], p.fSubMenu.Items[i]);
                        Inc(j);
                    end;
                end;
            end;
        end;
    end else begin
        if p.IsSubmenu or p.IsExpandable then begin
            for i := 0 to p.fSubMenu.Count - 1 do begin
                with p.fSubMenu.Items[i] do begin
                    if not IsLine and not IsGap then begin
                        Prefix := '';
                    end;
                end;
            end;
        end;
    end;
end;
procedure TACPopup.SetShowExpandedShortcuts(value : Boolean);
begin
    fShowExpandedShortcuts := value;
end;

function TACPopup.UseKeyboard : boolean;
begin
    result := frmconfig.cbUseKeyboard.Checked and (FrmMainPopup.GetNeedsFocus);
end;

{Drawing Routines}
procedure TACPopup.SetCanvas(c: TCanvas);
begin
    fNewCanvas := c;
end;
function TACPopup.GetCanvas: TCanvas;
begin
    //result := pnlMain.Canvas;
    if fNewCanvas = nil then begin
    	result := self.Canvas;
    end else begin
        result := fNewCanvas;
    end;
end;
function TACPopup.CalcHeight(fpi: TACPopupItem): integer;
begin
    result := max(
        UseCanvas.TextHeight('AZW') + fSpacing * 2,
        frmconfig.udMinHeight.Position
    );
end;
function TACPopup.CalcWidth(fpi: TACPopupItem; ClipIt : boolean=true): integer;
var wd : TSize;
    s : string;
begin
    result := LEFT_BORDER + LEFT_SPACE +LEFT_EDGE + fSpacing
            + UseCanvas.TextWidth(fpi.prefix) + fSpacing  + RIGHT_SPACE + RIGHT_BORDER;

    if (fpi.fClip<>nil) and (fpi.fClip.GetFormatType = FT_PICTURE) then begin
        // dib AAAxYYYY
        try
            s := fpi.fClip.GetFormatName;
            UnitToken.TokenString(s,')');
            s := UnitToken.TokenString(s,'x');
            result :=  StrToInt(s);
        except
        end;
    end else begin
        inc(result, UseCanvas.TextWidth(fpi.fCaption));
    end;

    if ClipIt then begin
    	result := min(result, self.fMaxItemWidth);
    end;
    result := max(150, result);
end;
procedure TACPopup.CalcTotalSize(FixSubmenu:boolean=false);
    function CalcRect(p : TACPopupItem; top,left : integer) : TRect;
    const GAP_SIZE = 20;
    var wd,ht,i,col, totalwd, maxht, originaltop : integer;
        sub : TACPopupItem;
        r : TRect;
        pop : TACPopupItem;
        procedure TallyColumn;
        begin
            AppendLog('New Column');
            inc(totalwd, wd);
            p.fColWidth[col] := wd;
            wd := 0;
            inc(col);
            ht := 0;
        end;
        procedure CalcItem(sub : TACPopupItem);
        var
            twd : integer;
            drawablePic : boolean;
            isWide : boolean;
        begin
           if sub.isBreak then begin
                sub.ftop := top+ht + TOP_BORDER;
                sub.fBottom := sub.fTop;
                sub.fleft := left + totalwd;
                sub.fMyColumn := col;
                TallyColumn;
            end else begin
                sub.ftop := top+ht + TOP_BORDER;
                if sub.IsLine then begin
                    ht := ht + fSpacing + 6;
                end else begin
                	twd := CalcWidth(sub, false);
                	wd := max(wd, min(self.fMaxItemWidth,twd));
                    drawablePic := (sub.Clip <> nil) and
                        (sub.Clip.GetFormatType = FT_PICTURE) and
                        (FrmConfig.cbNoThumbnails.Checked=false);
                    isWide := (twd > trunc(fMaxItemWidth*1.3));
                	if (sub.ItemType in TDoubleSized) and
                    	(drawablePic or (frmConfig.cbDoubleHeight.checked and isWide)) then begin
                    	ht := ht + (CalcHeight(sub)*2-fspacing*2);
                        sub.DoubleHeight := true;
                    end else begin
                    	ht := ht + CalcHeight(sub);
                        sub.DoubleHeight := false;
                    end;

                    if sub.Expanded then begin
                        inc(ht, 6); {expanded items are taller}
                    end;
                end;
                sub.fBottom := top+ht + TOP_BORDER;
                sub.fleft := left + totalwd;
                sub.fMyColumn := col;
            end;
            maxht := max(ht, maxht);
        end;

    begin
        wd := 0;
        ht := 0;
        originaltop := top;
        col := 0;
        totalwd := 0; maxht := 0;
        p.fColWidth[col] := 0;

        for i := 0 to p.fSubMenu.count - 1 do begin
            sub := p.fSubMenu.Items[i];
            CalcItem(sub);
        end;

        TallyColumn;
        result.top := originaltop;
        result.Left := left;
        result.Bottom := originaltop + maxht + TOP_BORDER+BOTTOM_BORDER;
        result.Right := left + (totalwd);
        p.fMaxHeight := maxht;

        if top=0 then begin
            fMainRect.Top := result.Top;
            fMainRect.left := result.Left;
            fMainRect.Right := result.Right;
            fMainRect.bottom := result.bottom;
        end;

        {no regions, no shadows}
       	if CheckWin32Version(6, 0) then
        if region1 = 0 then begin
            region1 := Windows.CreateRectRgnIndirect(result);
        end else begin
            region2 := Windows.CreateRectRgnIndirect(result);
            windows.CombineRgn(region1,region1,region2, RGN_OR);
            Windows.DeleteObject(region2);
        end;



        if p.fSubShowing <> nil then begin
            {lower the menu so the mouse won't hover it}
            if FixSubmenu then begin
                if fTopFixed and ((p.fSubShowing.fMaxHeight+p.fSubShowing.top)>fitems.fMaxHeight) then begin
                	{move the submenu up to be even with the bottom of the main popup}
                    {this prevents a recursvise shift-newhover loop}
                    r := CalcRect(p.fSubShowing,
                    	fitems.fMaxHeight - (p.fSubShowing.fMaxHeight+fspacing),
                        p.fsubshowing.right
                    );
                end else begin
//                    r := CalcRect(p.fSubShowing, p.fSubShowing.fbottom+1, p.fSubShowing.right);
                    if (p.fSubShowing.fSubMenu.Count <> 0) and
                        (not p.fSubShowing.fSubMenu.Items[0].isGap) then begin
                        pop := p.fSubShowing.add;
                        pop.SetIsGap(true);
                        p.fSubShowing.fSubMenu.Move(p.fSubShowing.fSubMenu.Count-1, 0);
                    end;

                    r := CalcRect(p.fSubShowing, p.fSubShowing.fTop, p.fSubShowing.right);
                end;

            end else begin
                r := CalcRect(p.fSubShowing,p.fSubShowing.ftop, p.fSubShowing.right);
            end;
            UnionRect(result, result, r);
        end;
    end;
var r : TRect;

    p : TACPopupItem;
    i : integer;
begin
    AppendLog('calcsize start');
    region1 := 0;
    r := CalcRect(fitems, 0, 0);

    self.Width := r.Right;
    self.Height := r.Bottom;
   	if CheckWin32Version(6, 0) then
		Windows.SetWindowRgn(self.Handle,region1,false);

    AppendLog('Find the last visible item');
    i := fitems.fSubMenu.count-1;
    if i >= 0 then begin
        repeat
            p := fitems.fSubMenu.GetItem(i);
            dec(i);
        until not p.IsBreak or (i=0);
    end;
    fLastVisible := p;

    {Windows.DeleteObject(region1);} // Windows owns this handle now
    self.EnsureVisibility(FixSubmenu);
    AppendLog('calcsize end');
end;
procedure TACPopup.EnsureVisibility(FixSubmenu:boolean=false);
var x,y : integer;
	m : TMonitor;
    expandedset : set of TItemType;
    i : integer;
begin
    {Fudge it up or left to keep it on the screen}
    x := fpoint.X;
    y := fpoint.y;
    fTopFixed := false;
    fLeftFixed := false;
    if (x + self.Width > screen.DesktopRect.right) then begin
        x := x - ((x+self.width+fspacing) - screen.desktoprect.right);
        fLeftFixed := true;
    end;

    // try turning on Clip Sets before altering the top position, if it's too tall
    // - make sure any previously expanded menus are re-expanded

    expandedset := [];
    if not self.fUseGrouping and (self.Height > screen.DesktopRect.Bottom) then begin
        if FrmConfig.cbEnableClipsSetsWhenTall.Checked then begin
            if (ClipQueue.GetQueueCount > FrmConfig.udClipsPerSet.Position) then begin
                self.SaveExpandedState;
                    Self.fUseGroupingOnce := True;
                    self.Clear;
                    self.AutoPopulate;
                self.LoadExpandedState;

                self.CalcTotalSize(FixSubmenu);
                Self.Invalidate;
            end;
        end;
    end;

    if (y + self.Height > screen.DesktopRect.Bottom) then begin
    	y := y - ((y+self.height+fspacing) - screen.desktoprect.bottom);
        fTopFixed := true;
        fIsTooTall := true;
    end;

    {update for mismatched monitor resolutions}
    m := screen.MonitorFromWindow(self.Handle);
    if (m <> nil) then begin
        if (y+self.Height) > m.BoundsRect.Bottom then begin
            //y := y - (m.BoundsRect.Bottom- (y+self.Height));
            fIsTooTall := true;
        end;
    end;



    {Recursive Call to fix visibility problems caused by submenus}
    if not FixSubmenu then begin
    	{Detect a submenu and a position change}
		FixSubmenu := fitems.fSubShowing <> nil;
        if FixSubmenu and (fLeftFixed or fTopFixed) then begin
            self.CalcTotalSize(FixSubmenu);
        end else begin
        	fIgnoreMouseMove := true;
            timIgnoreMouseMove.Enabled := true;
            self.Left := x;
            self.Top := y;
        end;
    end else begin
    	{Override the popup point until the popup is closed}
        fIgnoreMouseMove := true;
        timIgnoreMouseMove.Enabled := true;
        self.Left := x;
        self.Top := y;
        fpoint.x := x; {neither option is great, but this is the lesser evil}
        fpoint.y := y; {this causes a shift due to a side submenu to stay after}
                        {the submenu closes}
    end;
end;
procedure TACPopup.DrawCheckgroup(p: TACPopupItem);
var i : integer;
begin
    for i := 0 to p.fParent.fCheckgroup.count  - 1 do begin
        DrawItem( TACPopupItem(p.fParent.fCheckgroup[i]));
    end;
end;
procedure TACPopup.DrawDragmark;
var
	pa : Array of TPoint;
    r : TRect;
const wd = 7;
begin
	with UseCanvas do begin
        brush.Color := clActiveCaption;
        pen.color := fFontColor;
        r := fmainrect;
        InflateRect(r,-1,-1);
        setlength(pa,3);
        pa[0] := point(r.Right-wd,r.bottom);
        pa[1] := point(r.Right, r.bottom-wd);
        pa[2] := point(r.Right, r.Bottom);
        polygon(pa);
    end;
end;

procedure TACPopup.DrawItem(p: TACPopupItem);
var r : Trect;
    wc : TWideChar;
    s : string;
    len: integer;
    c : TColor;
var mainr : TRect;
    sd : TScrollDirection;
    prefixwidth : integer;
    clipboardleft : integer;
const
    BLEND_CLICKED = 97;
    BLEND_CLICKED_ORIGINAL = 99;
    procedure VerticalLines;
    begin
        // outside boundary line
        // reserve this space
        CopyRect(r,mainr);
        UseCanvas.pen.Color := fBorderColor;
        UseCanvas.MoveTo(r.Left,r.top);
        UseCanvas.LineTo(r.left,r.bottom);
        inc(r.Left);
        UseCanvas.pen.Color := dimColor(fBackgroundColor,1.02);
        UseCanvas.MoveTo(r.Left,r.top);
        UseCanvas.LineTo(r.left,r.bottom);

        {reserve this space}
        inc(mainr.Left);
    end;
    procedure LeftIconBackground;
    var c1, c2 : TColor;
    begin
        {icon column}
        CopyRect(r,mainr);
        inc(r.left);
        r.Right := r.left + (LEFT_SPACE-1);
        UseCanvas.brush.Style := bsSolid;
        UseCanvas.Brush.Color := fColumnColor;
        if p.IsInExpandedMenu then begin
            UseCanvas.Brush.Color := DimColorOrReverse(fBackgroundColor, 1.02);
        end;
        UseCanvas.FillRect(r);

        {focus rectangle}
        if (fhover = p) and (p.ItemType in TClickableLeftIcons) then begin
            if FLastPosition = pipLeftIcon then begin
                UseCanvas.Pen.Color := fFontColor;
                UseCanvas.Pen.style := psDot;
                UseCanvas.Rectangle(r);
                UseCanvas.Pen.style := psSolid;
            end;
        end;
    end;
    procedure CaptionBackground;
        procedure Fill(r : TRect);
        var b : boolean;
            c, c1,c2 : TColor;
        begin
            b := false;
            c := UseCanvas.Brush.Color;
            if (frmConfig.cbDoubleHeight.checked) and
                not ((p.fClip <> nil)and(p.fClip.CData.Clicked)) then begin
                if (p.fclip <> nil) and (p.ItemType = IT_MENU) and (p.fclip.GetFormatType <> FT_PICTURE) and
                    p.DoubleHeight then begin

                    UseCanvas.FillRect(r);
                    Inc(r.Top,(r.Bottom-r.Top) div 2);
                    c1 := c;
                    c2 :=  DimColorOrReverse(UseCanvas.Brush.Color, 0.96);
                    GraphUtil.GradientFillCanvas(UseCanvas,c1,c2,r,gdVertical);
                    with UseCanvas do begin
                        pen.Color := DimColorOrReverse(c2, 0.94);
                        MoveTo(r.Left+(fspacing+5), r.bottom-1);
                        LineTo(r.right-(fspacing+5), r.bottom-1);
                    end;
                end else begin
                    UseCanvas.FillRect(r);
                end;
            end else begin
                UseCanvas.FillRect(r);
            end;

        end;
    var c1, c2 : TColor;
    begin
        CopyRect(r,mainr);
        r.Left := p.fleft + LEFT_SPACE+LEFT_EDGE;
        UseCanvas.Brush.Style := bsSolid;
        UseCanvas.brush.Color := fBackgroundColor;
        if p.fClip <> nil then begin
            if p.fClip.CData.Clicked then begin
                UseCanvas.Brush.Color := unitMisc.Blend(fBackgroundColor, fClickedColor, BLEND_CLICKED_ORIGINAL);
                UseCanvas.brush.Color := dimcolor(UseCanvas.Brush.Color,1.01);
            end;
        end;


        if p.IsExpandable or (p.ItemType in [IT_CANCEL,IT_FULL,IT_SEARCH]) then begin
            if p.Expanded then begin
                c1 := DimColor(fBackgroundColor,0.94);
                c2 := Blend($3C3C3C,c1,55)
            end else begin
                c1 := DimColor(fBackgroundColor,0.975);
                c2 := Blend(clActiveCaption,fBackgroundColor,20);
            end;
            GraphUtil.GradientFillCanvas(UseCanvas,c1,c2,r,gdHorizontal);
        end else if p.IsInExpandedMenu then begin
            UseCanvas.brush.Color := fExpandedBackgroundColor;
            if (p.fClip <> nil) and p.fClip.CData.Clicked then begin
                UseCanvas.Brush.Color := unitMisc.Blend(UseCanvas.brush.Color, fClickedColor, BLEND_CLICKED);
            end;
            Fill(r);
        end else begin
            if (p.ItemType = IT_PINNED) then begin
                UseCanvas.brush.Color := DimColorOrReverse(fBackgroundColor,1.03);
            end;

            Fill(r);

            if (p.ItemType in TDoubleSized) and (frmConfig.cbDoubleHeight.checked) then begin
            	with UseCanvas do begin
					pen.Color := DimColorOrReverse(UseCanvas.brush.Color, 0.975);
                    MoveTo(r.Left+(fspacing+10), r.bottom-1);
                    LineTo(r.right-(fspacing+10), r.bottom-1);
                end;
            end;
        end;
    end;
    procedure LeftColumnEdge;
    begin
        CopyRect(r, mainr);
        r.Right :=  r.left + LEFT_SPACE;
        UseCanvas.pen.Color := fColumnEdgeColor;
        UseCanvas.MoveTo(r.right,r.top);
        UseCanvas.LineTo(r.right,r.bottom);
    end;
    procedure DrawLine;
    begin
        {edge in between the rows}
        Copyrect(r, mainr);
        dec(r.Right);
        r.Left := p.fleft + LEFT_SPACE+LEFT_EDGE;
        r.Top := p.fTop + fSpacing+1;
        r.Bottom := r.top+1;
        UseCanvas.Brush.Color := 0;
        UseCanvas.pen.Color := dimColor(fBackgroundColor,0.95);
        UseCanvas.MoveTo(r.Left, r.top);
        UseCanvas.LineTo(r.Right, r.top);
        UseCanvas.pen.Color := dimColor(fBackgroundColor,1.10);
        UseCanvas.MoveTo(r.Left, r.Bottom);
        UseCanvas.LineTo(r.Right, r.bottom);
    end;
    procedure DrawHoverHilite;
    var startc : TColor;
        back : TColor;
    begin
        // Since it's only called once, there's no need to optimized the
        // colors

        copyrect(r,mainr);
        r.Left := r.Left+ LEFT_SPACE+LEFT_EDGE;
        dec(r.right, RIGHT_BORDER);
        {focus rectangle}

        c := fHighlightColor;
        UseCanvas.pen.Color := DimColorOrReverse(c);
        UseCanvas.RoundRect(
            r.left,r.top,r.Right,r.bottom,
            2,2
        );
        {inner gradient}
        inc(r.Top);
        inc(r.left);
        dec(r.Right);
        dec(r.bottom);

        back := fBackgroundColor;
        if p.IsInExpandedMenu then begin
            back := DimColorOrReverse(fBackgroundColor, 1.04);
        end;
        back := unitMisc.Blend(back, fHighlightColor, 90);

        if (p.fClip <> nil) and
            (p.fclip.CData.Hash = CurrentClipboard.CData.hash) and
            (p.ItemType = IT_MENU) then begin

            startc := dimcolor(back,0.95);
            c := UnitMisc.Blend(startc,fHighlightColor,85); // was 85
        end else begin
            startc := ColorToRGB(back);
            c := UnitMisc.Blend(startc,fHighlightColor,35); // was 85
        end;
        GraphUtil.GradientFillCanvas(
            UseCanvas,startc,c,
            r,gdVertical
        );
    end;
    procedure LeftIcon;
    var ht : integer;
        wd : integer;
        bm : TBitmap;
        procedure DrawBox(checked : boolean);
        var
            center, chevronwd, chevronht : integer;
        begin
            copyrect(r,mainr);

            r.right := r.left+LEFT_SPACE;
            center := (r.Right+r.left) div 2;
            chevronwd := 14;
            chevronht := 14;
            inc(r.top, ((r.Bottom-r.Top) - chevronht) div 2);
            r.bottom := r.top + chevronht;
            r.left := (center - chevronwd div 2);
            r.Right := (center + chevronwd div 2);
            inc(r.right);

            UseCanvas.brush.Color := dimColor(fBackgroundColor,1.01);
            UseCanvas.Brush.Style := bsSolid;
            UseCanvas.pen.Color := saturate(clActiveCaption, 0.10);
            UseCanvas.Rectangle(r);

            UseCanvas.pen.Color := clBtnText;
            UseCanvas.Brush.Style := bsSolid;
            inc(r.Top, chevronht div 2);
            inc(r.Left, chevronwd div 2 - 3);
            if checked then begin
                GraphUtil.DrawCheck(UseCanvas,point(r.left,r.top),2);
            end;
        end;
        procedure DrawHIcon(wd,ht : integer);
        begin
            DrawIconEx(UseCanvas.handle,
                r.left, r.top,
                p.ficon, wd,ht, 0, 0, DI_NORMAL);
        end;
    begin
        copyrect(r,mainr);
        r.right := r.left + left_space;

        case p.Style of
        psNormal, psExpandable, psSubmenu : begin
            if ((p.fLeftIconOnlyOnHover) and (p <> Hover)) and
                not (p.Expanded) then begin
                {do nothing}
            end else begin
                wd := r.Right-r.left;
                ht := r.bottom-r.top;

                if (fhover=p) and (p.ItemType in TContainsAClip) and (self.FLastPosition = pipLeftIcon) then begin
                    if (ht > 18)  then begin
                        {center it vertically}
                        r.Top := r.top + (ht - 16) div 2;
                        r.Left := r.left + (wd - 16) div 2;
                    end else begin
                        r.Top := r.top + (ht-(ht-2)) div 2;
                        r.Left := r.left + (wd -(ht-2)) div 2;
                    end;
                    UseCanvas.Draw(r.Left,r.Top, ImgToBitmap(FrmMainPopup.ImgEdit));
                end else begin
                    if (ht > 18)  then begin
                        {center it vertically}
                        r.Top := r.top + (ht - 16) div 2;
                        r.Left := r.left + (wd - 16) div 2;
                        if p.fLeftBitmap <> nil then begin
                            r.right := r.Left + p.fLeftBitmap.Width;
                            r.bottom := r.top + p.fLeftBitmap.Height;

                            UseCanvas.draw(r.left, r.Top, p.fLeftBitmap);
                        end else if p.ficon <> 0 then begin
                            DrawHIcon(16,16);
                        end;
                    end else begin
                        r.Top := r.top + (ht-(ht-2)) div 2;
                        r.Left := r.left + (wd -(ht-2)) div 2;
                        {scale it to the menu height}
                        if p.fLeftBitmap <> nil then begin
                            r.right := r.Left + p.fLeftBitmap.Width;
                            r.bottom := r.top + p.fLeftBitmap.Height;
                            UseCanvas.StretchDraw(r,p.fLeftBitmap);
                        end else begin
                            DrawHIcon(ht-2,ht-2);
                        end;
                    end;

                    if p.ItemType = IT_PINNED then begin
                        Inc(r.left,8); //8x8 icon
                        Inc(r.Top,7);
                        UseCanvas.Draw(r.Left, r.Top, FrmClipboardManager.iPin.Picture.Graphic);
                    end;
                end;
            end;
        end;
        psCheckable: begin
            if not p.IsCheckGrouped  then begin
                DrawBox(p.checked);
            end else if p.Checked then begin
                DrawBox(p.checked);
            end;
        end;
        end;
    end;
    procedure Caption;
        procedure DrawDIB(canvas : TCanvas; pnt : Pointer; dest: TRect);  overload;
        var
            bi : PBitmapInfo;
            buf : Pointer;
            r, originalr : Trect;
            wd, ht : integer;
            ci : TClipItem;
            pic : TPicture;
            iratio : double;
            needsScale : boolean;
            tinywd, tinyht : boolean;
        begin
            with p.fClip.CData do begin
            if Assigned(thumb) then begin
                if (thumb.Width = dest.Width) and (thumb.Height = dest.Height) then begin
                    canvas.Draw(dest.Left,dest.Top, thumb);
                end else begin
                    canvas.StretchDraw(dest, thumb);
                end;
                EXIT;
            end;
            end;

            bi := pnt;
            buf := Pointer(Integer(bi) + SizeOf(Windows.BITMAPINFO));

            originalr := r;
            r.Left := 0;
            r.Top := max(0, bi^.bmiHeader.biHeight-dest.height);
            r.Height := bi^.bmiHeader.biHeight;
            r.Width := bi^.bmiHeader.biWidth;

            needsScale := false;
            if r.Height > dest.Height then begin
                r.Height := dest.Height;
            end;
            if r.Width > dest.Width then begin
                r.width := dest.Width;
                needsScale := true;
            end;
            if r.height < dest.height then begin
                // center a small picture in the middle of the destination box
                dest.top := dest.Top + (dest.height-r.height) div 2;
            end;

            tinywd := false;
            tinyht := false;
            if bi^.bmiHeader.biWidth < dest.Width then begin
                dest.Width := bi^.bmiHeader.biWidth;
                tinywd := true;
            end;
            if bi^.bmiHeader.biHeight < dest.Height then begin
                dest.top := dest.Top + (dest.height-r.height) div 2;
                dest.Height := bi^.bmiHeader.biHeight;
                tinyht := true;
            end;

            with p.fClip.CData do begin
                if true then begin
                    // performs slightly faster
                    // on my fast machine, it's 16ms without caching and 0ms with caching
                    if not Assigned(thumb) and not (tinywd and tinyht) then begin
                        thumb := TBitmap.Create;
                        thumb.PixelFormat := pf24bit;
                        thumb.Canvas.Brush.Style := bsSolid;
                        thumb.Canvas.Brush.Color := clNone;
                    end;
                    if assigned(thumb) and
                        ((thumb.Width <> dest.Width) or (thumb.Height <> dest.Height)) then begin
                        thumb.Width := dest.Width;
                        thumb.Height := dest.Height;
                        if (needsScale) then begin
                            // scale the destination box up to the size of the original
                            // compare the widths to caculate the percent growth needed
                            // apply this percent to the box's height
                            wd := bi^.bmiHeader.biWidth;
                            ht := trunc( (wd/dest.Width ) * dest.Height ) ;
                            r.top := bi^.bmiHeader.biHeight-ht;
                            r.Width := wd;
                            r.Height := ht;
                        end;
                        SetStretchBltMode(thumb.Canvas.Handle, STRETCH_HALFTONE);
                        StretchDIBits(thumb.Canvas.Handle,
                        0,0,dest.Width,dest.Height,
                        r.Left,r.Top,r.Width,r.Height,
                        buf, bi^, DIB_RGB_COLORS, SRCCOPY
                        );
                    end;
                    SetStretchBltMode(Canvas.Handle, STRETCH_DELETESCANS);
                    if (thumb<>nil) then begin
                        canvas.Draw(dest.Left,dest.Top, thumb);
                    end else begin
                        StretchDIBits(Canvas.Handle,
                            dest.left, dest.top, dest.width, dest.Height,
                            r.left, r.top, r.width, r.height,
                            buf, bi^, DIB_RGB_COLORS, SRCCOPY
                        );
                    end;
                end else begin
                    StretchDIBits(Canvas.Handle,
                        dest.left, dest.top, dest.width, dest.Height,
                        r.left, r.top, r.width, r.height,
                        buf, bi^, DIB_RGB_COLORS, SRCCOPY
                    );
                end;
            end;
        end;
    var
        //pic : TPicture;
    	rect2, rect3 : TRect;
        options : integer;
        procedure TextColorSet;
        var bt : byte;
        begin
            UseCanvas.brush.Style := bsClear;
            UseCanvas.pen.Color := 0;
            if p.Disabled then begin
                UseCanvas.pen.Color := fDisabledColor;
            end else begin
                UseCanvas.pen.Color := fFontColor;
                if (p.fClip <> nil) and p.fClip.CData.Clicked then begin
                    UseCanvas.pen.Color := fClickedColor;
                end;
            end;

//            if p.IsInExpandedMenu then begin
//                if Brightness(UseCanvas.Pen.Color) > 80 then begin
//                    UseCanvas.Pen.Color := dimColor(fFOntCOlor);
//                end;
//            end;

            SetTextColor(UseCanvas.Handle, ColorToRGB(UseCanvas.pen.Color));
        end;
        procedure DrawPic;
        var
            st : TStream;
            pnt : Pointer;
            sz : Int64;
            tempr : TRect;
        begin
            if not assigned(p.Clip.CData.thumb) then begin
                p.Clip.BuildMenuCaption;
            end;

            if assigned(p.Clip.CData.thumb) then begin
                //
                // Detect already scaled thumbs
                // Otherwise, just draw the entire picture or a crop of
                // the picture
                //
                with p.Clip.CData do begin
                    if (thumb.Width = r.Width) and (thumb.Height = r.Height) then begin
                        canvas.Draw(r.Left,r.Top, thumb);
                    end else if (thumb.Width + (RIGHT_SPACE + fSpacing) = r.Width) and
                        (thumb.Height = r.Height) then begin
                        canvas.Draw(r.Left,r.Top, thumb);
                    end else begin
                        tempr.Top := r.top;
                        tempr.Left := r.left;
                        tempr.Width := min(thumb.Canvas.ClipRect.Width, r.Width);
                        tempr.Height := min(thumb.Canvas.ClipRect.Height, r.Height);
                        if (tempr.Height < r.Height) then begin
                            tempr.Top := tempr.Top + ((r.Height-tempr.Height) div 2);
                        end;

                        Canvas.CopyRect(tempr, thumb.Canvas, rect(0,0,tempr.width, tempr.Height));
                    end;
                end;
            end else begin
                st := p.clip.getstream;
                sz := st.Size;
                pnt := GetMemory(sz);
                st.Read(pnt^, sz);
                p.Clip.FinishedWithStream;
                DrawDIb(UseCanvas, pnt, r);
                freemem(pnt);
            end;

            // Windows is doing something shady. Unless the BitmapInfo and the bit
            // buffer are continguous, the StretchDIBits call is failing
//            DrawDib(usecanvas, p.Clip.GetStream, r);
        end;
        procedure DrawPrefix;
        var b : boolean;
            c : TColor;
            bt : byte;
            ft : TFontStyle;
        begin
            wc.Appendunicode(p.Prefix);
            len := wc.StrLength;

            c := UseCanvas.Pen.Color;
            if c <> clGrayText then
                UseCanvas.Pen.Color := DimColorOrReverse(c, 1.05);



            ft := fsBold;
            b := not (ft in UseCanvas.Font.Style);
            if b then
                UseCanvas.Font.Style := UseCanvas.Font.Style + [ft];
            SetTextColor(UseCanvas.Handle, ColorToRGB(UseCanvas.pen.Color));

            prefixwidth := max(UseCanvas.TextWidth('ZZ')+1, UseCanvas.TextWidth(ReplaceStr(p.prefix,'&','')));
            r.Right := r.Left + prefixwidth;

            DrawTextW(UseCanvas.Handle, PWideChar(wc.Memory), len,
                r, DT_VCENTER or DT_SINGLELINE );

            if b then
                UseCanvas.Font.Style := UseCanvas.Font.Style - [ft];

            UseCanvas.Pen.Color := c;
            SetTextColor(UseCanvas.Handle, ColorToRGB(UseCanvas.pen.Color));
        end;
        procedure DrawClipboardMark;
        begin
            with FrmMainPopup.imgPasteMini.Picture do begin
                r.Left := clipboardleft;
                UseCanvas.Draw(
                    r.left-2,
                    Max(r.Top + (r.Height - Bitmap.height) div 2, 0),
                    Bitmap
                );
                r.Left := r.left + Width + fspacing*2;
            end;
        end;
    var i : integer;

    begin
        {prefix}
        CopyRect(r,mainr);
        inc(r.top, fspacing);
        dec(r.Bottom,fspacing);
        r.Left := r.left + LEFT_SPACE+LEFT_EDGE+fspacing;
        r.right := r.Right - (RIGHT_SPACE + fSpacing);

        TextColorSet;
        wc := TWideChar.Create;
        prefixwidth := 0;

        if (p.Prefix <> '') then begin
            if (fShowExpandedShortcuts) then begin
                if (p.fParent.ItemType <> self.fShowExpandedType) then begin
                    if self.fExpandedKeystrokes.IndexOf(p.Prefix[2]) <> -1 then begin
                        // already used by submenu, don't draw it
                    end else begin
                        if p.IsInSubmenu or p.IsInExpandedMenu and (p.ItemType <> IT_MENU) then begin

                        end else begin
                            DrawPrefix;
                        end;
                    end;
                end else begin
                    DrawPrefix;
                end;
            end else begin
                if p.IsInSubmenu or p.IsInExpandedMenu and (p.ItemType <> IT_MENU) then begin

                end else begin
                    DrawPrefix;
                end;
            end;
        end;

        myfree(wc);

        { main text }
        CopyRect(r,mainr);
        inc(r.top, fspacing);
        dec(r.Bottom,fspacing);
        if ((p.ItemType = IT_CLIPBOARD)
            or
            (
                (p.ItemType = IT_MENU) and (p.Clip.CData.Hash = CurrentClipboard.CData.Hash)
            ))
            then begin

            clipboardleft := r.left + LEFT_SPACE+LEFT_EDGE + prefixwidth;
            if not ((p.Clip <> nil) and (p.Clip.GetFormatType = FT_PICTURE)) then begin
                DrawClipboardMark;
            end else begin
                r.Left := r.left + LEFT_SPACE+LEFT_EDGE + prefixwidth+fspacing;
            end;
        end else begin
            r.Left := r.left + LEFT_SPACE+LEFT_EDGE + prefixwidth+fspacing;
        end;

        if ((p.Style in [psNormal]) and (p.fHasRightIcon))  then begin
        	r.right := r.Right - (RIGHT_SPACE + fSpacing);
        end else begin
        	r.right := r.Right - (fSpacing*3);
        end;


        wc := TWideChar.Create;
        if (p.fClip = nil) or (p.caption <> '') or (p.fClip.GetFormat <> CF_UNICODETEXT) then begin
            if (p.fClip = nil) or (p.caption <> '') then begin
                s := p.caption;
            end else begin
                if p.fClip.GetFormat = CF_DIB then begin
                    s := p.fClip.GetFormatName;
                end else begin
                    s := p.fclip.GetAsPlaintext;

                end;
            end;
            wc.AppendUnicode(s);
            wc.Replace(WideChar(' '),{NO BREAK SPACE}WideChar($00A0));
        end else begin
            wc.AppendUnicode(p.fClip.GetAsPlaintext);
        end;

        len := min(100, wc.StrLength);
        SetBkMode(UseCanvas.Handle, TRANSPARENT) ;


        options := 0;
        if p.ItemType in TContainsAClip then begin
        	options := options or DT_NOPREFIX;
        end;
        // rect2 is either the size of the caption, or smaller if the picture is smaller

        TextColorSet;
        if (FrmConfig.cbNoThumbnails.Checked=false) and
            ((p.Clip<>nil) and (p.Clip.GetFormatType = FT_PICTURE)) and
            (p.ItemType in TContainsAClip) then begin
            DrawPic;
            if (p.ItemType = IT_CLIPBOARD)
                or
            ((p.ItemType = IT_MENU) and
                (p.Clip.CData.Hash = CurrentClipboard.CData.Hash)) then begin
                r.Left := clipboardleft-9;
                DrawClipboardMark;
            end;
        end else begin
            if p.DoubleHeight then begin
                DrawTextW(UseCanvas.Handle, PWideChar(wc.Memory), len,
                    r, DT_VCENTER or DT_WORDBREAK or DT_EDITCONTROL or options  );
            end else begin
                DrawTextW(UseCanvas.Handle, PWideChar(wc.Memory), len,
                    r, DT_VCENTER or DT_SINGLELINE or options);
            end;
        end;


        myfree(wc);

        SetBkMode(UseCanvas.Handle, OPAQUE);
    end;
    procedure RightIcon;
    var center : integer;
        chevronwd, chevronht, wd, ht : integer;
        bm : TBitmap;
    begin
        copyrect(r,mainr);
        r.Left := r.Right - (RIGHT_SPACE+RIGHT_BORDER);

        UseCanvas.brush.Style := bsSolid;
        case p.Style of
        psSubmenu: begin
            UseCanvas.pen.Color := fFontColor;
            GraphUtil.DrawArrow(UseCanvas,sdRight,
                point(r.right-fspacing*2-RIGHT_BORDER-3,(r.top+r.bottom)div 2 -4),3);
        end;
        psExpandable: begin

            if p.Expanded  then begin
                sd := sdUp;
                bm := ImgToBitmap(FrmMainPopup.ImgUp);
            end else begin
                sd := sdDown;
                bm := ImgToBitmap(FrmMainPopup.imgDown);
            end;

            if not ((sd=sdup) and not p.ShowCollapseIcon) then begin

                copyrect(r,mainr);
                dec(r.right);
                r.left := (r.right - (RIGHT_SPACE)) and not 1;
                center := (r.Right+r.left) div 2;
                chevronwd := 14;
                chevronht := 14;
                inc(r.top, ((r.Bottom-r.Top) - chevronht) div 2);
                r.bottom := r.top + chevronht;
                r.left := (center - chevronwd div 2);
                r.Right := (center + chevronwd div 2);
                inc(r.right);

                UseCanvas.brush.Color := clNone;
                UseCanvas.Brush.Style := bsClear;
//                UseCanvas.pen.Color := saturate(clActiveCaption, 0.10);
                UseCanvas.pen.Color := DimColorOrReverse(fFontColor);
                UseCanvas.Rectangle(r);

                if p.Disabled then begin
                    UseCanvas.pen.Color := clGrayText;
                end else begin
                    UseCanvas.pen.Color := fFontColor;
                end;
                UseCanvas.Brush.Style := bsSolid;
                inc(r.Top, chevronht div 2 - 1);
                inc(r.Left, chevronwd div 2 - 3);
                GraphUtil.DrawChevron(UseCanvas,sd,point(r.left,r.top),3);
            end;
        end;
        psNormal: begin
            if p.fHasRightIcon  then begin
                wd := r.Right-r.left;
                ht := r.bottom-r.top;

                UseCanvas.Brush.Color := fBackgroundColor;
                {center it vertically}
                r.Top := r.top + (ht - p.fRightBitmap.Height) div 2;
                r.Left := r.left + (wd - p.fRightBitmap.Width) div 2;

                {focus rectangle}
                if (FLastPosition = pipRightIcon) and (self.fHover = p)  then begin
                    dec(r.top, 2); dec(r.left, 2);
                    r.Right := r.Left + p.fRightBitmap.Width + 4;
                    r.Bottom := r.Top + p.fRightBitmap.Height + 4;
                    UseCanvas.Pen.Color := fFontColor;
                    UseCanvas.Pen.style := psDot;
                    UseCanvas.Rectangle(r);
                    UseCanvas.Pen.style := psSolid;
                    inc(r.top, 2); inc(r.left, 2);
                end;

                UseCanvas.Draw(r.left, r.top, p.fRightBitmap);
                {
                DrawBitmap(UseCanvas.handle,p.fRightBitmap,r.top, r.left,
                    p.fRightBitmap.Width,p.fRightBitmap.Height
                );
                }
            end;
        end;
        end;
    end;
    procedure DrawHoverLine;
    begin
    	copyrect(r,mainr);
        inc(r.Top);
        inc(r.Left, LEFT_SPACE+LEFT_EDGE);
        dec(r.Right, fSpacing); 
    	if (fDragStarted) and (p=fDropMenuItem)  then begin
            if (fDragMenuItem.ItemType = IT_MENU) or (p.ItemType = IT_MENU) then begin
             	if (fDragMenuItem.ItemType = IT_MENU) and (p.ItemType = IT_MENU) then begin
                    UseCanvas.pen.Color := fFontColor;
                    UseCanvas.MoveTo(r.Left,r.top);
                    UseCanvas.LineTo(r.right,r.top);
                end;
            end else if (fDragMenuItem.ItemType = IT_PERMANENT) or (p.ItemType = IT_PERMANENT) then begin
             	if (fDragMenuItem.ItemType = IT_PERMANENT) and (p.ItemType = IT_PERMANENT) then begin
                    UseCanvas.pen.Color := fFontColor;
                    UseCanvas.MoveTo(r.Left,r.top);
                    UseCanvas.LineTo(r.right,r.top);
                end;
            end else begin
            	if not (p.ItemType in [IT_LINE, IT_NONE]) then begin
                    UseCanvas.pen.Color := fFontColor;
                    UseCanvas.MoveTo(r.Left,r.top);
                    UseCanvas.LineTo(r.right,r.top);
                end;
            end;
        end;
    end;
begin
    p.SetBoundsRect(self.Top, self.left);

    {main item rect}
    mainr.Top := p.fTop;
    mainr.Left := p.fLeft;
    mainr.Right := p.Right;
    mainr.Bottom := p.fbottom;
    p.fRight := mainr.Right;

    UseCanvas.Brush.Style := bsSolid;
    UseCanvas.Brush.Color := fBackgroundColor;

    VerticalLines;

    LeftIconBackground;
    LeftColumnEdge;
    CaptionBackground;

    if (p.IsBreak)  then begin
        {nothing needed}
    end else if (p.IsLine) then begin
        DrawLine;
    end else begin
        if (p = hover) and (not p.isGap) then begin
            DrawHoverHilite
        end;
        LeftIcon;
        Caption;
        RightIcon;
    end;
    DrawHoverLine;

    if p = fLastVisible then begin
    	DrawDragmark; 
    end;
end;
procedure TACPopup.DrawOnCanvas(fullmode : boolean=false);
begin
	self.fFullMode := fullmode;

    UseCanvas.Font := FrmConfig.GetFont;
    UseCanvas.Font.Size := FrmConfig.GetFont.Size;
    UseCanvas.Font.Style := FrmConfig.getfont.style;

    self.CalcTotalSize;
    self.Paint;
end;
procedure TACPopup.FullRedraw;
    procedure DrawMenu(p : TACPopupItem; const top,left : integer);
    var i : integer;
        sub, sub2 : TACPopupItem;
        r : TRect;





    begin
        if p.fSubMenu.Count = 0 then EXIT;

        r.top := top;
        r.Left := left;
        r.Right := p.fSubMenu[p.fSubMenu.Count-1].right;
        r.Bottom := top+p.fmaxheight;

        usecanvas.Brush.color := fBackgroundColor;
        UseCanvas.FillRect(r);

        {top line}
        UseCanvas.pen.color := fBorderColor;
        UseCanvas.Pen.Width := 1;
        UseCanvas.MoveTo(r.left, r.top);
        UseCanvas.lineto(r.right, r.top);
        inc(r.bottom);
        UseCanvas.pen.color := fBorderColor;
        UseCanvas.MoveTo(r.left-1, r.bottom);
        UseCanvas.lineto(r.right+1, r.bottom);
        {bottom line}
        UseCanvas.pen.color := fBorderColor;
        UseCanvas.MoveTo(r.left-1, r.bottom);
        UseCanvas.lineto(r.right+1, r.bottom);
        {right line}
        UseCanvas.pen.color := fBorderColor;
        UseCanvas.MoveTo(r.right, r.top);
        UseCanvas.lineto(r.right, r.bottom);
        // draw individual items
        for i := 0 to p.fsubmenu.count - 1 do begin
            sub := p.fsubmenu.Items[i];
            if sub.IsBreak  then begin
                // handle blank space
                if r.bottom - sub.fBottom > BOTTOM_BORDER then begin
                    sub.fBottom := r.Bottom;
                    self.DrawItem(sub);
                end;
            end else begin
                self.DrawItem(sub);
            end;
        end;
        // handle blank space
        if r.bottom - sub.fBottom > BOTTOM_BORDER then begin
            sub2 := TACPopupItem.create;
            sub2.ftop := sub.fbottom;
            sub2.fleft := sub.fleft;
            sub2.fRight := sub.fRight;
            sub2.fMyColumn := sub.fMyColumn;
            sub2.fParent := sub.fParent;
            sub2.fBottom := r.Bottom;
            sub2.IsBreak := true;
            sub2.ItemType := IT_MENU;

            self.DrawItem(sub2);
        end;

        if p.fSubShowing <> nil then begin
            DrawMenu(p.fSubShowing, p.fSubShowing.fTop, p.fSubShowing.Right);
        end;
    end;

begin

    drawmenu(fitems, 0,0);
    FirstPaint := true;

    DrawDragmark;


    if self.FormMode then begin
        SetWindowLong(self.Handle,GWL_EXSTYLE, GetWindowLong(self.handle, GWL_EXSTYLE) or WS_EX_LAYERED);
        SetLayeredWindowAttributes(self.handle,0,trunc(255 * 0.80), LWA_ALPHA);
    end else begin
		SetWindowLong(self.Handle,GWL_EXSTYLE, GetWindowLong(self.handle, GWL_EXSTYLE) and not WS_EX_LAYERED);
    end;


end;

procedure TACPopup.AddAScrollbar;
begin
   { r := self.fMainRect;
	i := GetSystemMetrics(SM_CYVTHUMB);

    self.Width := self.Width + i*2;
    inc(r.Right,i);

	scroll := CreateWindow('SCROLLBAR', nil,
    WS_CHILD or WS_VISIBLE or SBS_VERT or SBS_RIGHTALIGN,
    r.left,r.top,r.right-i,r.bottom,
    self.Handle,
     IDC_VSCROLL,
     hInstance,
     nil);

	SetScrollRange(scroll, SB_CTL, 0, self.Height, FALSE);
    }

end;


{Clip Menu Routines}
function TACPopup.ClipMenuVisible : Boolean;
begin
    result := fim.Visible;
end;
procedure TACPopup.ClipMenuReportKey(key : Char);
begin
    fim.ReportKeypress(key);
end;
procedure TACPopup.ClipMenuReportKey(key : word);
begin
    fim.ReportKeypress(key);
end;

procedure TACPopup.ClipMenuHide;
begin
//    FrmClipMenuNew.HideForm;
    fim.Hide;
end;
procedure TACPopup.ClipMenuShow(rightclick : boolean = false);
    procedure CheckForVisibility;
    begin
        fim.Top := max(screen.DesktopRect.top, fim.top);
        fim.Left := max(screen.DesktopRect.left, fim.Left);

        if fim.Width + fim.Left > screen.DesktopRect.right  then begin
            fim.left := screen.DesktopRect.right - fim.Width;
        end;
        if fim.Height + fim.Top > screen.DesktopRect.bottom then begin
            fim.top := screen.DesktopRect.bottom - fim.Height;
        end;
    end;
begin
	if fhover = nil then EXIT;

//    FrmClipMenuNew.Top := self.Top + fhover.Top;
//    FrmClipMenuNew.Left := mouse.CursorPos.X - 40;
//    FrmClipMenuNew.showform;
//    EXIT;

    {under or leftside}
    if rightclick then begin
        fim.Top := self.Top + fhover.top;
        {fim.Left := self.left + fhover.fleft;}
        fim.Left := mouse.CursorPos.X - 40;
    end else begin
        if fim.Top <> fhover.ftop then begin
            fim.Top := self.Top + fhover.ftop-2;
            fim.Left := self.left + fhover.fleft -  fim.Width;
        end;
    end;


    case fhover.ItemType of
        IT_PERMANENT: begin
            fim.SetPermanentClipMode;
            fim.SetPermanentClip(fhover.fPermanentGroupID, fhover.fPermanentID);
            fim.SetACPopupItem(fhover);
            CheckForVisibility;
            fim.show;
            ForceForeground(fim.handle);
        end;
        IT_MENU,
        IT_TEMP,
        IT_CLIPBOARD,
        IT_CLICKED,
        IT_PINNED: begin
                fim.SetPopupClipMode;
                fim.SetPopupClip(fHover.Clip);
                fim.SetACPopupItem(fhover);
                CheckForVisibility;
                fim.show;
                ForceForeground(fim.handle);
        end;
    end;

end;
procedure TACPopup.DeleteClipCallback(Sender: TObject);
begin
    self.AutoPopulate;
    self.CalcTotalSize;
    self.Invalidate;
end;
procedure TACPopup.MakePermanentCallback(Sender: Tobject);
begin
    self.hide;
end;
procedure TACPopup.EditCallback(Sender: Tobject);
begin
    self.Hide;
end;
procedure TACPopup.PasteCallback(Sender: Tobject);
begin
    self.hide;
end;
procedure TACPopup.ReserveAccel;
var c : char;
begin
	//inc(fReservedCount);
    c := self.GetAccelerator(0);
    while pos(c, fAccelCharsUsed) <> 0 do begin
        inc(fReservedCount);
    	c := self.GetAccelerator(0);
    end;
    inc(fReservedCount);
end;

procedure TACPopup.DestroyCallback(Sender: Tobject);
begin
    self.AutoPopulate;
    self.CalcTotalSize;
    self.Invalidate;
end;

{Mouse Events}
procedure TACPopup.CMMouseLeave(var Message: TMessage);
begin
    fToolTipNew.Hide;
    FLastPosition := pipNone;
end;
procedure TACPopup.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var p : TACPopupItem;

begin
    inherited;
    p := GetItemAt(x, y);

    if DragMouseUp(p) then EXIT;
    self.HandleMouseClick(p);
end;
procedure TACPopup.MouseDown(Button: TMouseButton; Shift: TShiftState; X,Y: Integer);
begin
    inherited;

    timHover.Enabled := false;

    fDownShift := shift;
    if fHover = nil then begin
    	fMouseDownOn := GetItemAt(x,y);
    end else begin
    	fMouseDownOn := fHover;
    end;

    self.DragMouseDown(fHover);
end;
procedure TACPopup.MouseLeave(Sender: TObject);
begin
    fToolTipNew.Hide;
end;
procedure TACPopup.MouseMove(Shift: TShiftState; X, Y: Integer);
var last : TACPopupItem;
    lastpos : TPopupItemPosition;
    pt : TPoint;

    procedure ClearSubmenus(pc : TACPopupItem);
    begin
        repeat
             pc.fSubShowing := nil;
             pc := pc.fParent;
        until pc = nil;
    end;
    function IsChild(parent, pc : TACPopupItem) : boolean;
    begin

        result := false;
        if pc = nil then EXIT;

        repeat
             if Parent = pc.fparent then begin
                //result := true;
                 BREAK;
             end;

             pc := pc.fParent;
        until pc.fParent = nil;

        result := not (pc = fItems);
    end;
var p : TACPopupItem;
	m : TMonitor;
begin
    inherited;

    if fIgnoreMouseMove then begin
        lastmousex := x;
        lastmousey := y;
    	EXIT;
    end;

    last := Hover;

    if (fIsTooTall) then begin
        pt := mouse.CursorPos;
        m := screen.MonitorFromPoint(pt);
        if m <> nil then begin
            pt.Y := pt.Y + m.top;
            if (pt.y = m.boundsrect.Top) or
            	(pt.y = m.BoundsRect.bottom+m.boundsrect.top-1) then begin
                if (pt.y = m.boundsrect.Top) then begin
                	self.top := m.top;

                end else begin
                	self.Top := m.top - (self.Height - m.Height);
                end;
                EXIT;
            end;
        end;
    end;
    if (Lastmousex=x) and (lastmousey=y) then begin
        {Mouse Move gets triggers after showing a tooltip, even though
        nothing has changed}
        EXIT;
    end;

    if self.fKeyActivated then begin
        self.fKeyActivated := false;
        self.ShowExpandedShortcuts := false;
        self.Invalidate;
    end;

    if not fDragStarted and (fDragMenuItem <> nil) then begin
        if (abs(fDragX - mouse.cursorpos.x) > mouse.DragThreshold) or
            (abs(fdragy - mouse.cursorpos.y) > mouse.DragThreshold) then begin

            fDragStarted := true;
            screen.Cursor := crDrag;
            //fMouseDownOn := nil;
        end;
    end;

    LastMouseX := x;
    LastMouseY := y;

    p := GetItemAt(x,y);
    if (p <> nil) and p.IsLine then begin
        SetHover(p); // Windows will show no hover over a line
        fToolTipNew.Hide;
        if not p.IsInSubmenu  then begin
            timHideSubmenu.Enabled := true;
        end;

        if last <> nil then
            DrawItem(last);

        EXIT;
    end;
    if self.DragMouseMove(p, last) then EXIT;

    if (last <> p) then begin
        if (p <> nil) then begin
            self.CancelShowSubmenu;
            {current is a submenu, nested submenu}
            if (p.IsSubmenu) then begin
                if (p.IsInSubmenu)  then begin 
                    {nested submenu}
                    {may not be possible}
                end else begin
                    {show me, if not showing}
                    if p <> p.fParent.fSubShowing then begin
                        self.ShowSubmenuDelayed(p);
                    end;
                end;
            end;
            if (not p.IsInSubmenu) then begin
                if (fLastSubmenu <> nil) and (fLastSubMenu <> p) then begin
                    timHideSubmenu.Enabled := true;
                end;
            end else begin
                if IsChild(p, fLastSubmenu) then begin
                    timHideSubmenu.Enabled := false;
                end else begin
                    if (p <> fLastSubmenu) then begin
                        self.ShowSubmenuDelayed(p);
                    end;
                end;
            end;

            if not p.isGap then begin
                SetHover(p);
            end else begin
                SetHover(p);
            end;
        end;
    end;
    { perform hit test }
    lastpos := FLastPosition;
    FLastPosition := pipNone;
    if (p <> nil) then begin
        FLastPosition := pipCaption;
        if (x - p.fLeft) < LEFT_SPACE then begin
            if p.ItemType in TClickableLeftIcons then begin
                FLastPosition := pipLeftIcon;
            end;
        end else if (p.fRight - x) < RIGHT_SPACE then begin
            if p.fHasRightIcon then begin
                FLastPosition := pipRightIcon;
            end;
        end;
    end;
    if (FLastPosition = pipNone) then begin
        timHideSubmenu.Enabled := true;
    end;
    if (lastPos <> FLastPosition) then begin
        if (fhover<>nil) then begin
            {update to show or remove focus rectangle}
            DrawItem(fhover);
        end;
        fToolTipNew.Hide;
        timHover.enabled := false;
        timHover.Enabled := true;
    end;
    if FLastPosition=pipLeftIcon then begin
        //ClipMenuShow
        if (fhover <> nil) then begin
            //DrawItem(fhover);
            timLeftIconHover.Enabled := false;
            timLeftIconHover.Enabled := true;
        end;
    end else begin
        timLeftIconHover.Enabled := false;
        self.ClipMenuHide;
    end;

end;


{Drag-and-Drop routines}
procedure TACPopup.DragMouseDown(p: TACPopupItem);
begin
    if (p <> nil) and (fDragMenuItem = nil) then begin
        if not (p.ItemType in [IT_NONE, IT_LINE]) then begin
            fDragMenuItem := p;
            if p <> nil then begin
            	fDragX := mouse.cursorpos.x; fDragY := mouse.cursorpos.y;
            end;
        end;
    end;
end;
function TACPopup.DragMouseMove(p, last: TACPopupItem) : boolean;
var temp : TACPopupItem;
begin
	result := false;
    if (p <> nil) and fDragStarted then  begin
    	temp := fDropMenuItem;
        fDropMenuItem := p;
        //last := fDropMenuItem;
        if (p <> last) then begin
            DrawItem(p);
            if IsLegalDrop(p) then begin


                screen.Cursor := crDrag;
                //appendlog('legaldrop');
            end else begin
                screen.Cursor := crNoDrop;
                appendlog('no legaldrop');
            end;

            //if last <> nil then DrawItem(last);
            if temp <> nil then DrawItem(temp);
        end else begin
            if p = fDragMenuItem then begin
            	screen.Cursor := crNoDrop;
            end;
            if temp <> nil then DrawItem(temp);
            DrawItem(p);

        end;

        result := true;
    end else if fDragStarted  then begin
       	screen.Cursor := crNoDrop;
        temp := fDropMenuItem;
        fDropMenuItem := p;
		if temp <> nil then DrawItem(temp);
    end;
end;
function TACPopup.DragMouseUp(p: TACPopupItem) : boolean;
var m1, m2 : TMenuOrderType;
begin
	result := false;
    screen.Cursor := crDefault;


    if fDragStarted and (p<>nil) then begin
        fDragStarted := false;
        if IsLegalDrop(p) then begin
            if fDragMenuItem.IsExpandable then begin
                m1 := TMenuOrderType(fdragmenuitem.integerdata);
            end else begin
                case fdragmenuitem.ItemType  of
                IT_MENU:	m1 := MOT_MENU;
                IT_PERMANENT: m1 := MOT_PERMANENT;
                IT_CLIPBOARD:	m1 := MOT_CURRENT;
                //IT_LAST: 	m1 := MOT_LAST;
                IT_SEARCH: 	m1 := MOT_SEARCH;
                IT_FULL: 	m1 := MOT_FULL;
                end;
            end;
            if p.IsExpandable then begin
            	m2 := TMenuOrderType(p.integerdata);
            end else begin
                case p.ItemType  of
                IT_MENU:	m2 := MOT_MENU;
                IT_PERMANENT: m2 := MOT_PERMANENT;
                IT_CLIPBOARD: m2 := MOT_CURRENT;
                //IT_LAST: 	m2 := MOT_LAST;
                IT_SEARCH: 	m2 := MOT_SEARCH;
                IT_FULL: 	m2 := MOT_FULL;
                end;
            end;
            if (m1 = MOT_MENU) then begin
                ClipQueue.Move(fDragMenuItem.IntegerData,p.IntegerData)
            end else if (m1 = MOT_PERMANENT) then begin
            	if fDragMenuItem.fPermanentGroupID = -1 then begin
                	FrmPermanent.MoveClip(fDragMenuItem.fPermanentID, p.fPermanentID);
                end;
            end else begin
        		FrmConfig.MoveItem(m1, m2);
            end;
            self.AutoPopulate;
            self.CalcTotalSize;
            self.Invalidate;
        end else begin
        	DrawItem(p);
        end;
        fDragMenuItem := nil;
        fdragx := 0; fdragy := 0;

    	result := true;
    end else begin
    	fDragStarted := false;
        fDragMenuItem := nil;
        fDropMenuItem := nil;
    end;
end;


{Timed Hide/Show Events}
procedure TACPopup.ShowSubmenuDelayed(p: TACPopupItem);
begin
    fQueuedSubmenu := p;
    timShowSubmenu.Enabled := true;
end;
procedure TACPopup.CancelShowSubmenu;
begin
    fQueuedSubmenu := nil;
    timShowSubmenu.Enabled := false;
end;
procedure TACPopup.timShowSubmenuTimer(Sender: TObject);
begin
    timShowSubmenu.Enabled  := false;

    if fQueuedSubmenu <> nil then begin
        fLastSubmenu := fQueuedSubmenu;
        fQueuedSubmenu.SetShowingSubMenu;
        if not fKeyActivated then
            fShowExpandedShortcuts := false;
        self.AssignSubmenuShortcuts(fQueuedSubmenu);
        fQueuedSubmenu := nil;



        self.CalcTotalSize;
        self.Invalidate;
    end;
end;
procedure TACPopup.timHideSubmenuTimer(Sender: TObject);
    procedure ClearSubmenus(pc : TACPopupItem);
    begin
        if pc = nil then EXIT;

        repeat
             pc.fSubShowing := nil;
             pc := pc.fParent;
        until pc = nil;
    end;
var P : TACPopupItem;
begin
    timHideSubmenu.Enabled := false;
    p := fItems;
    if p.fSubShowing = nil then begin

    end else begin
        while p.fSubShowing <> nil do begin
            p := p.fSubShowing;
        end;
        fLastSubmenu := p;
        if fLastSubmenu <> nil then begin
            ClearSubmenus(fLastSubmenu);
            fLastSubmenu := nil;
            self.CalcTotalSize;
            self.Invalidate;
        end;
    end;
end;
procedure TACPopup.timHoverTimer(Sender: TObject);
var pt : TPoint;
    ci : TClipItem;
    r : TRect;
    s : string;
    hotkeyname : string;
begin
    timHover.Enabled := false;

    if fhover = nil then EXIT;
    if FLastPosition = pipNone then EXIT;
	if fDragStarted then EXIT;

    r := fhover.BoundRect;
    pt := point(r.Right - trunc((r.Right-r.Left) / 3.5), r.bottom + 2);
    
    if fhover.IsExpandable and FrmConfig.cbAutoExpand.checked then begin
    	if not fhover.expanded then begin
        	self.HoverClick;
        end;
        timHover.Enabled := false;
    	EXIT;
    end;

    case fhover.ItemType  of
        IT_PERMANENT: begin
            case FLastPosition  of
            pipCaption: begin
                if fHover.fPermanentGroupID <> -1 then begin
                    FrmPermanent.PermFolderPush;
                    FrmPermanent.SetPermanentGroup(fHover.fPermanentGroupID);
                    hotkeyname := FrmPermanent.GetItemHotkeyName(fHover.fPermanentID);
                    FrmPermanent.PermFolderPop;
                end else begin
                    hotkeyname := FrmPermanent.GetItemHotkeyName(fHover.fPermanentID);
                end;


                if FrmPermanent.IsComplexItem(fhover.fPermanentText) then begin
                    ci := FrmPermanent.GetComplexItem(fhover.fPermanentText,fHover.fPermanentGroupID);
                    fToolTipNew.ShowTooltip(ci, pt, GetClipHeader(ci), hotkeyname);
                    myfree(ci);
                end else begin
                    fToolTipNew.ShowTooltip(fhover.fPermanentText,pt,GetClipHeader(nil),hotkeyname);
                end;
            end;
            end;
        end;
        IT_MENU,IT_PINNED,
        IT_PERMANENT_GROUP,
        IT_TEMP,
        //IT_SYS_SWITCH,
        IT_CLIPBOARD,
        IT_SELECTED,
        //IT_LAST,
        IT_CLICKED, IT_FULL: begin
            case FLastPosition of
            pipRightIcon: begin
                pt := point(fhover.Right+2,fhover.ftop);
                pt := self.ClientToScreen(pt);
                fToolTipNew.HideHeader;
                fToolTipNew.SmallFontOnce := true;
                fToolTipNew.ShowTooltip('[TAB or Click] Paste as ' + fhover.Clip.GetFormatName(false)  ,pt);
            end;
            pipCaption: begin
                if fhover.Hint <> '' then begin
                	fToolTipNew.HideHeader;
                    fTooltipNew.ShowTooltip(fHover.hint,pt);
                end else if fhover.fClip <> nil then begin
                	s := '';
                	if (fhover.Clip <> nil) and (fhover.clip.GetFormatType = FT_PICTURE) then begin
                    	s := fhover.Clip.GetFormatName;
                    end;



                    fToolTipNew.ShowTooltip(fhover.fclip, pt, GetClipHeader(fhover.fclip),'',
                        (fhover.ItemType in [IT_MENU, IT_TEMP]) and (fhover.fclip.CData.CreationDateUsed));
                end else begin

                end;
            end;
            pipLeftIcon: begin
                if fhover.ItemType = IT_FULL then begin
                	fToolTipNew.HideHeader;
                    fToolTipNew.SmallFontOnce := true;
                    fToolTipNew.ShowTooltip('Edit Menu Order', pt);
                end;
            end;
            //pipNone: fToolTip.ShowTooltip('None', pt,false,'');
            end;
        end else begin
            if fhover.Hint <> '' then begin
                fTooltipNew.HideHeader;
                fToolTipNew.SmallFontOnce := true;
                fTooltipNew.ShowTooltip(fHover.hint,pt);
            end;
        end;
    end;
end;
procedure TACPopup.timIgnoreMouseMoveTimer(Sender: TObject);
begin
	fIgnoreMouseMove := false;
    timIgnoreMouseMove.Enabled := false;
end;
procedure TACPopup.timLeftIconHoverTimer(Sender: TObject);
begin
    timLeftIconHover.Enabled := false;

    if FLastPosition = pipLeftIcon then
    	self.ClipMenuShow;
end;


{Click Event Processing}
procedure TACPopup.HandleMouseClick(p: TACPopupItem);
begin
	if lastHitTest = HTCAPTION then EXIT;

    self.fKeyActivated := false;
	{handle Icon clicks, pass of Caption cliks}
    {handle right clicks}
    if (p <> nil) and (p = fMouseDownOn) then begin
    	if p.Disabled  then EXIT;
        
        if (ssLeft in fDownShift) or(ssMiddle in fDownShift) then begin
            if (ssMiddle in fDownShift) then begin
               p.MiddleClicked := true;
            end;
			case FLastPosition of
				pipLeftIcon: begin
                	FrmMainPopup.SkipFocusReturnOnce;
                    self.hide;
                    case fhover.itemtype of
                        IT_PERMANENT:
                            FrmPermanent.ShowAndSelect(fhover.fPermanentGroupID, fhover.fPermanentID);
                        IT_FULL: begin
                            self.Hide;
                            FrmConfig.pcPanels.ActivePage := FrmConfig.tsMenuOrder;
                            frmConfig.Show;
                            UnitMisc.ForceForeground(frmconfig.Handle);
                            FrmMainPopup.SkipFocusReturnOnce;
                        end;
                        IT_LINE, IT_NONE: begin
                            // do nothing, Lines have not icons
                        end
                    else begin
                        if frmConfig.cbEditClipWindow.checked then begin
                            FrmEditItem.SetText(fhover.clip.GetAsPlaintext, nil);
                            FrmMainPopup.ShowPreviewEditForm;
                        end else begin
                            frmEditTextExternal.EditClip(fhover.clip);
                        end;
                    end;
                    end;
				end;
                pipRightIcon: begin
                    ACPopupClicks.ClipTypeClick(fHover);
                    self.Hide;
                end;
                pipCaption: begin
                	self.HandleCaptionClick(p);
                end;
                pipNone: ;
            end;
        end else if (ssRight in fDownShift) then begin
            self.CancelShowSubmenu;
            timHover.Enabled := false;
            timLeftIconHover.Enabled := false;
            if (p.Style = psNormal) and (p.ItemType in TUsesClipMenu) then begin
                self.ClipMenuShow(true);
            end;
        end;
    end;
end;
procedure TACPopup.HandleCaptionClick(p: TACPopupItem);
var wasChecked, wasExpanded, wasGrouped : boolean;
	i, j : integer;
    p2 : TACPopupItem;

    it : TItemType;
begin
    {handle special controls ourselves, pass off everything else to the click handler}



    p.fCtrlPressed := KeyboardQuery.IsPressed(VK_CONTROL);
    p.fShiftPressed := KeyboardQuery.IsPressed(VK_SHIFT);
    p.fAltpressed := KeyboardQuery.IsPressed(VK_MENU);

    case p.Style  of
    psLine: ;
    psBreak: ;
    psSubmenu: begin
        p.SetShowingSubMenu;

        fExpandedKeystrokes.Clear;
        fShowExpandedType := it;
        ShowExpandedShortcuts := self.fKeyActivated;
        self.AssignSubmenuShortcuts(p);

        self.CalcTotalSize;
        self.Invalidate;
    end;
    psExpandable: begin
    	timHover.Enabled := false;
        it := p.ItemType;
        fAutoExpandOnce := IT_NONE;
        fShowExpandedShortcuts := false;

        // Clip Sets special case
        // - the Current Group signifies the expanded set
        if (it = IT_CLIPSET_HEADER) then begin
            if not p.expanded then begin
                // find the collapsed group and trigger a refresh
                for i := 0 to self.fGroup.Count-1 do begin
                    p2 := TACPopupItem(self.fGroup.Items[i]);
                    if p2 = p then
                        Self.fCurrentGroup := i;
                end;
                self.AutoPopulate;
                self.CalcTotalSize;
                self.Invalidate;
            end;

            EXIT;
        end;


        unitMisc.AppendLog('Normal Expand');
        wasExpanded := p.Expanded;

        fShowExpandedType := it;

        if p.fDisabled then EXIT;

        if (FrmConfig.cbAutoCollapse.checked ) and (not p.Expanded)
             then begin
//        	simulate a show and an expand event - a known 'working' method for expanding
			self.AutoPopulate;
            p2 := nil;
            for i := 0 to fitems.fSubMenu.count - 1 do begin
                p2 := fitems.fSubMenu[i];
                if p2.ItemType = it then begin
                    BREAK;
                end;
            end;
            //self.CalcTotalSize;
            //self.Invalidate;
            if p2<>nil then begin
            	p2.Expanded := true;
                ShowExpandedShortcuts := self.fKeyActivated;

                self.CalcTotalSize;
                self.Invalidate;
                self.SetHover(p2);
                p := p2;
            end;

        end else begin
        	p.Expanded := not p.Expanded;
            ShowExpandedShortcuts := self.fKeyActivated and p.Expanded;
            self.CalcTotalSize;
            self.invalidate;
        end;

        fExpandedKeystrokes.Clear;
        self.AssignSubmenuShortcuts(p);
        fAutoExpandOnce := IT_NONE;
    end;
    psCheckable: begin
    	wasChecked := p.Checked;
        p.Checked := not p.Checked;
        DrawItem(p);
        DrawCheckgroup(p);
        if p.ItemType = IT_FORMMODE then begin
            self.FormMode := p.Checked;
            self.Invalidate; 
        end else begin
            if assigned(p.OnClick) then begin
                p.OnClick(p);
            end;
            if p.ItemType = IT_SWITCH then begin
            	if (waschecked = true) and (p.checked) then begin
                end else begin
                    self.AutoPopulate;
                    self.CalcTotalSize;
                    self.Invalidate;
                end;
            end;
        end;
        DrawItem(p);
    end;
    psNormal: begin
        case FLastPosition of
        pipLeftIcon: begin
            if p.ItemType = IT_FULL then begin
            	self.Hide;
                FrmConfig.pcPanels.ActivePage := FrmConfig.tsMenuOrder;
                frmConfig.Show;
                UnitMisc.ForceForeground(frmconfig.Handle);
                FrmMainPopup.SkipFocusReturnOnce;
            end;
        end;
        pipRightIcon: ;
        pipNone: ;
        pipCaption: begin
            self.fCaptionPoint.X := self.left + p.Left;
            self.fCaptionPoint.Y := self.top + p.Top;



            if p.IsInExpandedMenu and p.fCollapseParentOnClick then begin
                if assigned(p.onclick) then begin
                    p.OnClick(p);
                    p.fParent.Expanded := false;
                    self.AutoPopulate;
                    self.CalcTotalSize;
                    self.Invalidate;
                end;
            end else begin
                case p.ItemType  of
                IT_NEXTFIELD:begin
                    ForceForeground(FrmMainPopup.TargetData.ForegroundWindow);
                    Paste.SendTAB;
                    mysleep(30);

                    {if FormMode then begin
                        self.ShowPopup(fPoint);
            			self.FormMode := true;
                    end;}
                end;
                IT_FULL: begin
                    if FLastPosition=pipCaption then begin
                        fFullMode := not fFullMode;
                        if fFullMode then begin
                            p.Caption := 'Configured Mode';
                        end else begin
                            p.Caption := 'Full Mode'
                        end;
                        self.AutoPopulate;
                        self.CalcTotalSize;
                        self.Invalidate;
                    end else begin
                    end;
                end;
                IT_CANCEL: begin
                    self.Hide;
                end;
                else
                    HandleOnClick(p);
                end;
            end;
        end;

        end;
    end;
    end;
end;
procedure TACPopup.HandleOnClick(p: TACPopupItem);
begin
    if p.ItemType = IT_MENU then begin
        if p.Clip <> nil then
            p.Clip.CData.Clicked := true;
    end;

    if ACPopupClicks.PinClick(p) then begin
        p.OnClick(p);
        self.AutoPopulate;
        self.CalcTotalSize;
        self.Invalidate;
        EXIT;
    end;
	if ACPopupClicks.StayOpenOnClick(p) then begin
    	self.FormMode := true;

    	ACPopupClicks.StayOpenClick(p);

        self.FormMode := false;
        if FrmMainPopup.GetNeedsFocus then
            ForceForeground(self.Handle);
        self.AutoPopulate;
        self.CalcTotalSize;
        self.Invalidate;
        EXIT;
    end;
	if FormMode then begin
    	if p.ItemType in TPastesOnClickj then begin
            p.OnClick(p);
            {move to cursor?}
            self.ShowPopup(fPoint);
            self.FormMode := true;
        end;
    	EXIT;
    end;
    case p.ItemType  of
    IT_MENU,IT_PINNED,IT_PERMANENT, IT_CLIPBOARD,IT_TEMP,IT_SYSTEM,IT_SEARCH,
    IT_CLICKED, IT_PASTE_METHOD, IT_SELECTED{,IT_LAST} : begin
        case FLastPosition  of
        pipLeftIcon: ;
        pipRightIcon: ;
        pipCaption: begin
            if assigned(p.fOnClick) then begin
                self.Hide;
                ACPopupClicks.SetMenuPoint(self.fCaptionPoint);
                p.OnClick(p);
            end;
        end;
        pipNone: ;
        end;
    end;
    IT_NONE: begin {TODO thinkg about this}
        if assigned(p.fOnClick) then begin
            self.Hide;
            p.OnClick(p);
        end;
    end;
    IT_SWITCH: begin
        if assigned(p.fOnClick) then begin
            p.OnClick(p);
            if p.IsInExpandedMenu then
            	p.fParent.Expanded := false;

            self.AutoPopulate;
            self.CalcTotalSize;
            self.Invalidate;
        end;
    end;
    end;
end;

{Highlight routines}

procedure TACPopup.DetectAutoScroll(top,bottom : integer);
var
	pt : TPoint;
    m : TMonitor;
begin
	if fIsTooTall then begin
    	if top <> -1 then begin
           	pt.y := top;
            pt.x := 0;
            pt := self.ClientToScreen(pt);

            m := screen.MonitorFromWindow(self.handle);
            if (m <> nil) then begin
                if pt.y < m.top then begin
                	fIgnoreMouseMove := true;
                    timIgnoreMouseMove.Enabled := true;
                    self.Top := m.Top;
                    EXIT;
                end;
            end;
        end;

        if bottom <> -1 then begin
            pt.Y := bottom;
            pt.x := 0;
            pt := self.ClientToScreen(pt);


            m := screen.MonitorFromWindow(self.handle);
            if (m <> nil) then begin
                pt.Y := pt.y + m.top;
                if pt.y > m.BoundsRect.Bottom  then begin
                	fIgnoreMouseMove := true;
                    timIgnoreMouseMove.Enabled := true;
                    self.Top := m.top - (self.Height - m.Height);
                end;
            end;
        end;
    end;
end;

procedure TACPopup.SetHover(value: TACPopupItem);
var last : TACPopupItem;
begin
	if fShowing = false then EXIT;
    {hotkey pasting will call this method, even if the PopupItem hide the popup}
    
    last := fHover;
    fHover := value;
    if (last <> fHover) and (fHover <> nil) and (last <> nil) then begin
        timHover.Enabled := false;
        timHover.Enabled := true;
    end;

    fToolTipNew.Hide;
    if FLastPosition <> pipLeftIcon then begin
        self.ClipMenuHide;
    end;

    if (value <> nil) then begin
        // Detect when keystrokes need to shift up a level
        if (fKeyActivated) and
            (last<>nil) and (last.IsSubmenu) then begin
            if fHover.IsInExpandedMenu and not fhover.IsSubmenu and last.IsSubmenu then begin
                fShowExpandedType := fHover.ItemType;
                fShowExpandedShortcuts := True;
                Self.AssignSubmenuShortcuts(fhover.fparent);
                self.Invalidate;
            end;
        end;

        if last<>nil then DrawItem(Last);
        if fhover <> nil then DrawItem(fHover);

        if fhover.IsSubmenu then begin
            if (last<>nil) and (last.IsSubmenu) then begin
                self.CancelShowSubmenu;
                self.timHideSubmenuTimer(self);
            end;

            self.ShowSubmenuDelayed(fHover);
        end else if fhover.IsInSubmenu then begin

        end else begin
            self.CancelShowSubmenu;
            self.timHideSubmenuTimer(self);
        end;
    end;
end;
procedure TACPopup.HoverClick;
begin
    if Hover <> nil then begin
        FLastPosition := pipCaption;
        HandleCaptionClick(Hover);
    end;
end;
procedure TACPopup.HoverDown;
var i : integer;
    daddy, p : TACPopupItem;
begin
    if fItems.fSubMenu.Count = 0 then begin
        unitMisc.AppendLog('HoverDown: Empty Popup');
        EXIT;
    end;

    // init or select next item, no matter what it is
    if fHover = nil then begin
        i := 0;
        p := fItems.fSubMenu[i];
        daddy := fItems;
    end else begin
        daddy := fHover.VisibleParent;

        i := fHover.Index;
        i := (i + 1) mod daddy.fSubMenu.Count;
        p := daddy[i];
    end;

    while (p.Style in [psLine, psBreak]) do begin
        i := (i + 1) mod daddy.fSubMenu.Count;
        p := daddy.fSubMenu[i];
    end;


    FLastPosition := pipCaption;
    SetHover(p);
    self.DetectAutoScroll(p.top,p.bottom);
end;
procedure TACPopup.HoverLeft;
var p : TACPopupItem;
    i : integer;
    procedure CheckForExpandable(p : TACPopupItem);
    begin
        if (p.IsInExpandedMenu) then begin
            ShowExpandedShortcuts := self.fKeyActivated;
            fShowExpandedType := p.ItemType;
        end;
    end;
begin

    if fhover = nil then EXIT;
    {close submenu, if in submenu}
    if fHover.IsInSubmenu then begin
        ShowExpandedShortcuts := False;

        SetHover(fHover.fparent);
        self.CancelShowSubmenu;
        timHideSubmenuTimer(self);

        CheckForExpandable(fHover);
        if fHover.fParent.IsExpandable then
            self.AssignSubmenuShortcuts(fhover.fparent);
        self.Invalidate;
    end else if fHover.IsSubmenu and (fitems.fSubShowing <> nil) and (fhover.fSubMenu.Count <> 0)  then begin
        // close submenu only
        ShowExpandedShortcuts := False;
        CheckForExpandable(fHover);

        SetHover(fhover);
        self.CancelShowSubmenu;
        timHideSubmenuTimer(self);

        Self.AssignSubmenuShortcuts(fhover.fparent);
        self.Invalidate;

    end else begin
        if fhover.ColumnIndex = 0  then EXIT;

        p := getItemAt(fhover.fLeft - 5, (fhover.fTop+ fhover.fBottom)div 2);
        if p = nil then begin
            p := hover;
            i := p.Index;
            while (p.ColumnIndex = hover.ColumnIndex) or p.IsBreak do begin
            	dec(i);
                p := p.visibleparent[i];
            end;
        end;

        i := p.Index;
        while p.isline and (i > 0)do begin
            dec(i);
            p := p.VisibleParent[i];
        end;
        if p <> nil then begin
        	FLastPosition := pipCaption;
            SetHover(p);
        end;
    end;
end;
procedure TACPopup.HoverRight;
var i : integer;
    p : TACPopupItem;
begin
    if fhover = nil then EXIT;

    if fhover.IsSubmenu then begin
        if (fHover.fSubMenu.Count > 0) then begin
            fhover.SetShowingSubMenu;
            fLastSubmenu := fHover;
            self.AssignSubmenuShortcuts(fhover);
            self.CalcTotalSize;
            self.Invalidate;
            SetHover( fhover.fSubMenu[0] );
        end;
    end else begin
        p := getItemAt(fhover.right + 5, (fhover.fTop+ fhover.fBottom)div 2);
        if p = nil then begin
            if fitems.fColWidth.Count > 1 then begin
                p := fitems.fSubMenu[fitems.fsubmenu.count-1];
            end else begin
                EXIT;
            end;
        end;
        
        i := p.Index;
        while p.isline and (i > 0)do begin
            inc(i);
            p := p.VisibleParent[i];
        end;
        if p <> nil then begin
        	FLastPosition := pipCaption;
            SetHover(p);
        end;
    end;

end;
procedure TACPopup.HoverUp;
var i : integer;

    daddy, p : TACPopupItem;

begin

    // init or select next item, no matter what it is
    if fHover = nil then begin
        i := 0;
        daddy  := fItems;
        p := fItems.fSubMenu[daddy.fSubMenu.Count - 1];
    end else begin
        daddy := fHover.VisibleParent;
        i := fHover.index;
        dec(i);
        if i < 0 then i := daddy.fSubMenu.Count - 1;
        p := daddy[i];
    end;

    while (p.Style in [psLine, psBreak]) do begin
        dec(i);
        if i < 0 then i := daddy.fSubMenu.Count - 1;
        p := daddy.fSubMenu[i];
    end;



    FLastPosition := pipCaption;
    SetHover(p);
    self.DetectAutoScroll(p.top,p.bottom);
end;

{Keyboard Eventds}
procedure TACPopup.KeyDown(var Key: Word; Shift: TShiftState);

begin
    inherited;
    fLastKeyDown := key;

    if self.ClipMenuVisible then begin
        if key = VK_ESCAPE then begin

        end else begin
//            self.ClipMenuReportKey(key);
        end;
        key := 0;
    end else begin
        // allow for auto-repeat using the Down message
        fKeyActivated := true;
//        ShowExpandedShortcuts := false;
        case key  of
        VK_DOWN: self.HoverDown;
        VK_UP: self.HoverUp;
        VK_RIGHT: self.HoverRight;
        VK_LEFT: self.HoverLeft;
        else
            fKeyActivated := false;
        end;
    end;
end;
procedure TACPopup.KeyPress(var Key: Char);
var i, j : integer;
    p, p2 : TACPopupItem;

    it : TItemType;
begin
    inherited;
    if self.fIgnoreKeypresses then EXIT;

    if self.ClipMenuVisible then begin
        Self.ClipMenuReportKey(key);
        key := #0;
        EXIT;
    end;

    if (self.fShowExpandedShortcuts) then begin
        i := fExpandedKeystrokes.IndexOf(string(key));
        if i <> -1 then begin
            self.fKeyActivated := True;
            p := TACPopupItem(fExpandedKeystrokes.Objects[i]);

            self.FLastPosition := pipCaption;
            if Assigned(p.OnClick) then
                while KeyboardQuery.IsPressed(VK_SHIFT) do begin
                    Application.ProcessMessages;
                    Sleep(50);
                end;

            if p.IsSubmenu then begin
                SetHover(p);
                self.HandleCaptionClick(p);
            end else begin
                self.HandleCaptionClick(p);
            end;
            EXIT;
        end;
    end;

    windows.EnterCriticalSection(crit);

    i := fKeystrokes.IndexOf(string(key));
    if i <> -1 then begin
        self.fKeyActivated := true;
        p := TACPopupItem(fKeystrokes.Objects[i]);

        self.FLastPosition := pipCaption;
        if p.ItemType = IT_PERMANENT then begin
            if AnsiStartsText('[keys]', p.fPermanentText) then begin
				KeyboardQuery.WaitUntilRelease( VkKeyScan(key) );
            end;
        end;

        it := IT_NONE;
        if p.IsExpandable or p.IsSubmenu  then
        	it := p.ItemType;

    	if FALSE and fAltKeyDown then begin
            ACPopupClicks.ClipTypeClick(p);
            self.hide;
        end else begin
            {highlight expandable items}
            if it = IT_POPUP then begin
                SetHover(p);
                self.CancelShowSubmenu;
                Self.HandleCaptionClick(p);
            end else begin
                if Assigned(p.OnClick) then
                    while KeyboardQuery.IsPressed(VK_SHIFT) do begin
                        Application.ProcessMessages;
                        Sleep(50);
                    end;
                self.HandleCaptionClick(p);
                for j := 0 to fitems.fSubMenu.count - 1 do begin
                    p2 := fitems.fSubMenu[j];
                    if p2.IsExpandable and (p2.ItemType = it) then begin
                        SetHover(p2);
                        BREAK;
                    end;
                end;
            end;
        end;
    end;

    windows.LeaveCriticalSection(crit);
end;
procedure TACPopup.KeyUp(var Key: Word; Shift: TShiftState);
var ch : char;
begin
    inherited;
    if key = VK_ESCAPE then begin
        if self.ClipMenuVisible then begin
            ClipMenuReportKey(key);
        end else begin
            self.Hide;
        end;
        EXIT;
    end;
    if fhover = nil then EXIT;

//    if self.fim.Visible then begin
//        Self.ClipMenuReportKey(Char(key));
//        EXIT;
//    end;


    // Make sure Alt+Key was pressed on the popup
    // Detect when Alt+Key was pressed, flag it, and send a "Key" message to the popup
    fAltKeyDown := ssAlt in Shift;
    if FALSE and fAltKeyDown and (fLastKeyDown = Key) then begin
    	ch := chr(key);
        Paste.SendALT_Release;
        self.KeyPress(ch);
    end;

    case key of
    VK_RETURN, VK_SPACE: begin
        self.fKeyActivated := true;
        self.HoverClick;
    end;
    VK_DELETE: begin
        {Delete selected}
        case fhover.ItemType  of
            IT_MENU: begin
                ClipQueue.DeleteItem(ClipQueue.IndexOf(fhover.clip));
                self.DeleteClipCallback(nil);
            end;
            IT_TEMP: begin
                TempClipQueue.DeleteItem(TempClipQueue.IndexOf(fhover.clip));
                 self.DeleteClipCallback(nil);
            end;
            IT_CLIPBOARD: begin
                frmClipboardManager.ClearClipboard;
                self.DeleteClipCallback(nil);
            end;
            IT_CLICKED: begin
                ClickedClipQueue.DeleteItem2(ClickedClipQueue.IndexOf(fhover.clip), true);
                self.DeleteClipCallback(nil);
            end;
        end;
    end;
    VK_APPS: begin
        {right click menu}
        self.ClipMenuShow(false);
    end;
    VK_TAB: begin
        {clip type icon click}
        if (fhover.fItemType in TContainsAClip) then begin
            ACPopupClicks.ClipTypeClick(fHover);
            self.Hide;
        end;
    end;
    end;
end;

{Control Events}
procedure TACPopup.paint;
var c : Cardinal;
begin
    c := WIndows.GetTickCount;
	inherited;
    self.FullRedraw;
    AppendLog('paint ms=' + IntToStr(Windows.GetTickCount - c));
end;
procedure TACPopup.resize;
begin
    //inherited;
    {not handling gets rid of flicker}
    AppendLog('resize');
end;
procedure TACPopup.Hide;
begin
	if fShowing = false then begin
    	AppendLog('acpopup.hide called twice');
        EXIT;
    end;
    fShowing := false;

    //inherited;
    AppendLog('acpopup hide start');
    Windows.ShowWindow(self.Handle, SW_HIDE);

    timHover.Enabled := false;
    timShowSubmenu.Enabled := false;
    timHideSubmenu.Enabled := false;
    timLeftIconHover.Enabled := false;
    if assigned(fTooltipNew) then fToolTipNew.Hide;
    self.ClipMenuHide;

    fFullMode := false;

    self.visible := false;
    FirstPaint := false;
    screen.Cursor := crDefault;
    AppendLog('acpopup hide end');
end;
procedure TACPopup.WMActivate(var msg: tMessage);
begin
    {ignore all the inactive messages that appear until we've had a chance
    to paint and be visible to the user}
	if Msg.WParam = WA_INACTIVE then begin
    	if self.FirstPaint then begin
        	if self.FormMode = false then begin
        		self.hide;
            end;
        end;
    end;
end;
procedure TACPopup.WMGetDlgCode(var Msg: TMessage);
begin
	{Tell Windows we want arrow keys}
    inherited;
    Msg.Result:= Msg.Result or DLGC_WANTARROWS;
end;
procedure TACPopup.WMKeyDown(var Message: TWMKeyDown);
begin
	{TODO - what does this do again?}
    if message.CharCode in [VK_UP, VK_DOWN, VK_LEFT, VK_RIGHT] then begin
    end else begin
        inherited;
    end;
end;
procedure TACPopup.WMNCHitTest(var Message: TWMNCHitTest);
var pt : tpoint;
    i : integer;
const THRESHOLD = 5;
begin
	inherited;

    message.result := HTCLIENT;
    if not self.fDragStarted then begin

        pt := point(message.XPos, message.YPos);
        pt := self.ScreenToClient(pt);
        i := fMainRect.Right;
        if (pt.x >= i-THRESHOLD) and (pt.X <= i) then begin
            i := fMainRect.Bottom;
            if (pt.Y >= i-THRESHOLD) and (pt.Y <= i) then begin
            	message.Result := HTCAPTION;
            end;
        end;
    end;

end;
procedure TACPopup.WMSetCursor(var M: TWMSetCursor);
begin
	if m.CursorWnd = self.Handle then begin
        if m.HitTest = HTCAPTION then begin
            screen.Cursor := crSizeAll;
            m.Result := 1;
        end else begin
            screen.Cursor := crArrow;
            m.result := 1;
        end;
    end;
end;
procedure TACPopup.WMMove(var Msg : TMessage);
begin
   fpoint.X := self.Left;
   fpoint.Y := self.Top;
end;

{ TACPopupItemCollection }
function TACPopupItemCollection.Add: TACPopupItem;
begin
    result := TACPopupItem.Create;
    result.Visible := true;

    self.add(result);
end;
function TACPopupItemCollection.GetItem(Index: Integer): TACPopupItem;
begin
    result := TACPopupItem(inherited GetItem(Index));
end;
procedure TACPopupItemCollection.SetItem(Index: Integer; const Value: TACPopupItem);
begin
    inherited SetItem(Index, Value);
end;


{ TACPopupItem }
constructor TACPopupItem.Create();
begin
    inherited;

    fSubMenu := TACPopupItemCollection.Create();
    fCheckgroup := TObjectList.Create;
    fCheckgroup.OwnsObjects := false;
    fColWidth := TIntList.Create;
    fStyle := psNormal;

    fShowCollapseIcon := True;
end;
destructor TACPopupItem.Destroy;
var i : integer;
begin
    fColWidth.Free;
    fCheckgroup.Free;
    if (fStyle = psExpandable) and (fexpanded = false) then begin
        fSubMenu.Free;
    end;

    inherited;
end;

function TACPopupItem.GetBoundsRect: TRect;
begin
    result := fBoundsRect;
end;

function TACPopupItem.GetChecked: boolean;
begin
    result := fChecked;
end;
function TACPopupItem.GetColumnWidth: integer;
begin
    result := fParent.fColWidth[fMyColumn];
end;
function TACPopupItem.GetExpanded: boolean;
begin
    result := fExpanded;
end;

function TACPopupItem.GetIndex: integer;
begin
    {The index refers to the visible parent, not the owner}
    result := GetVisibleParent.fSubMenu.IndexOf(self);
end;

function TACPopupItem.GetOwnerIndex : integer;
begin
    result := fParent.fSubMenu.IndexOf(self);
end;

function TACPopupItem.GetIsBreak: boolean;
begin
    result := fStyle = psBreak;
end;
function TACPopupItem.GetIsExpandable: boolean;
begin
    result := fStyle = psExpandable;
end;
function TACPopupItem.GetIsInExpandedMenu: boolean;
begin
    result := fParent.IsExpandable;
end;
function TACPopupItem.GetIsInSubMenu: boolean;
begin
    result := (fparent.fParent <> nil) and not fparent.IsExpandable;
end;
function TACPopupItem.GetIsLine: boolean;
begin
    result := fStyle = psLine;
end;
function TACPopupItem.GetIsSubmenu: boolean;
begin
    result := fStyle = psSubmenu;
end;
function TACPopupItem.GetIsGap : boolean;
begin
    result := fStyle = psGap;
end;

function TACPopupItem.GetItem(Index : integer): TACPopupItem;
begin
    result := fSubMenu[index];
end;

function TACPopupItem.GetVisibleParent: TACPopupItem;
begin
    {In an expanded menu, the visible parent will be twice removed}
    result := fParent;
    if self.IsInExpandedMenu  then
        result := fParent.fParent;
end;


procedure TACPopupItem.CheckGrouped;
begin
    fParent.fCheckgroup.Add(self);
end;
function TACPopupItem.IsCheckGrouped : boolean;
begin
    result := fparent.fCheckgroup.IndexOf(self) <> -1;
end;
function TACPopupItem.Add: TACPopupItem;
begin
    result := self.fSubMenu.Add;
    result.fParent := self;

    if self.IsSubmenu then begin
        self.Disabled := false;
    end;
end;


function TACPopupItem.AddBreak: TACPopupITem;
begin
    result := self.add;
    result.IsBreak := true;
end;
function TACPopupItem.AddLine: TACPopupITem;
begin
    result := self.Add;
    result.IsLine := true;
end;
function TACPopupItem.CalcRight: integer;
begin
    if self.fParent.IsExpandable  then begin
        result := self.fLeft + self.fParent.fparent.fColWidth[self.fMyColumn] - RIGHT_BORDER;
    end else begin
        result := self.fLeft + self.fParent.fColWidth[self.fMyColumn] - RIGHT_BORDER;
    end;
    self.fRight := result;
end;
procedure TACPopupItem.SetBoundsRect(parenttop, parentleft: integer);
begin
    with fBoundsRect do begin
        top := ftop + parenttop;
        bottom := fbottom + parenttop;
        left := fleft + parentleft;
        right := self.right + parentleft;
    end;
end;

procedure TACPopupItem.SetCaption(value: string);
begin
    fCaption := value;
    if (length(fCaption) > 200) then begin
        fCaption := leftstr(fCaption, 200) + '...';
    end;
    // tabs make the windows too damn long
    // the #10 will make a funky box in the tab's place
    while pos(#9#9, fCaption) > 0 do begin
        fCaption := stringreplace(fCaption, #9#9, #9, [rfreplaceall]);
    end;
    fCaption := stringreplace(fCaption, #9,chr($B0), [rfReplaceAll]); {unicode middle dot}

    {escape the special hotkey character}
    fCaption := stringreplace(fCaption, '&', '&&', [rfReplaceAll]);
end;
procedure TACPopupItem.SetCaptionFromClip(value: TClipItem);
begin
    if value.CData.displaytext <> '' then begin
        fCaption := value.CData.displaytext;
        EXIT;
    end;
    case value.GetFormatType of
    FT_UNICODE, FT_RICHTEXT, FT_HTML:
        begin
            value.BuildMenuCaption;
            fCaption := value.CData.displaytext;
        end;
    FT_PICTURE:
        begin
            fCaption := value.GetFormatName()
        end;
    else
        begin
            SetCaption(value.GetAsPlaintext);
        end;
    end;
    value.CData.displaytext := fCaption;
end;

procedure TACPopupItem.SetChecked(value: boolean);
var i : integer;
    p : TACPopupItem;
begin
    if (fParent.fCheckgroup.IndexOf(self) <> -1) then begin
        if (self.fChecked = false) and value then begin
            for i := 0 to fParent.fCheckgroup.Count - 1 do begin
                p := TACPopupItem(fParent.fCheckgroup[i]);
                p.fChecked := false;
            end;
            self.fChecked := true;
        end;
    end else begin
        self.fChecked := value;
    end;

    self.fStyle := psCheckable;
end;
procedure TACPopupItem.SetExpanded(value: boolean);
var i : integer;
begin
    fExpanded := value;

    if value then begin
        for i := fSubMenu.count - 1 downto 0 do begin
            fparent.fSubMenu.Insert(self.index+1,fsubmenu[i]);
        end;
    end else begin
        for i := fSubMenu.count - 1 downto 0 do begin
            fparent.fSubMenu.Extract(fSubmenu[i]); // Don't Free the object
        end;
    end;
end;

procedure TACPopupItem.SetIsBreak(value: boolean);
begin
    if value then
        fStyle := psBreak;
end;
procedure TACPopupItem.SetIsExpandable(value: boolean);
begin
    if value then begin
        fStyle := psExpandable;
        fLeftIconOnlyOnHover := true;
    end;
end;
procedure TACPopupItem.SetIsLine(value: boolean);
begin
    if value then
        fStyle := psLine;
end;
procedure TACPopupItem.SetIsSubmenu(value: boolean);
begin
    if value then
        fStyle := psSubmenu;
end;
procedure TACPopupItem.SetIsGap(value : boolean);
begin
    if value then
        fStyle := psGap;
end;
procedure TACPopupItem.SetItem(Index: integer; value: TACPopupItem);
begin
    fSubMenu[index] := value;
end;

procedure TACPopupItem.SetLeftBitmap(value: TBitmap);
begin
    self.fHasLeftIcon := true;
    self.fLeftBitmap := value;
end;
procedure TACPopupItem.SetLeftIcon(value: hicon);
begin
    self.fHasLeftIcon := true;
    fIcon := value;
end;
procedure TACPopupItem.SetRightIcon(value: TBitmap);
begin
    self.fRightBitmap := value;
    self.fHasRightIcon := true;
end;
procedure TACPopupItem.SetShowingSubMenu;
begin
    if self.IsInExpandedMenu  then begin
        fParent.fParent.fSubShowing := self;
    end else begin
        fParent.fSubShowing := self;
    end;

end;

constructor TCustomTransparentControl.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    ControlStyle := ControlStyle + [csOpaque];
    Brush.Style := bsClear; {required for XP to make backgroun transparent}
end;
procedure TCustomTransparentControl.CreateParams(var Params: TCreateParams);
begin
    inherited CreateParams(Params);
    {Params.ExStyle := Params.ExStyle or WS_EX_TRANSPARENT;}
    {can't use WS_EX_TRANSPARENT in XP}

    if CheckWin32Version(5, 1) then
       params.WindowClass.Style := params.WindowClass.style or CS_DROPSHADOW;
end;
procedure TCustomTransparentControl.WMNCHitTest(var Message: TWMNCHitTest);
begin
    if not FInterceptMouse then begin
        Message.Result := HTTRANSPARENT
    end else begin
        inherited;
    end;
end;
procedure TCustomTransparentControl.Invalidate;

begin
    AppendLog('invalidate');
	inherited Invalidate;
end;

initialization
begin

end;
end.
