unit UnitFrmPermanentNew;
{
    Purpose:
        This unit stores/reads/edits the permanent items.
        The form is not a dummy form.


    Updates:
        New: Permanent Items Groups sorted by name
        --------------
        Fix: Up button did not range check
        --------------
        Changes for complex item as permanent items,
        show them when it's shown or hovered

        Folder list not cleared when "LoadPermanent" was called
        ----------------
        Made scalable
        New Keystroke support

        -----------------
        Updates for autotamically switching PIGs for a specific program

        -------------------
        Complete Rewrite. Old form was too poorly designed
        to save.


}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, INIFiles, ComCtrls {for THahsedStringList},
  UnitClipQueue, UnitFrmHotkey, Menus, Buttons, ClipBrd, Contnrs,
  ImgList, UnitFrmTooltipNew, UnitFrameClipDisplay;

const DEFAULT_FOLDER = 'Default';
      ADDNEW_FOLDER = '<add new>';
      PERM0_FILE = 'perm0.ini';
      PERM1_FILE = 'perm1.ini';
      PERM0_UNICODEFILE = 'perm0_unicode.ini';
      PERM1_UNICODEFILE = 'perm1_unicode.ini';

      HOTKEY_FILE = 'hotkey.ini';
      EXEPIG_FILE = 'exepigs.ini';

type TDataFileName = (DFN_PERM0, DFN_PERM1, DFN_HOTKEYS);




type
  TFrmPermanent = class(TForm)
    {$REGION 'DEFAULT'}
    pcPermanent: TPageControl;
    TabSheet2: TTabSheet;
    lvAutoSwitch: TListView;
    Label5: TLabel;
    pmGetAs: TPopupMenu;
    PlaintText1: TMenuItem;
    DIBPicture1: TMenuItem;
    CopiedFiles1: TMenuItem;
    RichTExt1: TMenuItem;
    tsEditGroups: TTabSheet;
    tvGroups: TTreeView;
    btnNewItemx: TButton;
    btnEditItemx: TButton;
    TabSheet3: TTabSheet;
    pnlText: TPanel;
    pnlHotkey: TPanel;
    lblHotkey: TLabel;
    Label11: TLabel;
    Shape2: TShape;
    pmTree: TPopupMenu;
    miDelete: TMenuItem;
    miMoveto: TMenuItem;
    btnFile: TSpeedButton;
    ilPerm: TImageList;
    Memo1: TMemo;
    Label1: TLabel;
    Memo2: TMemo;
    btnNewItem: TSpeedButton;
    btnEditItem: TSpeedButton;
    timCloseTooltip: TTimer;
    fClipDisplay: TFrameClipDisplay;
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure btnAddPIGClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure tsEditGroupsShow(Sender: TObject);
    procedure tvGroupsExpanding(Sender: TObject; Node: TTreeNode;
      var AllowExpansion: Boolean);
    procedure tvGroupsDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure tvGroupsEndDrag(Sender, Target: TObject; X, Y: Integer);
    procedure tvGroupsCollapsing(Sender: TObject; Node: TTreeNode;
      var AllowCollapse: Boolean);
    procedure btnNewItemxClick(Sender: TObject);
    procedure miDeleteClick(Sender: TObject);
    procedure btnEditItemxClick(Sender: TObject);
    procedure miMovetoItemClick(Sender: TObject);
    procedure pmTreePopup(Sender: TObject);
    procedure btnFileClick(Sender: TObject);
    procedure tvGroupsCustomDraw(Sender: TCustomTreeView; const ARect: TRect;
      var DefaultDraw: Boolean);
    procedure tvGroupsCustomDrawItem(Sender: TCustomTreeView; Node: TTreeNode;
      State: TCustomDrawState; var DefaultDraw: Boolean);
    procedure tvGroupsChange(Sender: TObject; Node: TTreeNode);
    procedure timCloseTooltipTimer(Sender: TObject);

    //procedure Button1Click(Sender: TObject);
    {$ENDREGION}
  private
  { Private declarations }
        SelectedIndex : integer;
        SelectedGroup : string;
        AppPath : string;
        CurrentPermPath : string;
        FolderList : TStringList;
        ItemNameList : TStringList;
        ItemDataList : TStringList;
        ItemHotKeyList : TStringList;
        EXEPigList : THashedStringList;
        EXEPasteList : THashedStringList;
        //Tooltip : TTooltipWindow;

        HotkeyObjects : TStringList;
        HotkeyObjects2 : TStringList;
        PushpopPath : string;
        EditModeOn : boolean;
        hkd : THotkeyData;

        //treenodedrag : TTreeNode;
        accepteddrag : boolean;
        // util functions

        LastSavedGroup : string;
        LastSavedItemIndex : integer;
        ignorecollapse : boolean;
        isrightclick : boolean;

        hotkeyTooltip : TFrmTooltipNew;

        function GetOldDataFilename (i: integer): string;
        function GetDataFilename (i: TDataFileName; ForRead : boolean = true; overridefolder : string = ''): string;

        function ToHKString(hkd : THotkeyData) : string;
        function FromString(s : string) : THotKeyData;
        procedure UpdateFolderList;
        procedure LoadPermanentItemsGroup;
        procedure SavePermanentItemsGroup;
        function IsPermanentPathFolder(folder : string) : boolean;
        procedure RefreshFormData;
        //
//        procedure SetEditMode(value : boolean);
//        procedure ShowSelectedItem;
//        procedure RefreshCurrentItem;

//        procedure PopupItemClick(sender: TObject);

        function GetNextComplexIndex : integer;
        function ExtractComplexIndex(s : string) : integer;

        procedure RefreshHotkeys;
        procedure RefreshPermanentGroups(selectCurrentGroup : boolean);
        procedure SelectTreeItem(group : string; index : integer);
        procedure ShowSelectedTreeItem;
        procedure HandleItemSelect;

        function MakeClipReference(i : integer) : string;
  protected
        procedure CreateParams(var Params: TCreateParams); override;
  public
  { Public declarations }
        procedure ShowForm;
        procedure SetPermanentPath( path : string );
        procedure SetPermanentGroup(i : integer);
        function GetPermanentPath : string;
        function GetPermanentPathFull : string;
        // to enumerate permanent item names
        function PermFoldersGetCount : cardinal;
        function PermFoldersGetItem(index : cardinal) : string;
        function IsComplexItem(s : string) : boolean;
        function GetComplexItem(s : string) : TClipItem; overload;
        function GetComplexItem(s : string; group : integer) : TClipItem; overload;


        procedure SaveNewItem(
            group : string;
            name, text : string;
            hotkey : THotKeyData;
            usekeys : boolean;
            ci : TClipItem = nil
            );
        procedure SaveItem(
            group : string; index : integer;
            name, text : string;
            hotkey : THotKeyData;
            usekeys : boolean;
            ci : TClipItem = nil
            );

        procedure PermFolderPush;
        procedure PermFolderPop;

        // to enumerate items in the current permanent folder
        function GetCount: integer;
        function GetItemName(i: integer): string;
        function GetItemText(i: integer): string; overload;
        function GetItemText(i: integer; group : integer): string; overload;
        function GetItemHotkeyName(i : integer) : string;
        function GetTextFrom(name: string): string;
        function GetTextIndex(name : string) : integer;

        // display form with a newly created item
        procedure ShowWithNewItem(item : string; name : string = ''; hotkey : string = ''); overload;
        procedure ShowWithNewComplexItem(ci : TClipItem; name : string = ''; hotkey : string = ''); overload;
        procedure ShowAndSelect(groupID, itemID : integer);


        procedure AutoSwitch(EXEName : string);
        procedure AssignPIG(EXEName : string);

        function GetHotkeyCount : integer;
        function GetHotkeyIdent(index : integer) : integer;
        function IsHotkeyIdent(ident : integer) : boolean;
        procedure ReportHotkey(index : integer);

        procedure MoveClip(oldi, newi : integer);
  end;

var
  FrmPermanent: TFrmPermanent;
const EMPTY_HOTKEY_NAME = 'none';
{////////////////////}
{//}implementation{//}
{////////////////////}

uses UnitMisc, StrUtils, UnitFrmClipboardManager, UnitFrmMove, UnitHotKey,
   UnitFrmMainPopup, UnitPopupGenerate, UnitTWideChar,
  UnitFrmPermanentEdit, UnitKeyboardQuery, System.Generics.Collections,
  System.Character;

{$R *.dfm}

const NEW_GROUP_TEXT = '[New Group ...]';

//
// public API
//
procedure TFrmPermanent.SetPermanentGroup(i: integer);
begin
    self.SetPermanentPath( self.PermFoldersGetItem(i) );
end;
procedure TfrmPermanent.SetPermanentPath( path : string );
begin
    if (self.EditModeOn) then begin
        ShowMessage('Cannot switch groups while editing a permanent item.');
        EXIT;
    end;

    if Self.IsPermanentPathFolder(path) then  begin
        CurrentPermPath := path;
        self.LoadPermanentItemsGroup;
    end else begin
        CurrentPermPath := DEFAULT_FOLDER;
        self.LoadPermanentItemsGroup;
    end;

    //lbItemNames.ItemIndex := -1;
end;
function TfrmPermanent.GetPermanentPath : string;
begin
    result := self.CurrentPermPath;
end;
function TFrmPermanent.GetPermanentPathFull: string;
begin
    result := IncludeTrailingPathDelimiter(
        IncludeTrailingPathDelimiter(self.AppPath) + self.GetPermanentPath
    );
end;


function TfrmPermanent.PermFoldersGetCount : cardinal;
begin
    result := FolderList.Count;
end;
function TfrmPermanent.PermFoldersGetItem(index : cardinal) : string;
begin
    result := FolderList[index];
end;

function TFrmPermanent.GetComplexItem(s: string; group: integer): TClipItem;
begin
    if group <> -1 then begin
        self.PermFolderPush;
        self.SetPermanentGroup(group); 
    end;

    result := GetComplexItem(s);

    if group <> -1 then begin
        self.PermFolderPop;
    end;
end;
function TfrmPermanent.GetCount: integer;
begin
    result := ItemNameList.Count;
end;
function TfrmPermanent.GetItemName(i: integer): string;
begin
    result := ItemNameList[i];
end;
function TFrmPermanent.GetItemText(i, group: integer): string;
begin
    if (group <> -1) then begin
        self.PermFolderPush;
        self.SetPermanentGroup(group);
    end;
    
    result := GetItemName(i);

    if (group <> -1) then begin
        self.permfolderpop;
    end;
end;
function TfrmPermanent.GetItemText(i: integer): string;
begin
    result := ItemDataList[i];
end;
function TFrmPermanent.GetItemHotkeyName(i: integer): string;
var hkd : THotkeyData;
begin
    hkd := FromString(ItemHotKeyList[i]);
    if hkd.name = EMPTY_HOTKEY_NAME then
        result := ''
    else
        result := hkd.name;
end;
function TFrmPermanent.GetTextIndex(name: string): integer;
var i: integer;
begin
    result := -1;
    for i := 0 to ItemNameList.count - 1 do begin
        if (lowercase(name) = lowercase(ItemNameList[i])) then begin
            result := i;
            BREAK;
        end;
    end;
end;
function TfrmPermanent.GetTextFrom(name: string): string;
begin
    result := ItemDataList[ GetTextIndex(name) ];
end;



procedure TFrmPermanent.ShowAndSelect(groupID, itemID: integer);
var s : string;
begin
    self.Show;

    Windows.SetForegroundWindow(self.Handle); 

    self.pcPermanent.ActivePageIndex := 0;

    s := self.CurrentPermPath;
    if groupID <> -1 then
        s := self.PermFoldersGetItem(groupID);
    self.SelectTreeItem(s, itemID);

end;
procedure TFrmPermanent.ShowForm;
begin
    self.Show;
end;
procedure TfrmPermanent.ShowWithNewItem(item : string; name : string = '';hotkey : string = '');
begin
    self.Show;
    self.pcPermanent.ActivePageIndex := 0;
    self.SelectTreeItem(self.CurrentPermPath, -1);
    FrmPermanentEdit.SetNewItem(item);
    self.btnNewItem.Click;
end;
procedure TFrmPermanent.ShowWithNewComplexItem(ci: TClipItem; name : string = '';hotkey : string = '');
var i : integer;
begin
	if ci.GetFormatType = FT_UNICODE then begin
		self.ShowWithNewItem(ci.GetAsPlaintext);
    end else begin
        i := self.GetNextComplexIndex;
        if (ci <> nil) then begin
            ci.SaveToFile(self.GetPermanentPathFull, i, CI_FILEMASK_ALL);
            ci.FinishedWithStream;
        end;

        self.ShowWithNewItem(MakeClipReference(i), name, hotkey);
	end;
end;

procedure TFrmPermanent.HandleItemSelect;
begin
    btnNewItem.Enabled := false;
    btnEditItem.Enabled := false;
    if tvgroups.Selected = nil then EXIT;
    if tvgroups.Selected.Text = NEW_GROUP_TEXT  then begin
        self.btnAddPIGClick(self);
        self.tsEditGroupsShow(self);
    end;

    if tvgroups.selected.Parent = nil then begin
        btnNewItem.Enabled := true;
    end else begin
        btnEditItem.Enabled := true
    end;

    self.ShowSelectedTreeItem;
end;

//
// Create / Destroy
//
procedure TFrmPermanent.CreateParams(var Params: TCreateParams);
begin
  inherited;
    // allow context taskbar context menu and
    // show me on the taskbar - independant of main form
    with Params do begin
        ExStyle := ExStyle or WS_EX_APPWINDOW;
        WndParent := GetDesktopwindow;
    end;
end;
procedure TFrmPermanent.FormCreate(Sender: TObject);
var name: string;
    lc : TListColumn;

    procedure EmptyFile(name : string);
    var t : textfile;
    begin
        assignfile(t,name);
        rewrite(t);
        closefile(t);
    end;
    procedure ConvertClipsToV2;
    var i,j : integer;
        ci : TClipItem;
        s : string;

        procedure ConvertClip(s : string; i : integer);
        var idx, j, k : integer;
            ci : TClipITem;
        begin
            ci := TClipItem.Create;
            ci.LoadFromFile(self.GetPermanentPathFull, ExtractComplexIndex(s) );

            // find where reference to this clip
            idx := -1;
            for j := 0 to ItemDataList.Count - 1 do begin
                if ItemDataList[j] = s then begin
                    idx := j;
                    BREAK;
                end;
            end;
            if (idx <> -1) then begin
                // save to the new format and update the reference
                k := self.GetNextComplexIndex;
                ci.SaveToFile(self.GetPermanentPathFull, k, CI_FILEMASK_ALL);
                ItemDataList[idx] := self.MakeClipReference(k);
            end;
            myfree(ci);
        end;
    begin
        for i := 0 to self.PermFoldersGetCount-1 do begin
            self.SetPermanentGroup(i);
            for j := 0 to self.GetCount - 1 do begin
                s := self.GetItemText(j);
                if IsComplexItem(s) then begin
                    if TCharacter.ToLower(s[4])='l' then
                        convertclip(s, j); // trigers the conversion
                end;
            end;
            self.SavePermanentItemsGroup;
        end;
    end;
begin
    UnitMisc.AppendLog('FrmPermanent - creating');

    self.Font.Size := 8;

    self.FolderList := TStringList.Create;
    self.ItemNameList := TStringList.Create;
    self.ItemDataList := TStringList.Create;
    self.ItemHotKeyList := TStringList.Create;
    self.EXEPigList := THashedStringList.Create;
    self.EXEPasteList := THashedStringList.Create;
    //self.Tooltip := TTooltipWindow.Create;
    self.HotkeyObjects := TStringList.Create;
    self.HotkeyObjects2 := TStringList.Create;

    self.CurrentPermPath := DEFAULT_FOLDER;
    self.AppPath := UnitMisc.GetAppPath;

    // fix
    if FileExists(self.AppPath + EXEPIG_FILE) then begin
        self.EXEPigList.LoadFromFile(self.AppPath + EXEPIG_FILE);
    end;

    //
    // make the new Default directory and import and
    // current permanent items
    //
    if not DirectoryExists( self.AppPath + DEFAULT_FOLDER) then begin
        mkdir(self.AppPath + DEFAULT_FOLDER);
        self.PermFolderPush;
        self.CurrentPermPath := DEFAULT_FOLDER;

        name := GetDataFilename(DFN_PERM0);
        EmptyFile(name);
        name := GetDataFilename(DFN_HOTKEYS);
        EmptyFile(name);
        name := GetDataFilename(dfn_perm1);
        EmptyFile(name);
        self.PermFolderPop;


        name := GetOldDataFilename(0);
        if FileExists(name) then
            copyfile(pchar(name), PChar(GetDataFileName(DFN_PERM0)), true);

        name := GetOldDataFilename(1);
        if fileExists(name) then
            copyfile(pchar(name), PChar(GetDataFilename(DFN_PERM1)), true);
    end;

    self.UpdateFolderList;
    self.RefreshHotkeys;

    //
    // init the Auto Switch window
    //
    lc := lvAutoSwitch.Columns.Add;
    lc.Caption := 'Program Name';
    lc.Width := 100;

    lc := lvAutoSwitch.Columns.Add;
    lc.Caption := 'Group';
    lc.Width := 100;

    //lblClipType.Caption := '';

    pcPermanent.ActivePageIndex := 0;
    pcPermanent.Align := alClient;

    //self.SetEditMode(false);
    self.PushpopPath := '';

    fClipDisplay.ShowClip('');
    fClipDisplay.PicStretched := true;

    tvGroups.DoubleBuffered := true;

    ConvertClipsToV2;
end;
procedure TFrmPermanent.FormDestroy(Sender: TObject);
begin
    self.EXEPigList.SaveToFile(self.AppPath + EXEPIG_FILE );
    MyFree(self.EXEPasteList);
    MyFree(self.ItemNameList);
    MyFree(self.ItemDataList);
    MyFree(self.ItemHotkeyList);
    MyFree(self.FolderList);
    //MyFree(self.Tooltip);
//    MyFree(self.Clipboard);
end;
//
// Util Functions
//

function TfrmPermanent.GetOldDataFilename(i: integer): string;
begin
    result := self.AppPath + 'perm' + IntToStr(i) + '.ini';
end;
function TfrmPermanent.GetDataFilename(i: TDataFileName; ForRead : boolean = true; overridefolder : string = ''): string;
var
    s : string;
    folder : string;
begin
    folder := CurrentPermPath;
    if overridefolder <> '' then begin
        folder := overridefolder;
    end;

    case i of
    DFN_PERM0 : result := IncludeTrailingPathDelimiter(self.AppPath + folder) + PERM0_FILE;
    DFN_PERM1 : result := IncludeTrailingPathDelimiter(self.AppPath + folder) + PERM1_FILE;
    DFN_HOTKEYS : result := IncludeTrailingPathDelimiter(self.AppPath + folder) + HOTKEY_FILE;
    end;


    case i of
    DFN_PERM0: begin
        s := IncludeTrailingPathDelimiter(self.AppPath + folder) + PERM0_UNICODEFILE;
        if (forread and fileexists(s)) or (not forread) then begin
            result := s;
        end;
    end;
    DFN_PERM1: begin
        s := IncludeTrailingPathDelimiter(self.AppPath + folder) + PERM1_UNICODEFILE;
        if (forread and fileexists(s)) or (not forread) then begin
            result := s;
        end;
    end;
    end;
end;

procedure TfrmPermanent.LoadPermanentItemsGroup;
var name, itemText, s : string;
    lineCount : cardinal;
    i, cnt : integer;
    tf : textfile;

    sl : TStringList;
begin

    // use default when none given
    // ensure the default exists
    if CurrentPermPath = '' then begin
        CurrentPermPath := DEFAULT_FOLDER;
    end;
    if not DirectoryExists(self.AppPath + CurrentPermPath) then begin
        if CurrentPermPath = DEFAULT_FOLDER then
            ForceDirectories(self.AppPath + CurrentPermPath);
    end;



    //
    // load permanent items
    //
    ItemNameList.Clear;
    name := GetDataFilename(DFN_PERM0);
    if FileExists(name) then begin
        ItemNameList.LoadFromFile(name);
    end;

    ItemHotKeyList.Clear;
    name := GetDataFilename(DFN_HOTKEYS);
    if FileExists(name) then begin
    	if ContainsText(name,'_unicode') then begin
			ItemHotKeyList.LoadFromFile(name, tencoding.Unicode);
        end else begin
        	ItemHotKeyList.LoadFromFile(name);
        end;
    end;
    while ItemHotKeyList.Count < ItemNameList.Count do begin
        ItemHotKeyList.Add('');
    end;

    //
    // abort reading and show message on error
    // always close the file
    //
    ItemDataList.Clear;
    name := GetDataFilename(DFN_PERM1);
    if FileExists(name) then begin
    	if ContainsText(name, '_unicode.') then begin
            sl := TStringList.Create;
            sl.LoadFromFile(name, TEncoding.Unicode);

            i := 0;
            while i < sl.Count do begin
                s := sl[i];
                inc(i);
                cnt := StrToInt(s);

                itemtext := '';
                while cnt <> 0 do begin
                    s := sl[i];
                    inc(i);
                	if itemtext = '' then begin
                        itemtext := s;
                    end else begin
                    	itemtext := itemtext + #13#10 + s;
                    end;
                    dec(cnt);
                end;
                ItemDataList.Add(itemText);
            end;
        end else begin

            AssignFile(tf, name);
            Reset(tf);

            try
                while not eof(tf) do begin
                    try
                        Readln(tf, s);
                        itemText := '';
                        lineCount := StrToInt(s);

                        for i := 0 to lineCount - 1 do begin
                            Readln(tf, s);
                            if (itemText = '') then begin
                                itemText := s;
                            end else begin
                                itemText := itemText + chr(13) + chr(10) + s;
                            end;
                        end;

                        ItemDataList.Add(itemText);
                    except
                         on E: Exception do begin
                            ShowMessage('The "Permanent Item" file for group ' + CurrentPermPath + ' is corrupted - ' + name + #13#10#13#10 +
                                        'Error Message: ' + E.Message);
                            break;
                         end;
                    end;
                end;
            finally
                CloseFile(tf);
            end;
        end;
    end;
end;
procedure TfrmPermanent.SavePermanentItemsGroup;
var name: string;
    s : string;
    cnt : cardinal;
    sl : TStringList;
    i,j: longint;

begin
    if (CurrentPermPath = '') then
        EXIT;

    if not DirectoryExists(IncludeTrailingPathDelimiter(AppPath) + CurrentPermPath) then begin
        mkdir(IncludeTrailingPathDelimiter(AppPath) + CurrentPermPath);
    end;
    //
    // save items
    //
    name := GetDataFilename(DFN_PERM0, false);
    ItemNameList.SaveToFile(name, TEncoding.Unicode);

    name := GetDataFilename(DFN_HOTKEYS);
    ItemHotKeyList.SaveToFile(name);

    name := GetDataFilename(dfn_perm1, false);
    //AssignFile(tf, name);
    //Rewrite(tf);
    sl := TStringList.Create;

    for i := 0 to ItemDataList.Count - 1 do begin
        s := ItemDataList[i];

        cnt := 1;
        for j := 1 to length(s) - 1 do begin
// Unix LF causing issues
//            if (s[j] = #13) and (s[j+1]= #10) then inc(cnt);
            if (s[j+1]= #10) then inc(cnt);
        end;


        sl.Add(IntToStr(cnt));
        sl.Add(s);
        //writeln(tf, cnt);
        //writeln(tf, s);
    end;

    //CloseFile(tf);
    sl.SaveToFile(name, TEncoding.Unicode);
    sl.Free;
end;


procedure TfrmPermanent.UpdateFolderList;
var rec : TSearchRec;
    r : integer;
begin
    //
    // scan each subfolder and look for permanent item config files
    // this will generate a list of permanent item groups (using their
    // folder name)
    //
    FolderList.clear;
    FolderList.Sorted := true;
    r := FindFirst(AppPath + '*.*', faDirectory, rec);
    while (r = 0) do begin
        // is file a subfolder?
        if (rec.Attr and faDirectory) > 0 then begin
            if (rec.name <> '.') and (rec.name <> '..') then begin
                if IsPermanentPathFolder(rec.name) then begin
                    FolderList.Add(rec.name);
                end;
            end;
        end;

        r := FindNext(rec);
    end;
end;

function TFrmPermanent.IsPermanentPathFolder(folder : string) : boolean;
begin
    result := fileexists( GetDataFilename(DFN_PERM0, true, folder) );
end;

procedure TFrmPermanent.FormShow(Sender: TObject);
begin
    self.LoadPermanentItemsGroup;
    self.RefreshFormData;
end;
procedure TFrmPermanent.RefreshFormData;
var i : integer;
    li : TListItem;
begin
    // load the current item names
    //
    lvAutoSwitch.Items.clear;
    for i := 0 to EXEPigList.Count - 1 do begin
        li := lvAutoSwitch.items.Add;
        li.Caption := EXEPigList.Names[i];
        li.SubItems.Add(EXEPigList.Values[EXEPigList.Names[i]]);
    end;
end;
procedure TFrmPermanent.btnEditItemxClick(Sender: TObject);
var s : string;
    b : boolean;
    prevPerm : string;
begin
    FrmPermanentEdit.top := self.top;
    FrmPermanentEdit.left := self.left;

    prevPerm := self.CurrentPermPath;

    self.PermFolderPush;
    self.SetPermanentPath(self.SelectedGroup);
    s := ItemHotKeyList[self.selectedIndex];
    hkd := self.FromString(s);
    hkd.name := FrmHotKey.GetHumanReadable(
        hkd.alt,hkd.shft,hkd.ctrl,hkd.win,hkd.key, b
    );
    self.PermFolderPop;


    FrmPermanentEdit.showform(
        false, self.SelectedGroup, self.SelectedIndex, hkd, nil
    );
    if FrmPermanentEdit.ModalResult <> mrCancel then begin
        //self.CurrentPermPath := self.SelectedGroup;
        //self.tsEditGroupsShow(self);
        //self.ShowSelectedTreeItem;
        self.RefreshPermanentGroups(true);
        self.SelectTreeItem(self.LastSavedGroup, self.LastSavedItemIndex );
    end;
end;

procedure TFrmPermanent.btnFileClick(Sender: TObject);
begin
    UnitMisc.ShowPopupRight(btnFile, pmTree);
end;

procedure TFrmPermanent.SelectTreeItem(group: string; index: integer);
var i : integer;
    tn : TTreeNode;
begin

    for i  := 0 to tvgroups.Items.count - 1 do begin
        tn := tvgroups.Items[i];
        if tn.Parent = nil then begin
            if tn.Text = group then begin
                tvgroups.Selected := tn;
                tn.Expand(false);
                if index >= 0 then tvgroups.Selected := tn.item[index];

                self.ShowSelectedTreeItem;

                BREAK;
            end;
        end;
    end;
end;



//
// add / delete Permanent Item Group
//
procedure TFrmPermanent.btnAddPIGClick(Sender: TObject);
var newgroup : string;
begin
    if
    Dialogs.InputQuery('Add Group','Please enter a new group name', newgroup)
    then begin
        if not IsPermanentPathFolder(newgroup) then begin
            self.CurrentPermPath  := newgroup;
            self.ItemNameList.Clear;
            self.ItemDataList.clear;
            self.SavePermanentItemsGroup;
            self.UpdateFolderList;
            self.RefreshFormData;
        end else begin
            Dialogs.showmessage('Group name already exists');
        end;
    end;
end;
procedure TFrmPermanent.AutoSwitch(EXEName: string);
var path : string;
begin
    // given an EXE name, change to a PIG if it's associated with one
    // Associate the current PIG with the EXE if no association exist

    path := EXEPigList.Values[EXEName];
    if (path <> '') then begin
        self.SetPermanentPath(path);
    end else begin
        if EXEName <> '' then begin
            EXEPigList.Values[EXEName] := self.GetPermanentPath;
        end;
    end;
end;
procedure TFrmPermanent.AssignPIG(EXEName: string);
begin
    EXEPigList.Values[EXEName] := self.GetPermanentPath;
end;

procedure TFrmPermanent.PermFolderPop;
begin
    if self.PushpopPath <> self.CurrentPermPath then begin
        self.SetPermanentPath(self.PushpopPath);
    end;
    self.PushpopPath := '';
end;
procedure TFrmPermanent.PermFolderPush;
begin
    if self.PushpopPath <> '' then begin
        self.PushpopPath := '';
        raise Exception.Create('ERROR: Permanent Item folder restore failed');
    end;
    self.PushpopPath := self.CurrentPermPath;
end;

// Clip Routines
function TfrmPermanent.MakeClipReference(i : integer) : string;
begin
    result := '[FIV2=' +IntToStr(i)+ ']';
end;
function TFrmPermanent.GetNextComplexIndex: integer;
var i, j : integer;
//    sl : TStringList;
    sl : TList<integer>;
begin
//    sl := TStringList.Create;
    sl := TList<Integer>.create;
    for i := 0 to (ItemDataList.Count - 1) do begin
        if self.IsComplexItem(ItemDataList[i]) then begin
            j := self.ExtractComplexIndex(ItemDataList[i]);
            sl.Add(j);
//            sl.Add(IntToStr(j));
        end;
    end;

    result := 0;
    while sl.IndexOf(result) <> -1 do inc(result);

//    UnitMisc.MyFree(sl);
end;
function TFrmPermanent.ExtractComplexIndex(s: string): integer;
begin
    s := midstr(s,7,length(s)-7);
    result := StrToInt(s);
end;
function TFrmPermanent.GetComplexItem(s: string): TClipItem;
var i : integer;
    prefix, indexstr : string;
begin
    result := TClipItem.Create;

    prefix := uppercase(leftstr(s,6));

    indexstr := midstr(s,7,length(s)-7);
    i := StrToInt(indexstr);
    if (prefix='[FILE=') then begin
        // unused
    end else if (prefix='[FIV2=') then begin
        result.LoadFromFile(self.GetPermanentPathFull, i, CI_FILEMASK_ALL);
    end;
end;
function TFrmPermanent.IsComplexItem(s: string): boolean;
var s2 : string;
begin
    s2 := uppercase(leftstr(s,6));
    result :=  (s2='[FILE=') or (s2='[FIV2=');
end;


procedure TFrmPermanent.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
{    if btnCancel.Visible then begin
        btnCancel.Click;
    end;
}
    CanClose := true;
end;




//
// Hotkey Management / Enumeration
//
function TFrmPermanent.GetHotkeyCount: integer;
begin
    result := HotkeyObjects.Count;
end;
function TFrmPermanent.GetHotkeyIdent(index: integer): integer;
begin
    result := TACHotKey(hotkeyobjects.Objects[index]).Getid;
end;
procedure TFrmPermanent.RefreshHotkeys;
var i, j, k : integer;
    s : string;
    hk : TACHotKey;
    hkd : THotkeyData;
begin
    for i := 0 to HotkeyObjects2.count - 1 do begin
        hk := TACHotKey(
            HotkeyObjects2.Objects[i]
        );
        //hk.UnSet;
        MyFree(hk);
    end;
    HotkeyObjects2.Clear;
    HotkeyObjects.Clear;
    self.PermFolderPush;
    k := 0;
    for i := 0 to self.PermFoldersGetCount - 1 do begin
        self.SetPermanentPath(self.PermFoldersGetItem(i));

        for j := 0 to self.GetCount - 1 do begin
            s := ItemHotKeyList[j];
            if s <> '' then begin
                hkd := self.FromString(s);
                try
                    if (hkd.name <> EMPTY_HOTKEY_NAME) and
                        (hkd.alt or hkd.ctrl or hkd.shft or hkd.win) then begin

                        hk := TACHotKey.Create(
                            frmmainpopup.Handle,
                            'ArsClip.PermItem' + hkd.name,
                            hkd.ctrl, hkd.alt, hkd.shft, hkd.win, hkd.key
                        );
                        HotkeyObjects.Values[IntToStr(hk.GetID)] := IntToHex(i,8) + IntToHex(j,8);

                        HotkeyObjects2.AddObject(IntToStr(k), hk);
                        inc(k);

                        UnitMisc.AppendLog('perm hotkey: ' + hkd.name);
                    end;
                except
                end;
            end;
        end;
    end;

    self.PermFolderPop;
end;
function TFrmPermanent.ToHKString(hkd: THotkeyData): string;
begin
    result :=
        IntToStr(Integer(hkd.alt)) +
        IntToStr(Integer(hkd.ctrl)) +
        IntToStr(Integer(hkd.shft)) +
        IntToStr(Integer(hkd.win)) +
        IntToHex(hkd.key,2) +
        hkd.name;

end;
function TFrmPermanent.FromString(s: string): THotKeyData;
begin
    result.name := EMPTY_HOTKEY_NAME;
    if length(s) < 6 then begin
        EXIT;
    end;

    result.alt := Boolean(StrToInt(s[1]));
    result.ctrl  := Boolean(StrToInt(s[2]));
    result.shft  := Boolean(StrToInt(s[3]));
    result.win  := Boolean(StrToInt(s[4]));
    result.key := StrToInt('$' +  copy(s,5,2));
    delete(s,1,6);
    result.name := s;
end;
function TFrmPermanent.IsHotkeyIdent(ident: integer): boolean;
begin
    result := HotkeyObjects.Values[IntToStr(ident)] <> '';
end;
procedure TFrmPermanent.ReportHotkey(index: integer);
var s, sname, shotkey : string;
    i,j : integer;
begin
    try
        s := HotkeyObjects.Values[IntToStr(index)];
    except
    end;
    if (trim(s) <> '') then begin
        i := StrToInt('$' + copy(s,1,8));
        j := StrToInt('$' + copy(s,9,8));

        self.PermFolderPush;
        try
            self.SetPermanentPath( self.PermFoldersGetItem(i) );
            s := self.GetItemText(j);
            sname := self.GetItemName(j);
            shotkey := GetItemHotkeyName(j);
        finally
            self.PermFolderPop;
        end;


        if not assigned(hotkeyTooltip) then begin
            hotkeyTooltip := TFrmTooltipNew.Create(application);
        end;
        hotkeyTooltip.HideHeader;
        hotkeyTooltip.ShowTooltip(sname, mouse.CursorPos,shotkey);


        Application.ProcessMessages;
        FrmMainPopup.SendText(s);

        timCloseTooltip.enabled := true;
    end;
end;
procedure TFrmPermanent.timCloseTooltipTimer(Sender: TObject);
begin
    timCloseTooltip.enabled := false;
    hotkeyTooltip.hide;
    application.ProcessMessages;
end;

//===========================================================
//  new Edit form
//===========================================================

procedure TFrmPermanent.tsEditGroupsShow(Sender: TObject);
begin
    self.RefreshPermanentGroups(true);
end;
procedure TFrmPermanent.tvGroupsChange(Sender: TObject; Node: TTreeNode);
begin
    if self.visible and tvgroups.Visible then begin
        KeyboardQuery.WaitUntilRelease(leftButton);
        tvGroups.SetFocus;
        self.HandleItemSelect;
    end;
end;
procedure TFrmPermanent.tvGroupsExpanding(Sender: TObject; Node: TTreeNode; var AllowExpansion: Boolean);
begin
    //self.HandleItemSelect;
    KeyboardQuery.WaitUntilRelease(leftButton);
    AllowExpansion := not isrightclick;

    if (AllowExpansion) then begin
        self.ignorecollapse := true;
        tvGroups.FullCollapse;
        self.ignorecollapse := false;
    end;
end;

procedure TFrmPermanent.tvGroupsCollapsing(Sender: TObject; Node: TTreeNode; var AllowCollapse: Boolean);
begin
    //self.HandleItemSelect;

    AllowCollapse := false;
    if ignorecollapse  then begin
        KeyboardQuery.WaitUntilRelease(leftButton);
        AllowCollapse := true;
        EXIT;
    end;
    if isrightclick then begin
        EXIT;
    end;

    {AllowCollapse := false;
    if tvGroups.Selected <> nil then begin
        AllowCollapse := tvGroups.Selected.HasChildren
        and (tvGroups.selected <> node);
    end;}
    AllowCollapse := true;
end;
procedure TFrmPermanent.tvGroupsCustomDraw(Sender: TCustomTreeView; const ARect: TRect; var DefaultDraw: Boolean);
begin
    defaultdraw := true;
end;
procedure TFrmPermanent.tvGroupsCustomDrawItem(Sender: TCustomTreeView; Node: TTreeNode; State: TCustomDrawState; var DefaultDraw: Boolean);
begin
    defaultdraw := true;
    if (cdsFocused in state) or
        (cdsHot in state) or
        (cdsSelected in state)  then begin
        if (cdsHot in state) then begin
                sender.canvas.brush.Color := dimColor(ColorToRGB(clInactiveCaption),1.05);
            end else begin
                sender.canvas.brush.Color := dimColor(ColorToRGB(cl3DLight ));
            end;
        //sender.canvas.brush.Color := DimColor(ColorToRGB(clInactiveCaption),1.05);
        if (cdsHot in state)  then begin
                //sender.Canvas.font.color := ColorToRGB(clHighlighttext);
            end else begin
                sender.Canvas.font.color := ColorToRGB(clBtnText);
            end;
    end;
    if node.parent = nil then begin
        sender.Canvas.Brush.color := ColorToRGB(clBtnFace);
        if (cdsSelected in state) or
        (cdsHot in state) then begin
            if (cdsHot in state) then begin
                sender.canvas.brush.Color := dimColor(ColorToRGB(clInactiveCaption),1.05);
            end else begin
                sender.canvas.brush.Color := dimColor(ColorToRGB(cl3DLight ));
            end;
            sender.Canvas.font.color := ColorToRGB(clBtnText);
            {
            if (cdsHot in state)  then begin
                sender.Canvas.font.color := ColorToRGB(clHighlighttext);
            end else begin
                sender.Canvas.font.color := ColorToRGB(clBtnText);
            end;
            }
        end;
    end;
end;
procedure TFrmPermanent.pmTreePopup(Sender: TObject);
var i : integer;
    mi : TMenuItem;
    s : string;
    cur : string;
begin
   miMoveto.Enabled := false;
    if tvgroups.Selected.Parent = nil then begin
        midelete.Caption := '&Delete Group';
    end else begin
        miDelete.Caption := '&Delete Item';
        miMoveto.Enabled := true;
        miMoveto.Clear;
        cur := tvgroups.selected.parent.text;

        for i := 0 to FrmPermanent.PermFoldersGetCount - 1 do begin
            s := FrmPermanent.PermFoldersGetItem(i);
            if cur <> s then begin
                mi := pmTree.CreateMenuItem;
                mi.Caption := s;
                mi.OnClick := miMovetoItemClick;
                mi.AutoHotkeys := maManual;
                mimoveto.Add(mi);

            end;
        end;
    end;
    //tvgroups.Selected := tn;
end;
procedure TFrmPermanent.tvGroupsDragOver(Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
var tn : TTreeNode;
begin
    tn := tvGroups.GetNodeAt(x,y);
    Accept := tn <> nil;
    Accept := Accept and (tvgroups.Selected <> nil); //
    Accept := Accept and (tvgroups.selected.parent <> nil); // root nodes don't move
    Accept := Accept and (tn.Text <> NEW_GROUP_TEXT);

    accepteddrag := Accept;
    if accept then begin


    end;
end;
procedure TFrmPermanent.tvGroupsEndDrag(Sender, Target: TObject; X, Y: Integer);
var drag, drop : TTreeNode;
    dragi, dropi : integer;
    group, prevgroup : string;
    ci : TClipItem;
    name, text : string;
    hotkey : THotkeyData;
    usekeys : boolean;
begin
    if not self.accepteddrag  then EXIT;

    drop := tvGroups.GetNodeAt(x,y);
    drag := tvGroups.Selected;
    //showmessage('selected:' + tvGroups.Selected.Text + ' drop:' + drop.text);

    if (drop <> nil) then begin

        if (drop.Parent <> nil) then begin
            // move item within group

            dragi := drag.Index;
            dropi := drop.Index;
            group := drop.Parent.text;
            //showmessage('drag: ' + IntToStr(drag.Index) + ' drop:' + IntToStr(drop.Index));
            self.PermFolderPush;
            self.SetPermanentPath(group);
            ItemNameList.Move(dragi, dropi);
            ItemDataList.Move(dragi, dropi);
            ItemHotKeyList.Move(dragi, dropi);
            self.SavePermanentItemsGroup;
            self.PermFolderPop;

            //self.SetPermanentPath(group);
            self.RefreshPermanentGroups(false);
            self.SelectTreeItem(group, dropi);
        end else begin
            // move item to new group
            
            dragi := drag.Index;
            group := drop.Text;
            prevgroup := drag.Parent.Text;
            if group = prevgroup then EXIT;
            
            ci := nil;
            self.PermFolderPush;
            self.SetPermanentPath(drag.Parent.text);
                name := ItemNameList[dragi];
                text := ItemDataList[dragi];
                if self.IsComplexItem(text)  then
                    ci := self.GetComplexItem(text);
                hotkey := self.FromString(ItemHotKeyList[dragi]);
                usekeys := lowercase(leftstr(text,6)) = '[keys]';
            self.PermFolderPop;

            self.SaveNewItem(
                group,name,text,hotkey,usekeys,ci
            );

            self.PermFolderPush;
                self.SetPermanentPath(group);
                dropi := ItemNameList.Count - 1;
            self.PermFolderPop ;

            self.PermFolderPush;
                self.SetPermanentPath(prevgroup);
                ItemNameList.Delete(dragi);
                ItemDataList.Delete(dragi);
                ItemHotKeyList.Delete(dragi);
                self.SavePermanentItemsGroup;
            self.PermFolderPop;
            self.RefreshPermanentGroups(false);
            self.SelectTreeItem(group, dropi);
        end;
    end;


end;




procedure TFrmPermanent.ShowSelectedTreeItem;
var tn : TTreeNode;
    txt : string;
    ci : TClipItem;
  wc: TWideChar;
  r : Trect;
  hotkey : THotkeyData;
begin
    tn := tvGroups.selected;
    if tn <> nil then begin
        btnEditItem.Enabled := false;
        btnNewItem.enabled := false;
        lblhotkey.Caption := EMPTY_HOTKEY_NAME;
        pnltext.Visible := false;
        pnlhotkey.Visible := false;

        if tn.Parent <> nil then begin
            pnltext.Visible := true;
            pnlhotkey.Visible := true;

            self.PermFolderPush;
            self.SetPermanentPath(tn.Parent.text);
            self.SelectedGroup := tn.Parent.text;
            self.SelectedIndex := self.GetTextIndex(tn.text);

            hotkey := self.FromString(ItemHotKeyList[self.selectedindex]);
            lblHotkey.Caption := hotkey.name;
            txt := self.GetItemText(self.SelectedIndex);
            btnEditItem.Enabled := true;

            ci := nil;
            fClipDisplay.OverrideFont(self.font);
            if IsComplexItem(txt) then begin
                ci := self.GetComplexItem(txt);
                fClipDisplay.PicStretched := true;
                fClipDisplay.ShowClip(ci);
            end else begin
                fClipDisplay.Showclip(txt);
            end;
            MyFree(ci);

            self.PermFolderPop;
        end else begin
            btnNewItem.Enabled := true;
        end;
    end;

end;


procedure TFrmPermanent.SaveNewItem(group, name, text: string; hotkey: THotKeyData; usekeys: boolean; ci : TClipItem = nil);
begin
    if usekeys and (pos('[keys]', lowercase(text)) <> 0) then begin
        text := rightstr(text, length(text)- length('[keys]'));
    end;

    self.SaveItem(group,-1,name,text,hotkey,usekeys, ci);
end;
procedure TFrmPermanent.SaveItem(group: string; index: integer; name, text: string; hotkey: THotKeyData; usekeys: boolean; ci : TClipItem = nil);
var asnew : boolean;
    i : integer;
    hkd2 : THotkeyData;

begin
    self.PermFolderPush;
    self.SetPermanentPath(group);
    asnew := (index = -1);

    // create complex clip in cache, if needed
    if (ci <> nil) then begin
        i := self.GetNextComplexIndex;

        while FileExists(
            ci.GetFilenameV2(self.GetPermanentPathFull, i)
        ) do
            inc(i);

        ci.SaveToFile(self.GetPermanentPathFull, i, CI_FILEMASK_ALL);
        text := self.MakeClipReference(i);
    end;


    // create new or save name and text
    self.LastSavedItemIndex := index;
    self.LastSavedGroup := group;    
    if (asnew) then begin
        index := ItemNameList.count;
        self.LastSavedItemIndex := index;

        // Name must exist, name must be unique for a new item
        //
        if name = '' then name := 'Default';
        i := 1;
        if ItemNameList.IndexOf(name) <> -1 then begin
            while ItemNameList.IndexOf(name + IntToStr(i) ) <> -1 do begin
                inc(i);
            end;
            name := name + IntToStr(i);
        end;

        ItemNameList.Add(name);
        ItemDataList.Add(text);
        ItemHotKeyList.Add('');
    end else begin
        ItemNameList[index] := name;
        ItemDataList[index] := text;
    end;

    // handle keystroke flag for plain text items
    if usekeys and (ci = nil) then begin
        ItemDataList[index] := '[KEYS]' + text;
    end;



    hkd2 := self.FromString(ItemHotKeyList[index]);
    if ToHKString(hkd2) <> ToHKString(hotkey) then begin
        if (hotkey.name <> EMPTY_HOTKEY_NAME) then begin
            ItemHotKeyList[index] := self.ToHKString(hotkey);
        end else begin
            ItemHotKeyList[index] := '';
        end;
    end;


    self.SavePermanentItemsGroup;
    self.PermFolderPop;
    self.RefreshHotkeys;

end;

procedure TFrmPermanent.RefreshPermanentGroups(selectCurrentGroup: boolean);
var i, j : integer;
    tn, subtn : TTreeNode;
    ci : TClipItem;
begin
    tvGroups.Items.clear;
    //if (self.GetCount = 0) and (self.CurrentPermPath <> DEFAULT_FOLDER)  then EXIT;// can't do this

    self.PermFolderPush;
    for i := 0 to FolderList.Count - 1 do begin
        tn := tvGroups.items.Add(nil, folderlist[i]);
        tn.ImageIndex := 0;
        tn.SelectedIndex := tn.ImageIndex;

        self.SetPermanentPath(folderlist[i]);
        for j := 0 to self.GetCount  - 1 do begin
            subtn := tvGroups.Items.AddChild(tn, self.GetItemName(j));
            subtn.ImageIndex := 1;
            subtn.SelectedIndex := subtn.ImageIndex;

            if IsComplexItem(self.GetItemText(j))  then begin
                ci := self.GetComplexItem(self.GetItemText(j));
                if ci.GetFormat = CF_DIB then begin
                    subtn.ImageIndex := 3;
                    subtn.SelectedIndex := subtn.ImageIndex;
                end else if (ci.GetFormat = CF_UNICODETEXT) or (ci.getformat = GetCF_HTML) or (ci.GetFormat= GetCF_RICHTEXT)  then begin
                    subtn.ImageIndex := 2;
                    subtn.SelectedIndex := subtn.ImageIndex;
                end;
                MyFree(ci);

            end;
        end;


        if (lowercase(tn.Text) = lowercase(CurrentPermPath)) and
            selectCurrentGroup then begin
            tn.Selected := true;
            tn.Expand(true);
            tvGroups.Selected := tn;
        end else begin
            tn.Selected := false;
        end;
    end;
    self.PermFolderPop;


    tn := tvGroups.items.add(nil, NEW_GROUP_TEXT);
    tn.ImageIndex := 4;
    tn.SelectedIndex := tn.ImageIndex;

    btnNewItem.Enabled := false;
    btnEditItem.Enabled := false;

    //self.tvGroupsClick(self);
    self.HandleItemSelect;
    self.ShowSelectedTreeItem;
end;
procedure TFrmPermanent.btnNewItemxClick(Sender: TObject);
begin
    FrmPermanentEdit.top := self.top;
    FrmPermanentEdit.left := self.left;

    self.SelectedGroup := tvGroups.Selected.Text;


    FrmPermanentEdit.showform(true, self.SelectedGroup, -1);
    if FrmPermanentEdit.ModalResult <> mrCancel then begin
        self.RefreshPermanentGroups(false);
        self.SelectTreeItem(self.SelectedGroup, -1);
    end;
end;

procedure TFrmPermanent.miDeleteClick(Sender: TObject);
var g : string;
    sr : TSearchRec;
    i : integer;
begin
     if MessageDlg('Delete the selected item "' +tvGroups.selected.text+'"?', mtConfirmation,[mbYes, mbNo],0,mbNo) <> mrYes  then
        EXIT;

    if tvGroups.Selected.parent = nil then begin
        //
        // Delete Group
        //
        g := IncludeTrailingPathDelimiter(AppPath) + tvGroups.Selected.text;
        if DirectoryExists(g) then begin
            i := findfirst(g + '\*.*', faAnyFile,  sr);
            while i = 0 do begin
                if (sr.Attr and faDirectory) = 0 then
                    DeleteFile( g + '\' + sr.name );
                i := FindNext(sr);
            end;
            FindClose(sr);

            //deletefile( GetDataFilename(dfn_perm0) );
            //deleteFile( GetDataFilename(dfn_perm1) );
            RmDir(g + '\');

            self.UpdateFolderList;
            self.RefreshPermanentGroups(false); 
        end;
    end else begin
        self.PermFolderPush;
            g := tvGroups.Selected.parent.text;
            self.SetPermanentPath(g);
            i := tvgroups.Selected.Index;
            ItemNameList.Delete(i);
            ItemDataList.Delete(i);
            ItemHotKeyList.Delete(i);
            //lbItemNames.Items.Delete(i);
            self.SavePermanentItemsGroup;
        self.PermFolderPop;
        self.RefreshPermanentGroups(false);
        self.SelectTreeItem(g,-1);
    end;
end;
procedure TFrmPermanent.miMovetoItemClick(Sender: TObject);
var mi : TMenuITem;
    i : integer;
    g, newgroup, name, text : string;
    ci : TClipItem;
    hotkey : THotkeydata;
    usekeys : boolean;

begin
    ci := nil;
    mi := TMenuItem(sender);
    newgroup := mi.Caption;


    // gather data from selected item
    self.PermFolderPush;
        g := tvGroups.Selected.parent.text;
        self.SetPermanentPath(g);
        i := tvgroups.Selected.Index;
        name := ItemNameList[i];
        text := ItemDataList[i];
        if self.IsComplexItem(text)  then
            ci := self.GetComplexItem(text);
        hotkey := self.FromString(ItemHotKeyList[i]);
        usekeys := lowercase(leftstr(text,6)) = '[keys]';
    self.PermFolderPop;

    // save in new group
    self.SaveNewItem(
        newgroup,name,text,hotkey,usekeys,ci
    );

    // delete item
    self.PermFolderPush;
        self.SetPermanentPath(g);
        ItemNameList.Delete(i);
        ItemDataList.Delete(i);
        ItemHotKeyList.Delete(i);
        self.SavePermanentItemsGroup;
    self.PermFolderPop;
    
    // select new location
        self.SetPermanentPath(g);
        self.RefreshPermanentGroups(true);



end;
procedure TFrmPermanent.MoveClip(oldi, newi: integer);
begin
	ItemNameList.Move(oldi, newi);
    ItemDataList.Move(oldi, newi);
    ItemHotKeyList.Move(oldi, newi);
    self.SavePermanentItemsGroup;
end;

end.
