unit UnitPaste;
{
    Purpose:
        Move the Pasting logic into it's own class. The FrmMainPopup unit
        was getting a little bloated.
    NOTES;
        Use EmptyClipboard (or Clipboard.Clear) before pasting - this marks
        clips as "ours" so the ClipboardManager can ignore them

    Updates:
        Don't use Simple Text version when pasting CF_DIBs
        --------------------
        Pasting complex items will now also include a plain text (for programs
        that don't understand what, for example, CF_HTML is)
        --------------------
        New option to alter the clipboard (and trigger the Clipboard Manager)
        Win9x Mimic characters shift keys didn't work
        -----------
        Clear the clipboard before pasting
        This is the only place an item may be placed on the clipboard

        ------
        Another fix for SHIFT+Insert on Win2k
        --------------------
        - New support for pasting with keystrokes embedded in the text
        - Fix for SHIFT+Insert on Win2k
}
interface
uses UnitClipQueue, INIFiles, Classes;

type TPasteMethod = (
    PASTE_CTRL_V=0,
    PASTE_SHIFT_INS=1,
    PASTE_MIMIC=2,
    PASTE_CLIPBOARD=3,
    PASTE_DEFAULT=4,
    PASTE_DISABLED=5
);

type TPasteVariable = (
pv_NONE,
pv_date, pv_wait, pv_prompt, pv_popupitem, pv_clipbrd, pv_clipbrdcurrent, pv_clear, pv_clearpoup,
pv_windowconfig, pv_windowremoved, pv_windowhistory, pv_windowclipboard, pv_windowpermanent,
pv_windowsearch, pv_selectall, pv_selectleft, pv_selectright, pv_windowpastesel,
pv_clipboardonly, pv_left, pv_right, pv_up, pv_down, pv_backspace, pv_home, pv_end,
pv_tab, pv_delete, pv_insert, pv_space, pv_enter, pv_key, pv_cut, pv_copy, pv_paste, pv_run,
pv_clipboardfind, pv_disablemonitoring, pv_deletepopupclip, pv_trimclipboard,
pv_clipboardupper, pv_clipboardlower, pv_clipboardcapwords, pv_clipboardinverse
);


type TPaste = class(TObject)
    private
        {Configurations for pasting method}
        UseKeyboardMimic,
        UsePastingSI,
        UsePastingCV,
        UseClipboardOnly,

        {single time use overrides for pasting}
        PasteCVOnce,
        PasteSIOnce,
        ClipboardOnlyOnce,
        KeyboardMimicOnce : boolean;


        EXEPasteList : THashedStringList;

        VariableStart : TStringHash;
        fIsPasting, fIsPastingKeystrokes : Boolean;
        procedure ClearMethods;

    public
        constructor Create;
        destructor Destroy; override;
        procedure DefineVariableStart(starttext : string; vartype : TPasteVariable);
        function IsVariable(starttext : string) : boolean;
        function VariableType(starttext : string) : TPasteVariable;
        procedure SendText(s: string; ci  : TClipItem = nil; noPlaintext : boolean = false);
        procedure SendTextWithKeystrokes(maintext : string);
        procedure PlaceOnClipboardDontBypassClipboardManager(s : string);
        procedure SendSHIFT_INSERT;
        procedure SendCTRL_V;
        procedure SendCTRL_C;
        procedure SendCTRL_X;        
        procedure SendCTRL_A;
        procedure SendCTRL_Release;
        procedure SendShift_Release;
        procedure SendALT_Release;
        procedure SendKeyCustom(keys : string);
        procedure SendTAB;
        procedure SendENTER;
        procedure SendINSERT;
        procedure SendDELETE;
        procedure SendBACKSPACE;
        procedure SendHOME;
        procedure SendEND;
        procedure SendUP;
        procedure SendDOWN;
        procedure SendDOWNSpecial;
        procedure SendLEFT;
        procedure SendRIGHT;
        procedure SendESC;
        procedure SendSPACE;
        procedure SendCTRL_SHIFT_LEFT;
        procedure SendCTRL_SHIFT_RIGHT;
        procedure SendCTRL_RIGHT;
        {Used by FrmCONFIG to set pasting method}
        procedure SetClipboardOnly;
        function GetClipboardOnly : boolean;
        procedure SetMimicTyping;
        procedure SetUsePaste;
        procedure SetUsePasteSI;

        {Used by FrmMainPopup}
        procedure SetClipboardOnlyOnce;
        procedure SetKeyboardMimicOnce;
        function GetKeyboardMimicOnce : boolean;
        procedure SetUsePasteCVOnce;
        procedure SetUsePasteSIOnce;
        procedure ClearOnceFlags;
        function GetClipboardOnlyOnce : boolean;

        procedure AssignPaste(EXEName : string; method : TPasteMethod);
        function GetPasteMethod(EXEName : string) : TPasteMethod;
        function GetDefaultPasteMethod : TPasteMethod;
        procedure GetEXEPasteList(var sl : TStringList);

        function IsPasting : boolean;
end;

var Paste : TPaste;

implementation

uses Windows, SysUtils, StrUtils, UnitFrmMainPopup, UnitPopupGenerate, clipbrd,
    UnitFrmClipboardManager, UnitOtherQueue, UnitMisc, UnitToken, Forms, Dialogs,
  UnitFrmConfig, UnitFrmNewEditHistory, UnitFrmSysTrayMenu, UnitFrmPermanentNew,
  UnitFrmSearch, UnitFrmHotKey, UnitKeyboardQuery, UnitFrmEditTextExternal,
  UnitFrmPasteSelected, System.Character;

const  EXEPASTE_FILE = 'exepaste.ini';
{ TPaste }

procedure TPaste.PlaceOnClipboardDontBypassClipboardManager(s: string);
begin
    frmClipboardManager.DisablePasteProtectionOnce;
    // No clearing, do not prevent the Clipboard from detecting the change
    clipboard.SetTextBuf(PChar(s));
end;



constructor TPaste.Create;
var s : string;
begin
    self.EXEPasteList := THashedStringList.Create;
    s := UnitMisc.GetAppPath;
    s := s + EXEPASTE_FILE;
    if FileExists(s) then begin
        self.EXEPasteList.LoadFromFile(s);
    end;

    VariableStart := TStringHash.Create;
    self.DefineVariableStart('date', pv_date);
    self.DefineVariableStart('wait', pv_wait);
    self.DefineVariableStart('prompt', pv_prompt);
    self.DefineVariableStart('popupitem', pv_popupitem);
    self.DefineVariableStart('clipbrd', pv_clipbrd);
    self.DefineVariableStart('CLIPBRDCURRENT', pv_clipbrdcurrent);
    self.DefineVariableStart('clear', pv_clear);
    self.DefineVariableStart('clearpopup', pv_clearpoup);
    self.DefineVariableStart('windowconfig', pv_windowconfig);
    self.DefineVariableStart('windowremoved', pv_windowremoved);
    self.DefineVariableStart('windowhistory', pv_windowhistory);
    self.DefineVariableStart('windowclipboard', pv_windowclipboard);
    self.DefineVariableStart('windowpermanent', pv_windowpermanent);
    self.DefineVariableStart('windowsearch', pv_windowsearch);
    self.DefineVariableStart('selectall', pv_selectall);
    self.DefineVariableStart('selectleft', pv_selectleft);
    self.DefineVariableStart('selectright', pv_selectright);
    self.DefineVariableStart('windowpastesel', pv_windowpastesel);
    self.DefineVariableStart('clipboardonly', pv_clipboardonly);
    self.DefineVariableStart('left', pv_left);
    self.DefineVariableStart('right', pv_right);
    self.DefineVariableStart('up', pv_up);
    self.DefineVariableStart('down', pv_down);
    self.DefineVariableStart('back', pv_backspace);
    self.DefineVariableStart('home', pv_home);
    self.DefineVariableStart('end', pv_end);
    self.DefineVariableStart('tab', pv_tab);
    self.DefineVariableStart('del', pv_delete);
    self.DefineVariableStart('insert', pv_insert);
    self.DefineVariableStart('space', pv_space);
    self.DefineVariableStart('enter', pv_enter);
    self.DefineVariableStart('key', pv_key);
    self.DefineVariableStart('run', pv_run);
    self.DefineVariableStart('copy', pv_copy);
    self.DefineVariableStart('cut', pv_cut);
    Self.DefineVariableStart('clipboardfind', pv_clipboardfind);
    Self.DefineVariableStart('deleteclip', pv_deletepopupclip);
    self.DefineVariableStart('trimclipboard', pv_trimclipboard);
    self.DefineVariableStart('clipboardupper', pv_clipboardupper);
    self.DefineVariableStart('clipboardlower', pv_clipboardlower);
    self.DefineVariableStart('clipboardcapwords', pv_clipboardcapwords);
    self.DefineVariableStart('clipboardinverse', pv_clipboardinverse);
end;

destructor TPaste.Destroy;
var s : string;
begin
    s := UnitMisc.GetAppPath;
    s := s + EXEPASTE_FILE;
    self.EXEPasteList.SaveToFile(s);
    MyFree(EXEPasteList);
end;


//
// Either send text S or use ClipItem if specificed
//
function TPaste.IsPasting : boolean;
begin
    result := self.fIsPasting or self.fIsPastingKeystrokes;
end;
procedure TPaste.SendText(s: string; ci : TClipItem = nil; NoPlaintext : boolean = false);
    procedure SendUsingKeyboardMimic(s : string);
    var c : char;
        w : smallint;
        i : integer;
        delay : Integer;
        ShiftPressed, EnterPressed, AltPressed, CtrlPressed : boolean;
    begin
        //UnitMisc.AppendLog('mimic typing');
        delay := FrmConfig.GetMimicDelayMS;
        for i := 1 to length(s) do begin
            c := s[i];
            if c=#0 then BREAK; // issue with nulls showing up in text

            w := VkKeyScan(c);
            ShiftPressed := (hi(w) and 1) > 0;
            CtrlPressed := (hi(w) and 2) > 0;   {this is for German/Swiss keyboards with an AltGr (CTLR+ALT) keystroke}
            AltPressed := (hi(w) and 4) > 0;


            EnterPressed := (byte(c) = VK_RETURN);
            {VkKeyScan: The first bit of the hi byte set means shift is pressed}
            {Ditch LF - assume CR came first}
            if (c <> #10) then begin
                if ShiftPressed and (not EnterPressed) then begin
                    keybd_event(VK_SHIFT, 0,0,0);
                end;
                if AltPressed and (not EnterPressed) then begin
                    keybd_event(VK_MENU, 0, 0, 0);
                end;
                //sleep(1);
                if CtrlPressed and (not EnterPressed) then begin
                    keybd_event(VK_CONTROL , 0, 0, 0);
                end;
                //sleep(1);

                {Press and release key}
                keybd_event(lo(w), 0, 0, 0);
                sleep(delay);
                keybd_event(lo(w), 0, KEYEVENTF_KEYUP, 0);


                if CtrlPressed and (not EnterPressed) then begin
                    keybd_event(VK_CONTROL, 0, KEYEVENTF_KEYUP, 0);
                end;

                if AltPressed and (not EnterPressed) then begin
                    keybd_event(VK_MENU, 0, KEYEVENTF_KEYUP, 0);
                end;
                if ShiftPressed and (not EnterPressed) then begin
                    //keybd_event(VK_RSHIFT, 0, KEYEVENTF_KEYUP, 0);
                     keybd_event(VK_SHIFT, 0, KEYEVENTF_KEYUP, 0);
                end;

            end;

            if (i mod 10 = 0) then sleep(1); // give the keyboard buffer a little break
        end;
    end;

    procedure PutOnClipboard;
    begin
        clipboard.Clear; // <-- Makes use OWNER and flags the data
                         // to be ignored by us
        clipboard.SetTextBuf(PChar(s));
    end;

    procedure SendUsingPaste(s : string);
    begin
        //
        // place text on clipboard and paste via CTRL+P
        //

        UnitMisc.AppendLog('clearing and placing selected text on clipboard');
        UnitMisc.AppendLog('---"' + s + '"---');
        PutOnClipboard;

        sleep(10);
        if (self.PasteSIOnce) then begin
            Self.SendSHIFT_INSERT;
        end else if (self.PasteCVOnce) then begin
            Self.SendCTRL_V;
        end else if (self.UsePastingCV)  then begin
            Self.SendCTRL_V;
        end else if (self.UsePastingSI) then begin
            Self.SendSHIFT_INSERT;
        end;
    end;

    function PutClipOnCLipboard(ci : TClipItem; Format : word = 0) : boolean;
    var duph : THandle;
        dups : cardinal;
        pc : PChar;
        s : string;
        st : TStream;

        procedure TextOnClipboard(s : ansistring);
        begin
            dups := length(s) + 1;
            //dups := StrSize(s+#0) + 1;
            duph := Windows.GlobalAlloc(GMEM_MOVEABLE or GMEM_ZEROINIT, dups);
            pc := Windows.GlobalLock(duph);
            Windows.MoveMemory(pc, pansichar(s+#0), dups);
            Windows.GlobalUnlock(duph);
            Windows.SetClipboardData(CF_TEXT, duph);

        end;
    var rawstring : RawByteString;
    begin
        result := false;

        if (Format = CF_TEXT) then begin
            s := ci.GetAsPlaintext +#0;
            duph := UnitMisc.DupPointerToHandle(@s[1], length(s) * SizeOf(Char) );
            Windows.SetClipboardData(CF_UNICODETEXT, duph);
        end else begin
            UnitMisc.AppendLog('SendText: format = ' + ci.GetFormatName );
            st := ci.GetStream;
            duph := 0;
            if ci.GetFormatType = FT_UNICODE then begin
                s := ci.GetAsPlaintext + #0;
                duph := UnitMisc.DupPointerToHandle(@s[1], length(s) * SizeOf(Char) );
            end else if (st.size <> 0) then begin
                duph := UnitMisc.DupStreamToHandle(st);
            end else begin
                UnitMisc.AppendLog('SendText: Missing Pointer/Unicode string');
            end;
            ci.FinishedWithStream;

            if (duph <> 0) then begin
                Windows.SetClipboardData(ci.GetFormat, duph);
            end else begin
                UnitMisc.AppendLog('SendText: Unable to dup ClipItem to place on clipboard');
                EXIT;
            end;
        end;

        result := true;
    end;
begin
    if (s = '') and (ci = nil) then begin
        ClipboardOnlyOnce := false;
        KeyboardMimicOnce := false;
        self.PasteCVOnce := false;
        self.PasteSIOnce := false;
        EXIT;
    end;
    UnitMisc.AppendLog('[Paste Start]');
try
    self.fIsPasting := true;
    s := s + #0;
    UnitMisc.MySleep(FrmConfig.GetPasteDelay);

    // send text using configured method
    // Update: ClipboardOnlyWindow will override and force
    //          a clipboard only operation
    if (ci = nil) then begin
        UnitMisc.AppendLog('Plain text version');
        if (self.ClipboardOnlyOnce) then begin
            UnitMisc.AppendLog('Paste: Clipboard Once');
            self.ClipboardOnlyOnce := false;
            Windows.EmptyClipboard;
            PutOnClipboard;
        end else if (self.UseKeyboardMimic or self.KeyboardMimicOnce) then begin
            UnitMisc.AppendLog('Paste: Mimic');
            SendUsingKeyboardMimic(s);
            self.KeyboardMimicOnce := false;
        end else if (self.UsePastingCV or self.UsePastingSI
            or self.PasteCVOnce or self.PasteSIOnce) then  begin
            Windows.EmptyClipboard;
            UnitMisc.AppendLog('Paste: default mode');
            SendUsingPaste(s);
            self.PasteCVOnce := false;
            self.PasteSIOnce := false;
        end else begin
            // send clipboard only
            Windows.EmptyClipboard;
            PutOnClipboard;
            UnitMisc.AppendLog('Paste: Clipboard Only');
        end;
    end else begin
        UnitMisc.AppendLog('complex text version');
        //
        // No calls to unit Clipbrd can be used to in this section
        // since it will try to open and already open clipboard.
        // Place both the plain text version and the complex item on the clipboard
        //
        Windows.OpenClipboard(Application.Handle);
        Windows.EmptyClipboard;
        //Clipboard.Clear; // <- Makes use owner
        if not PutClipOnCLipboard(ci) then begin
            Windows.CloseClipboard;
            EXIT;
        end;

        if not noPlaintext then begin
            if ci.HasText and (ci.GetFormat <> Windows.CF_DIB)  then begin
                if not PutClipOnCLipboard(ci, CF_TEXT) then begin
                    Windows.CloseClipboard;
                    EXIT;
                end;
            end;
        end;

        Windows.CloseClipboard;



        // check the PasteOnce options first
        // they override normal pasting settings

        if (self.ClipboardOnlyOnce) then begin
            self.ClipboardOnlyOnce := false;
            UnitMisc.AppendLog('Paste: Clipboard Only');
        end else if (self.PasteCVOnce) then  begin
            self.SendCTRL_V;
            self.PasteCVOnce := false;
            UnitMisc.AppendLog('Paste: CTRL+V Once');
        end else if (self.PasteSIOnce) then  begin
            self.SendSHIFT_INSERT;
            self.PasteSIOnce := false;
            UnitMisc.AppendLog('Paste: Shift Insert once');
        end else if (self.KeyboardMimicOnce) then begin
            if (ci.HasText) then begin
                SendUsingKeyboardMimic(ci.GetAsPlaintext);
                UnitMisc.AppendLog('Paste: Mimic Type');
            end else begin
                // I can't mimic the clipitem - default to ctrl+v
                self.SendCTRL_V;
                UnitMisc.AppendLog('Paste: CTRL+V');
            end;
            self.KeyboardMimicOnce := false;
        end else if (self.UsePastingCV) then  begin
            self.SendCTRL_V;
            self.PasteCVOnce := false;
            UnitMisc.AppendLog('Paste: CTRL+V');
        end else if (self.UsePastingSI) then  begin
            self.SendSHIFT_INSERT;
            self.PasteSIOnce := false;
            UnitMisc.AppendLog('Paste: Shift Insert');
        end  else if (self.UseKeyboardMimic) then begin
            if (ci.HasText) then begin
                SendUsingKeyboardMimic(ci.GetAsPlaintext);
                UnitMisc.AppendLog('Paste: Mimic Typing');
            end else begin
                // I can't mimic the clipitem - default to ctrl+v
                self.SendCTRL_V;
                UnitMisc.AppendLog('Paste: CTRL+V');
            end;
            self.KeyboardMimicOnce := false;
        end;

    end;
finally
    self.fIsPasting := false;
end;


    UnitMisc.AppendLog('[Paste End]');
    //frmClipboardManager.InformOfPaste;
    //frmClipboardManager.SetIgnoreClipboard(false);
end;


function SetKeyUp(key : byte; up : boolean = true) : boolean;
begin
    result := false;
    if up then begin
        result := KeyboardQuery.IsPressed(key);
        if result  then begin
            case key of
                VK_CONTROL:
                begin
                    if Keyboardquery.IsPressed(VK_LCONTROL) then begin
                        keybd_event(VK_LCONTROL,0,KEYEVENTF_KEYUP, 0);
                    end else begin
                        keybd_event(VK_RCONTROL,0,KEYEVENTF_KEYUP, 0);
                    end;
                    sleep(15);
                end;
                VK_SHIFT:
                begin
                    if KeyboardQuery.IsPressed(VK_LSHIFT) then begin
                        keybd_event(VK_LSHIFT,0,KEYEVENTF_KEYUP, 0);
                    end else begin
                        keybd_event(VK_RSHIFT,0,KEYEVENTF_KEYUP, 0);
                    end;
                    sleep(15);
                end;
                VK_MENU:
                begin
                    keybd_event(VK_MENU,0,KEYEVENTF_KEYUP, 0);
                    if KeyboardQuery.IsPressed(VK_LMENU) then begin
                        keybd_event(VK_LMENU,0,KEYEVENTF_KEYUP, 0);
                    end else begin
                        keybd_event(VK_RMENU,0,KEYEVENTF_KEYUP, 0);
                    end;
                    sleep(16);
                end;
            else
                begin
                    keybd_event(key,key,KEYEVENTF_KEYUP, 0);
                    sleep(10);
                end;
            end;
        end;
    end else begin
        keybd_event(key,0,0, 0);
        sleep(10);
    end;
end;

//
// used by SendText and OtherItemClickEvent to simulate the user
// pressing CTRL+V to paste
//
procedure TPaste.SendCTRL_V;
var w : word;
begin
    UnitMisc.AppendLog('sending CTRL+V');

    // clear any phantom keystrokes
    keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);

    // press CTRL, press V, release V, release CTRL
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), 0, 0);
    sleep(10);
            w := VkKeyScan('V');
            keybd_event(lo(w), w, 0, 0);
            sleep(50);

            keybd_event(lo(w), w, KEYEVENTF_KEYUP, 0);
            sleep(50);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent CTRL+V');
end;
procedure TPaste.SendCTRL_C;
var w : word;
begin
    UnitMisc.AppendLog('sending CTRL+C');

    // clear any phantom keystrokes
    keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);

    // press CTRL, press V, release V, release CTRL
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), 0, 0);
    sleep(10);
            w := VkKeyScan('C');
            keybd_event(lo(w), w, 0, 0);
            sleep(10);

            keybd_event(lo(w), w, KEYEVENTF_KEYUP, 0);
            sleep(10);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    //sleep(10);
    UnitMisc.AppendLog('sent CTRL+C');
end;
procedure TPaste.SendCTRL_X;
var w : word;
begin
    UnitMisc.AppendLog('sending CTRL+X');

    // clear any phantom keystrokes
    keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);

    // press CTRL, press V, release V, release CTRL
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), 0, 0);
    sleep(10);
            w := VkKeyScan('X');
            keybd_event(lo(w), w, 0, 0);
            sleep(10);

            keybd_event(lo(w), w, KEYEVENTF_KEYUP, 0);
            sleep(10);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    //sleep(10);
    UnitMisc.AppendLog('sent CTRL+X');
end;
procedure TPaste.SendCTRL_A;
var w : word;
begin
    UnitMisc.AppendLog('sending CTRL+A');

    // clear any phantom keystrokes
    keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);

    // press CTRL, press V, release V, release CTRL
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), 0, 0);
    sleep(10);
            w := VkKeyScan('A');
            keybd_event(lo(w), w, 0, 0);
            sleep(50);

            keybd_event(lo(w), w, KEYEVENTF_KEYUP, 0);
            sleep(50);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent CTRL+A');
end;

procedure TPaste.SendCTRL_SHIFT_RIGHT;
var w : word;
begin
     // clear any phantom keystrokes
    keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);

    // press CTRL, press V, release V, release CTRL
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), 0, 0);
    sleep(1);
    keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)),0,0);  // win9x
    sleep(10);
            //w := VkKeyScan('C');
            w := VK_RIGHT;
            keybd_event(lo(w), w, KEYEVENTF_EXTENDEDKEY, 0);
            sleep(10);

            keybd_event(lo(w), w, KEYEVENTF_EXTENDEDKEY or KEYEVENTF_KEYUP, 0);
            sleep(10);
    keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);

    //sleep(10);
    UnitMisc.AppendLog('sent CTRL+SHIFT+LEFT');
end;
procedure TPaste.SendCTRL_SHIFT_LEFT;
var w : word;
begin
     // clear any phantom keystrokes
    keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);

    // press CTRL, press V, release V, release CTRL
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), 0, 0);
    sleep(1);
    keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)),0,0);  // win9x
    sleep(10);
            //w := VkKeyScan('C');
            w := VK_LEFT;
            keybd_event(lo(w), w, KEYEVENTF_EXTENDEDKEY, 0);
            sleep(10);

            keybd_event(lo(w), w, KEYEVENTF_EXTENDEDKEY or KEYEVENTF_KEYUP, 0);
            sleep(10);
    keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);

    //sleep(10);
    UnitMisc.AppendLog('sent CTRL+SHIFT+LEFT');
end;

procedure TPaste.SendCTRL_Release;
begin
	SetKeyUp(VK_CONTROL, true);
end;
procedure TPaste.SendShift_Release;
begin
	SetKeyUp(VK_SHIFT, true);
end;
procedure TPaste.SendALT_Release;
begin
    SetKeyUp(VK_MENU, true);
end;

procedure TPaste.SendKeyCustom(keys: string);
var w : word;
    hk : THotkeydata;
var flags : cardinal;
begin

// These keys require are known to require KEYEVENTF_EXTENDEDKEY
// like for SHIFT+HOME and SHIFT+END
//
//HOME & END
//Left Win 		E0 5B
//Right Win 		E0 5C
//Application 	E0 5D
//ACPI Power 	E0 5E
//ACPI Sleep 	E0 5F
//ACPI Wake 	E0 63
//Numeric / 	E0 35
//Up Arrow 	E0 48
//Dn Arrow 	E0 50
//Page Up 	E0 49
//Page Down 	E0 51
//R Arrow 	E0 4D
//L Arrow 	E0 4B


    hk := FrmHotkey.FromString(keys);
    // unpress mod keys
   { keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_MENU , VkKeyScan(char(VK_MENU)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_LWIN, VkKeyScan(char(VK_MENU)), KEYEVENTF_KEYUP, 0);}
    sleep(10);



    // simulate modifiers down
    if hk.win then
        keybd_event(VK_LWIN, VkKeyScan(char(VK_LWIN)), 0, 0);
    sleep(10);
    if hk.alt then
        keybd_event(VK_MENU, VkKeyScan(char(VK_MENU)), 0, 0);
    sleep(10);
    if hk.ctrl then
        keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), 0, 0);
    sleep(10);
    if hk.shft  then
        keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)),0,0);  // win9x
    sleep(10);

            // main key press
            //w := VkKeyScan('C');
            w := hk.key;

            flags := 0;
            if (hk.key >= ord('A')) and (hk.key <= ord('Z')) then begin
                w := VKKeyScan(char(lo(w)));
            end;
            if hk.key in [VK_LEFT, VK_RIGHT, VK_HOME, VK_END, VK_NEXT, VK_PRIOR ] then
                flags := KEYEVENTF_EXTENDEDKEY;

            keybd_event(lo(w), w, flags, 0);
            sleep(10);

            keybd_event(lo(w), w, flags or KEYEVENTF_KEYUP, 0);
            sleep(10);
    //keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)), KEYEVENTF_KEYUP, 0);
    //keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);

    // release modifiers
    if hk.shft then
        keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    if hk.ctrl then
        keybd_event(VK_CONTROL, VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    if hk.alt then
        keybd_event(VK_MENU, VkKeyScan(char(VK_MENU)), KEYEVENTF_KEYUP, 0);
    if hk.win then
        keybd_event(VK_LWIN, VkKeyScan(char(VK_LWIN)), KEYEVENTF_KEYUP, 0);

    //sleep(10);
    UnitMisc.AppendLog('sent Kestroke');
end;


procedure TPaste.SendSHIFT_INSERT;
begin
    UnitMisc.AppendLog('sending SHIFT+INSERT');

    // clear any phantom keystrokes
    keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);

    // press SHIFT, press INSERT, release INSERT, release SHIFT

    // Label VK_INSERT as extended, othewise it freaks out when combined with SHIFT

    keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)),0,0);  // win9x
    sleep(1);
            keybd_event(VK_INSERT, Lo(MapVirtualKey(VK_INSERT,0)),KEYEVENTF_EXTENDEDKEY, 0);
            sleep(1);

            keybd_event(VK_INSERT, Lo(MapVirtualKey(VK_INSERT,0)),KEYEVENTF_EXTENDEDKEY or KEYEVENTF_KEYUP,0);
            sleep(1);

    keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)), KEYEVENTF_KEYUP, 0);
    sleep(5);
    UnitMisc.AppendLog('sent SHIFT+INSERT');
end;
procedure TPaste.SendCTRL_RIGHT;
var w : word;
begin
     // clear any phantom keystrokes
    keybd_event(VK_SHIFT, VkKeyScan(char(VK_SHIFT)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);
    sleep(10);

    // press CTRL, press V, release V, release CTRL
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), 0, 0);
    sleep(1);
    //keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)),0,0);  // win9x
    sleep(10);
            //w := VkKeyScan('C');
            w := VK_RIGHT;
            keybd_event(lo(w), w, KEYEVENTF_EXTENDEDKEY, 0);
            sleep(10);

            keybd_event(lo(w), w, KEYEVENTF_EXTENDEDKEY or KEYEVENTF_KEYUP, 0);
            sleep(10);
    //keybd_event(VK_SHIFT, Lo(MapVirtualKey(VK_SHIFT,0)), KEYEVENTF_KEYUP, 0);
    keybd_event(VK_CONTROL , VkKeyScan(char(VK_control)), KEYEVENTF_KEYUP, 0);

    //sleep(10);
    UnitMisc.AppendLog('sent CTRL+SHIFT+LEFT');

end;

procedure TPaste.SendTAB;
begin
    UnitMisc.AppendLog('sending tab');
    keybd_event(VK_TAB , 0, 0, 0);
    sleep(30);

    keybd_event(VK_TAB, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent tab');
end;
procedure TPaste.SendDELETE;
begin
    UnitMisc.AppendLog('sending del');
    keybd_event(VK_DELETE, 0, 0, 0);
    sleep(30);

    keybd_event(VK_DELETE, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent del');

end;
procedure TPaste.SendINSERT;
begin
    UnitMisc.AppendLog('sending INSERT');
    keybd_event(VK_INSERT, 0, 0, 0);
    sleep(30);

    keybd_event(VK_INSERT, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent INSERT');
end;
procedure TPaste.SendENTER;
begin
    UnitMisc.AppendLog('sending ENTER');
    keybd_event(VK_RETURN, MapVirtualKey(VK_RETURN,0), 0, 0);
    sleep(30);

    keybd_event(VK_RETURN, MapVirtualKey(VK_RETURN,0), KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent ENTER');
end;
procedure TPaste.SendSPACE;
begin
    UnitMisc.AppendLog('sending tab');
    keybd_event(VK_SPACE , 0, 0, 0);
    sleep(30);

    keybd_event(VK_SPACE, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sending space');
end;

procedure TPaste.ClearMethods;
begin
    self.UseKeyboardMimic := false;
    self.UsePastingSI := false;
    self.UsePastingCV := false;
    self.UseClipboardOnly := false;
end;
procedure TPaste.SetClipboardOnly;
begin
    self.ClearMethods;
    self.UseClipboardOnly := true;
end;
procedure TPaste.SetMimicTyping;
begin
    self.ClearMethods;
    self.UseKeyboardMimic := true;
end;
procedure TPaste.SetUsePaste;
begin
    self.ClearMethods;
    self.UsePastingCV := true;
end;
procedure TPaste.SetUsePasteSI;
begin
    self.ClearMethods;
    self.UsePastingSI := true;
end;

procedure TPaste.SetKeyboardMimicOnce;
begin
    ClearOnceFlags;
    self.KeyboardMimicOnce := true;
end;
procedure TPaste.SetUsePasteCVOnce;
begin
    ClearOnceFlags;
    self.PasteCVOnce := true;
end;
procedure TPaste.SetUsePasteSIOnce;
begin
    ClearOnceFlags;
    self.PasteSIOnce := true;
end;
procedure TPaste.SetClipboardOnlyOnce;
begin
    ClearOnceFlags;
    AppendLog('[Paste]: ClipboardOnce set');
    self.ClipboardOnlyOnce := true;
end;
function TPaste.GetClipboardOnlyOnce : boolean;
begin
    result := self.ClipboardOnlyOnce;
end;
function TPaste.GetClipboardOnly: boolean;
begin
    result := self.UseClipboardOnly;
end;



{
//
// Not used, but may need to be refered to later
//
procedure TFrmMainPopup.PlaceTextOnClipboard(s : string);
    procedure SetBuffer(format: Word; var buffer; Size: Integer);
    var Data: THandle;
        DataPtr: Pointer;
    begin
        clipboard.Open;
        try
            Data := Windows.GlobalAlloc(GMEM_MOVEABLE+GMEM_DDESHARE, Size);
            try
                DataPtr := GlobalLock(Data);
                try
                    Move(Buffer, DataPtr^, Size);
                    clipboard.Clear;
                    Windows.SetClipboardData(Format, Data);
                finally
                    Windows.GlobalUnlock(Data);
                end;
            except
                Windows.GlobalFree(Data);
                raise;
            end;
        finally
            clipboard.Close;
        end;
    end;

begin

    s := s + #0 + 'ArsClip' + #0;
    SetBuffer(CF_TEXT, PChar(s)^, length(s) + 1);
end;
}



function TPaste.GetKeyboardMimicOnce: boolean;
begin
    result := self.UseKeyboardMimic;
end;

type TKeyProc = procedure;

type TNugget = class(TObject)
    vartype : TPasteVariable;
    vartext : string;
end;

procedure TPaste.SendTextWithKeystrokes(maintext: string);
var pastestr : string;
    key, clip : string;
    prompts : TStringList;
    ShowSearchWindow, MimicActivate : boolean;
const SLEEPMS = 60;

    procedure PreKey(pastestr : string);
    var b : boolean;
    begin
        if pastestr <> '' then begin
            frmClipboardManager.IgnoreClipboardOnce;
            b := self.ClipboardOnlyOnce;
            if MimicActivate then self.KeyboardMimicOnce := true;
            self.SendText(pastestr);
            self.ClipboardOnlyOnce := b;
            sleep(SLEEPMS);
        end;
    end;
    procedure PostKey;
    begin
        Application.ProcessMessages;
        pastestr := '';
    end;
    function url_encode(url : string) : string;
    var
    	i: integer;
    begin
        result:='';
        for i := 1 to length(url) do begin
        case url[i] of
        'a'..'z','A'..'Z','0'..'9','/','.','&','-' :
        	result:=result+ url[i];
        else
        	result := result+'%'+uppercase(inttohex(ord(url[i]),2));
        end;
        end;
    end;
    procedure ExecuteNugget(var i : integer; nugget : TNugget; nuggets : TList);
    var
        pastestr, temptoken, temptoken2, before : string;
        j : Integer;
        h : THandle;
    begin
        case nugget.vartype of
        pv_NONE: begin  // literal text -- concat all text nuggets into a single item
            pastestr := nugget.vartext;
            while (i+1 < nuggets.count) and
                (TNugget(nuggets[i+1]).vartype in
                [pv_NONE, pv_clipbrdcurrent, pv_clipbrd]) do begin
                Inc(i);
                nugget := TNugget(nuggets[i]);

                if nugget.vartype = pv_clipbrdcurrent then begin
                    pastestr := pastestr + Clipboard.AsText;
                end else begin
                    pastestr := pastestr + nugget.vartext;
                end;
            end;
            PreKey(pastestr);
        end;
        pv_run: begin   // run all text items afterwords
            temptoken := '';
            while (i+1 < nuggets.count) and
                (TNugget(nuggets[i+1]).vartype in
                [pv_NONE, pv_clipbrdcurrent, pv_clipbrd]) do begin
                Inc(i);
                nugget := TNugget(nuggets[i]);
                if (nugget.vartype = pv_clipbrdcurrent) then begin
                    nugget.vartext := Clipboard.AsText;
                end;
                // URL encode clipboard items if the run command is a URL (http:)
                if (nugget.vartype in [pv_clipbrdcurrent, pv_clipbrd]) and
                    AnsiStartsText('http:', temptoken) then begin
                    temptoken := temptoken +url_encode(nugget.vartext);
                end else begin
                    temptoken := temptoken + nugget.vartext;
                end;
            end;
            UnitMisc.ShellExecute(FrmMainPopup.handle, temptoken);
        end;
        pv_wait: begin //WAIT or WAIT=100
            j := 100;
            if ContainsText(temptoken, '=') then begin
                TokenString(nugget.vartext,'=');
                temptoken := TokenString(nugget.vartext,']');
                try
                    j := StrToInt(temptoken);
                except
                end;
            end;
            sleep(j);
        end;
        pv_prompt: begin    //PROMPT="Prompt Text"
            TokenString(nugget.vartext,'"');
            temptoken := TokenString(nugget.vartext,'"');
            pastestr := prompts.Values[LowerCase(temptoken)];
            if pastestr = '' then begin
                h := GetForegroundWindow;
                ForceForeground(FrmMainPopup.Handle);
                pastestr := InputBox('ArsClip Prompt', temptoken, '');
                if h<>0 then ForceForeground(h);

                prompts.Values[LowerCase(temptoken)] := pastestr;
            end;
            if pastestr <> '' then begin
                PreKey(pastestr);
            end;
        end;
        pv_clipboardfind: begin //CLIPBOARDFIND="" REPLACE=""
            // detect \n or \t
            TokenString(nugget.vartext, '"', false);
            temptoken := TokenString(nugget.vartext, '"', false);
            TokenString(nugget.vartext, '"', false);
            temptoken2 := TokenString(nugget.vartext, '"', false);

            pastestr := Clipboard.AsText;
            before := pastestr;
            // auto-detect Unix style or Windows style linefeeds
            if ContainsStr(temptoken,'\n') then begin
                if ContainsStr(pastestr, #10) then begin
                    if ContainsStr(pastestr, #13) then begin
                        temptoken := StringReplace(temptoken,'\n',#13#10,[rfReplaceAll, rfIgnoreCase]);
                    end else begin
                        temptoken := StringReplace(temptoken,'\n',#10,[rfReplaceAll, rfIgnoreCase]);
                    end;
                end;
            end;
            temptoken2 := StringReplace(temptoken2,'\t', #9, [rfReplaceAll, rfIgnoreCase]);
            temptoken2 := StringReplace(temptoken2,'\n', #13#10, [rfReplaceAll, rfIgnoreCase]);

            pastestr := StringReplace(pastestr,temptoken,temptoken2,[rfReplaceAll, rfIgnoreCase]);

            if before <> pastestr then
                Clipboard.SetTextBuf(@pastestr[1]);
        end;
        pv_clipboardupper: begin
            pastestr := UpperCase(Clipboard.AsText);
            Clipboard.SetTextBuf(@pastestr[1]);
        end;
        pv_clipboardlower: begin
            before := Clipboard.AsText;
            pastestr := LowerCase(before);
            if before <> pastestr then
                Clipboard.SetTextBuf(@pastestr[1]);
        end;
        pv_clipboardcapwords: begin
            before := Clipboard.AsText;
            pastestr := UnitToken.CapitalizeWords(before);
            if before <> pastestr then
                Clipboard.SetTextBuf(@pastestr[1]);
        end;
        pv_clipboardinverse: begin
            pastestr := Clipboard.AsText;
            before := pastestr;
            for j := 1 to length(pastestr) do begin
                if TCharacter.IsUpper(pastestr[j]) then begin
                    pastestr[j] := TCharacter.ToLower(pastestr[j]);
                end else begin
                    pastestr[j] := TCharacter.ToUpper(pastestr[j]);
                end;
            end;
            if before <> pastestr then
                Clipboard.SetTextBuf(@pastestr[1]);
        end;
        pv_clipboardonly: begin end; // not used
        pv_popupitem: begin end;
        pv_deletepopupclip: begin //
            TokenString(nugget.vartext, '=');
            try
                j := StrToInt(nugget.vartext);
                if j < ClipQueue.GetQueueCount then begin
                    ClipQueue.DeleteItem(j);
                end;
            except
            end;
        end;
        pv_key: begin   //KEY="datahere"
            TokenString(nugget.vartext, '"');
            pastestr := TokenString(nugget.vartext, '"');
            self.SendKeyCustom(pastestr);
        end;
        pv_clipbrd: begin end;
        pv_clipbrdcurrent: begin
            sleep(10); // breathing room for any previous COPY command
            pastestr := clipboard.AsText;
            PreKey(pastestr);
        end;
        pv_trimclipboard: begin
            frmClipboardManager.IgnoreClipboardOnce;  // TODO: Is this needed?
            before := Clipboard.AsText;
            pastestr := Trim(before);
            if before <> pastestr then
                Clipboard.SetTextBuf(@pastestr[1]);
        end;
        pv_clear: frmClipboardManager.ClearClipboard;
        pv_clearpoup: ClipQueue.ClearQueue;
        pv_windowconfig: begin
            FrmConfig.Show;
            UnitMisc.ForceForeground(FrmConfig.Handle);
        end;
        pv_windowremoved: begin
            frmEditHistory.ShowRemoved;
            UnitMisc.ForceForeground(frmEditHistory.Handle);
        end;
        pv_windowhistory: begin
            frmEditHistory.ShowRemoved;
            UnitMisc.ForceForeground(frmEditHistory.Handle);
        end;
        pv_windowclipboard: begin
            frmSysTrayMenu.EditClipboardMenuItemClickEvent(nil);
        end;
        pv_windowpermanent: begin
            frmPermanent.Show;
            UnitMisc.ForceForeground(frmPermanent.Handle);
        end;
        pv_windowsearch: begin
            ShowSearchWindow := true;
        end;
        pv_windowpastesel: FrmPasteSelected.ShowaAutomated;
        pv_selectall: Self.SendCTRL_A;
        pv_selectleft: self.SendCTRL_SHIFT_LEFT;
        pv_selectright: self.SendCTRL_SHIFT_RIGHT;
        pv_left: self.SendLEFT;
        pv_right: self.SendRIGHT;
        pv_up: self.SendUP;
        pv_down: self.SendDOWN;
        pv_backspace: self.SendBACKSPACE;
        pv_home: self.SendHOME;
        pv_end: self.SendEND;
        pv_tab: self.SendTAB;
        pv_delete: self.SendDELETE;
        pv_insert: Self.SendINSERT;
        pv_space: self.SendSPACE;
        pv_enter: self.SendENTER;
        pv_cut: self.SendCTRL_X;
        pv_copy: Self.SendCTRL_C;
        end;
    end;


var
    i, j : integer;
    temps, temptoken,
    varname, vartext, literaltext, datastr
    : string;


    h : THandle;
    disablemontoring, monitoring : Boolean;

    nugget : TNugget;
    nuggets : TList;
    nuggettype : TPasteVariable;
begin

try
    self.fIsPastingKeystrokes := true;
    ShowSearchWindow := false;
    UnitMisc.MySleep(FrmConfig.GetPasteDelay);

    // key rid of '[KEYS]'
    maintext := rightstr(maintext,length(maintext) - 6);
    prompts := TStringList.Create;

    if pos('[CLIPBOARDONLY]',maintext) <> 0 then begin
        self.ClipboardOnlyOnce := true;
    	maintext := StringReplace(maintext, '[CLIPBOARDONLY]','',[rfIgnoreCase, rfReplaceAll]);
	end;

    if pos('[MIMIC]',maintext) <> 0 then begin
        MimicActivate := true;
    	maintext := StringReplace(maintext, '[MIMIC]','',[rfIgnoreCase, rfReplaceAll]);
	end;

    disablemontoring := false;
    if pos('[DISABLEMONITORING]',maintext) <> 0 then begin
        disablemontoring := true;
    	maintext := StringReplace(maintext, '[DISABLEMONITORING]','',[rfIgnoreCase, rfReplaceAll]);
        monitoring := frmClipboardManager.GetMonitoring;
        frmClipboardManager.SetDisableMonitoring(true);
	end;


    //
    // turn remaining text into a list of nuggets commands
    //
    clip := Clipboard.AsText;
    nuggets := TList.Create;
    temps := maintext;
    literaltext := '';
    while temps <> '' do begin
        literaltext := TokenString(temps,'[', false);

        // text ahead of the '[' is always litteral
        // '[' and following may also be litteral if it's not a variable
        i := 0;
        while (i<Length(temps)) and
            CharInSet(LowerCase(temps[i+1])[1], ['a'..'z']) do
            Inc(i);
        varname := Copy(temps,1,i);
        {$REGION 'Handle Variable or Literal Text'}
        if self.IsVariable(varname) then begin
            // create the prefix literal nugget
            if literaltext <> '' then begin
                nugget := TNugget.Create;
                nugget.vartype := pv_NONE;
                nugget.vartext := literaltext;
                literaltext := '';
                nuggets.Add(nugget);
            end;

            nuggettype := self.VariableType(varname);
            vartext := TokenString(temps,']', false);

            // handle the insert variables and convert to literal text
            case nuggettype of
            pv_clipbrd: begin
                nugget := TNugget.Create;
                nugget.vartype := pv_clipbrd;
                nugget.vartext := clip;
                nuggets.Add(nugget);
            end;
            pv_date: begin  //DATE="mm/yy/ddd"
                datastr := TokenString(vartext, '"', false);
                datastr := TokenString(vartext, '"', false);
                nugget := TNugget.Create;
                nugget.vartype := pv_NONE;
                nugget.vartext := '';
                try
                    nugget.vartext := FormatDateTime(datastr, now);
                except
                end;
                nuggets.Add(nugget);
            end;
            pv_POPUPITEM: begin   //POPUPITEM=X
                datastr := TokenString(vartext, '=', false);
                datastr := vartext;
                //datastr := TokenString(vartext, ']"', false);
                nugget := TNugget.Create;
                nugget.vartype := pv_NONE;
                nugget.vartext := '';
                try
                    j := StrToInt(datastr);
                    if j < ClipQueue.GetQueueCount then
                        nugget.vartext := ClipQueue.GetItemText( j );
                except
                end;
                nuggets.Add(nugget);
            end;
            else
                begin
                    nugget := TNugget.Create;
                    nugget.vartype := nuggettype;
                    nugget.vartext := vartext;
                    nuggets.Add(nugget);
                end;
            end;
        end else begin
            if temps <> '' then begin
                literaltext := literaltext + '['; // add back removed
            end;

            if literaltext <> '' then begin
                nugget := TNugget.Create;
                nugget.vartype := pv_NONE;
                nugget.vartext := literaltext;
                literaltext := '';
                nuggets.Add(nugget);
            end;
        end;
        {$ENDREGION}
    end;


    //
    // Execute nugget commands
    //
    i := 0;
    while i < nuggets.count do begin
        nugget := TNugget(nuggets.items[i]);
        nuggettype := nugget.vartype;
        ExecuteNugget(i, nugget, nuggets);
        PostKey;
        inc(i);
    end;

    if ShowSearchWindow then  FrmSearch.ShowAutomatted;
    if self.ClipboardOnlyOnce  then self.ClipboardOnlyOnce := false;
    if disablemontoring then frmClipboardManager.SetDisableMonitoring(not monitoring);
finally
    Self.fIsPastingKeystrokes := false;
end;
end;

procedure TPaste.SendBACKSPACE;
begin
    UnitMisc.AppendLog('sending back');
    keybd_event(VK_BACK , 0, 0, 0);
    sleep(30);

    keybd_event(VK_BACK, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent back');
end;
procedure TPaste.SendEND;
begin
    UnitMisc.AppendLog('sending end');
    keybd_event(VK_END  , 0, 0, 0);
    sleep(30);

    keybd_event(VK_END, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent end');
end;
procedure TPaste.SendHOME;
begin
    UnitMisc.AppendLog('sending VK_HOME');
    keybd_event(VK_HOME  , 0, 0, 0);
    sleep(30);

    keybd_event(VK_HOME, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent VK_HOME');
end;
procedure TPaste.SendDOWN;
begin
     UnitMisc.AppendLog('sending VK_DOWN');
    keybd_event(VK_DOWN   , 0, 0, 0);
    sleep(30);

    keybd_event(VK_DOWN, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent VK_DOWN');
end;
procedure TPaste.SendDOWNSpecial;
var lwindown, rwindown : boolean;


begin
    UnitMisc.AppendLog('sending VK_DOWN');

    lwindown := SetKeyUp(VK_LWIN);
    rwindown := SetKeyUp(VK_RWIN);

    keybd_event(VK_DOWN, 0, 0, 0);
    sleep(10);
    keybd_event(VK_DOWN, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    if lwindown then SetKeyUp(VK_LWIN, false);
    if rwindown then SetKeyUp(VK_RWIN, false);


    UnitMisc.AppendLog('sent VK_DOWN');
end;
procedure TPaste.SendLEFT;
begin
  UnitMisc.AppendLog('sending VK_LEFT');
    keybd_event(VK_LEFT   , 0, 0, 0);
    sleep(30);

    keybd_event(VK_LEFT, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent VK_LEFT');
end;
procedure TPaste.SendRIGHT;
begin
   UnitMisc.AppendLog('sending VK_RIGHT');
    keybd_event(VK_RIGHT   , 0, 0, 0);
    sleep(30);

    keybd_event(VK_RIGHT, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent VK_RIGHT');
end;
procedure TPaste.SendUP;
begin
   UnitMisc.AppendLog('sending VK_UP');
    keybd_event(VK_UP   , 0, 0, 0);
    sleep(30);

    keybd_event(VK_UP, 0, KEYEVENTF_KEYUP, 0);
    sleep(10);
    UnitMisc.AppendLog('sent VK_UP');
end;
procedure TPaste.SendESC;
begin
    keybd_event(VK_ESCAPE  , 0, 0, 0);
    sleep(150);

    keybd_event(VK_ESCAPE, 0, KEYEVENTF_KEYUP, 0);
    sleep(30);
end;

procedure TPaste.ClearOnceFlags;
begin
    PasteCVOnce := false;
    PasteSIOnce := false;
    ClipboardOnlyOnce := false;
    KeyboardMimicOnce := false;
end;

//
// Assigned Paste Methods
//


procedure TPaste.AssignPaste(EXEName: string; method: TPasteMethod);
begin
    EXEPasteList.Values[EXEName] := IntToStr(Integer(method));
end;
function TPaste.GetPasteMethod(EXEName : string) : TPasteMethod;
var s : string;
begin
    result := PASTE_DEFAULT;
    s := EXEPasteList.Values[EXEName];

    if (s <> '') then begin
        result := TPasteMethod(StrToInt(s));
    end;
end;
function TPaste.GetDefaultPasteMethod: TPasteMethod;
begin
    result := PASTE_DEFAULT; // this case should never fire

    if self.UseKeyboardMimic then begin
        result := PASTE_MIMIC;
    end else if self.UsePastingSI then begin
        result := PASTE_SHIFT_INS;
    end else if self.UsePastingCV then begin
        result := PASTE_CTRL_V;
    end else if self.UseClipboardOnly then begin
        result := PASTE_CLIPBOARD
    end;
end;
procedure TPaste.GetEXEPasteList(var sl : TStringList);
begin
    sl.AddStrings(EXEPasteList); 
end;



procedure TPaste.DefineVariableStart(starttext : string; vartype : TPasteVariable);
begin
    if VariableStart.ValueOf(LowerCase(starttext)) = -1 then begin
        VariableStart.Add(LowerCase(starttext), Integer(vartype));
    end;
end;
function TPaste.IsVariable(starttext : string) : boolean;
begin
    result := VariableStart.ValueOf(LowerCase(starttext)) <> -1;
end;
function TPaste.VariableType(starttext : string) : TPasteVariable;
var i : integer;
begin
    result := pv_NONE;
    i := VariableStart.ValueOf(LowerCase(starttext));
    if i <> -1 then
        Result := TPasteVariable(i);
end;



initialization
begin
    UnitMisc.AppendLog('TPaste - creating');
    Paste := TPaste.Create;
end;
finalization
begin
    MyFree(Paste);
end;
end.
